/******************************************************************************

Contents:
    Record the actions in a list and analyse that list
Created:
    Richard Cooper, Sat Aug 15 15:54:44 1998
Public procedures:
    void ActInitialiseList()
    void ActRecord(ActionType act, int cycle, char *arg1, char *arg2)
    void ActAnalyseList(FILE *fp)

*******************************************************************************/

#include "cs.h"

typedef enum a3_label {PREPARE_COFFEE, PREPARE_CEREAL} A3Label;

typedef enum a2_label {ANOMOLOUS_A2, ADD_COFFEE, ADD_SUGAR, ADD_MILK, DRINK_COFFEE, EAT_CEREAL, A2_BREAK, A2_DOUBLE_BREAK} A2Label;

typedef struct action_list {
    ActionType          act;
    int                 cycle;
    char               *arg1;
    char               *arg2;
    A2Label             subtask;
    A3Label             task;
    BOOLEAN             a2_open;
    BOOLEAN             a2_close;
    BOOLEAN             a3_open;
    BOOLEAN             a3_close;
    BOOLEAN             crux;
    BOOLEAN             independent;
    BOOLEAN             correct;
    struct action_list *next;
    struct action_list *prev;
} ActionList;

static ActionList *first_action = NULL;
static ActionList *last_action  = NULL;

/******************************************************************************/
/********* Constructing the action/event list: ********************************/

void ActInitialiseList()
{
    while (first_action != NULL) {
        ActionList *tmp = first_action;
        first_action = first_action->next;
        free(tmp);
    }
    last_action = NULL;
}

void ActRecord(ActionType act, int cycle, char *arg1, char *arg2)
{
    ActionList *new;

    if ((new = (ActionList *)malloc(sizeof(ActionList))) == NULL) {
        fprintf(stderr, "WARNING: Memory allocation failed in ActRecord()\n");
    }
    else {
        new->act = act;
        new->cycle = cycle;
        new->arg1 = arg1;
        new->arg2 = arg2;
        new->next = NULL;
        new->prev = last_action;
    }
    if (last_action == NULL) {
        first_action = new;
        last_action = new;
    }
    else {
        last_action->next = new;
        last_action = new;
    }
}

/******************************************************************************/
/********* Printing the action/event list: ************************************/

static char acs_code(ActionList *action)
{
    switch (action->subtask) {
        case ANOMOLOUS_A2: {
            return(action->crux?'X':'x');
        }
        case ADD_COFFEE: {
            return(action->crux?'G':'g');
        }
        case ADD_SUGAR: {
            return(action->crux?'S':'s');
        }
        case ADD_MILK: {
            return(action->crux?'M':'m');
        }
        case DRINK_COFFEE: {
            return(action->crux?'D':'c');
        }
        case EAT_CEREAL: {
            return(action->crux?'E':'e');
        }
        case A2_BREAK: {
            return('/');
	}
        case A2_DOUBLE_BREAK: {
            return('"');
        }
        default: {
            return(0);
        }
    }
}

static void print_action(FILE *fp, ActionList *action)
{
    if (action->a2_open) {
        fprintf(fp, "(\n");
    }

    fprintf(fp, "%c%c %4d: %c ", (action->correct?' ':'*'), (action->independent?'I':' '), action->cycle, acs_code(action));

    switch (action->act) {
        case NONE: {
            fprintf(fp, "none(%p, %p)\n", action->arg1, action->arg2);
            break;
        }
        case PICK_UP: {
            fprintf(fp, "pick_up(%s)\n", action->arg1);
            break;
        }
        case PUT_DOWN: {
            fprintf(fp, "put_down(%s)\n", action->arg1);
            break;
        }
        case DIP_SPOON: {
            fprintf(fp, "dip_spoon(%s)\n", action->arg1);
            break;
        }
        case EMPTY_SPOON: {
            fprintf(fp, "empty_spoon(%s, %s)\n", action->arg1, action->arg2);
            break;
        }
        case EAT_FROM_SPOON: {
            fprintf(fp, "consume(%s)\n", action->arg1);
            break;
        }
        case STIR: {
            fprintf(fp, "stir(%s)\n", action->arg1);
            break;
        }
        case POUR: {
            fprintf(fp, "pour(%s, %s)\n", action->arg1, action->arg2);
            break;
        }
        case TEAR: {
            fprintf(fp, "tear(%s)\n", action->arg1);
            break;
        }
        case SCREW_OPEN: {
            fprintf(fp, "screw_open(%s)\n", action->arg1);
            break;
        }
        case SCREW_CLOSED: {
            fprintf(fp, "screw_closed(%s)\n", action->arg1);
            break;
        }
        case SWAP: {
            fprintf(fp, "swap\n");
            break;
        }
        case DRINK: {
            fprintf(fp, "drink(%s)\n", action->arg1);
            break;
        }
        case PREPARE: {
            fprintf(fp, "prepare(%s)\n", action->arg1);
            break;
        }
        case PUSH: {
            fprintf(fp, "press(%s)\n", action->arg1);
            break;
        }
        case RELAX: {
            fprintf(fp, "relax(%s)\n", action->arg1);
            break;
        }
        default: {
            fprintf(fp, "default(%p, %p)\n", action->arg1, action->arg2);
            break;
        }
    }

    if (action->a2_close) {
        fprintf(fp, ")\n");
    }
}

static void print_list(FILE *fp)
{
    ActionList *tmp;

    fprintf(fp, "Actions attempted:\n");
    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        print_action(fp, tmp);
    }
}

/******************************************************************************/
/********* Analysing the action/event list: ***********************************/

static BOOLEAN is_implement(char *string)
{
    if (strcmp(string, "teaspoon") == 0) {
        return(TRUE);
    }
    else if (strcmp(string, "stirrer") == 0) {
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

static A2Label corresponding_subtask(ActionList *action)
{
    char *string = action->arg1;

    if (strcmp(string, "sugar") == 0) {
        return(ADD_SUGAR);
    }
    else if (strcmp(string, "sugar bowl") == 0) {
        return(ADD_SUGAR);
    }
    else if (strcmp(string, "sugar packet") == 0) {
        return(ADD_SUGAR);
    }
    else if (strcmp(string, "coffee") == 0) {
        return(ADD_COFFEE);
    }
    else if (strcmp(string, "coffee packet") == 0) {
        return(ADD_COFFEE);
    }
    else if (strcmp(string, "coffee jar") == 0) {
        return(ADD_COFFEE);
    }
    else if (strcmp(string, "milk") == 0) {
        return(ADD_MILK);
    }
    else if (strcmp(string, "milk carton") == 0) {
        return(ADD_MILK);
    }
    else {
        return(ANOMOLOUS_A2);
    }
}

static BOOLEAN is_error(ActionList *action)
{
    switch (action->act) {
        case NONE: {
            return(TRUE);
        }
        case PICK_UP: {
            return((action->subtask != ADD_MILK) && (action->subtask != ADD_SUGAR) && (action->subtask != ADD_COFFEE));
        }
        case PUT_DOWN: {
            return((action->subtask != ADD_MILK) && (action->subtask != ADD_SUGAR) && (action->subtask != ADD_COFFEE));
        }
        case DIP_SPOON: {
            return((action->subtask != ADD_MILK) && (action->subtask != ADD_SUGAR) && (action->subtask != ADD_COFFEE));
        }
        case EMPTY_SPOON: {
            return((strcmp(action->arg2, "coffee mug") != 0) || ((action->subtask != ADD_MILK) && (action->subtask != ADD_SUGAR) && (action->subtask != ADD_COFFEE)));
        }
        case EAT_FROM_SPOON: {
            return(TRUE);
        }
        case STIR: {
            return(TRUE);
        }
        case POUR: {
            return((strcmp(action->arg2, "coffee mug") != 0) || ((action->subtask != ADD_MILK) && (action->subtask != ADD_SUGAR) && (action->subtask != ADD_COFFEE)));
        }
        case TEAR: {
            return((action->subtask != ADD_MILK) && (action->subtask != ADD_SUGAR) && (action->subtask != ADD_COFFEE));
        }
        case SCREW_OPEN: {
            return(TRUE);
        }
        case SCREW_CLOSED: {
            return(TRUE);
        }
        case SWAP: {
            return(TRUE);
        }
        case DRINK: {
            return(TRUE);
        }
        case PREPARE: {
            return(TRUE);
        }
        case PUSH: {
            return(TRUE);
        }
        case RELAX: {
            return(TRUE);
        }
        default: {
            return(TRUE);
        }
    }
}

static A2Label get_subtask(ActionList *action)
{
    switch (action->act) {
        case NONE: {
            return(A2_BREAK);
        }
        case PICK_UP: {
            if (is_implement(action->arg1) && (action->next != NULL)) {
                if ((action->next->act == PUT_DOWN) && is_implement(action->next->arg1)) {
                    return(A2_BREAK);
                }
                else {
                    return(get_subtask(action->next));
                }
            }
            else {
                return(corresponding_subtask(action));
            }
        }
        case PUT_DOWN: {
            if (is_implement(action->arg1) && (action->prev != NULL)) {
                if ((action->prev->act == PUT_DOWN) && is_implement(action->prev->arg1)) {
                    return(A2_BREAK); 
                }
                else {
                    return(get_subtask(action->prev));
                }
            }
            else {
                return(corresponding_subtask(action));
            }
        }
        case DIP_SPOON: {
            return(corresponding_subtask(action));
        }
        case EMPTY_SPOON: {
            return(corresponding_subtask(action));
        }
        case EAT_FROM_SPOON: {
            return(A2_BREAK);
        }
        case STIR: {
            return(A2_BREAK);
        }
        case POUR: {
            return(corresponding_subtask(action));
        }
        case TEAR: {
            return(corresponding_subtask(action));
        }
        case SCREW_OPEN: {
            return(corresponding_subtask(action));
        }
        case SCREW_CLOSED: {
            return(corresponding_subtask(action));
        }
        case SWAP: {
            return(A2_BREAK);
        }
        case DRINK: {
            return(A2_BREAK);
        }
        case PREPARE: {
            return(A2_BREAK);
        }
        case PUSH: {
            return(A2_BREAK);
        }
        case RELAX: {
            return(A2_BREAK);
        }
        default: {
            return(A2_BREAK);
        }
    }
}

BOOLEAN is_crux(ActionType act)
{
    switch (act) {
        case EMPTY_SPOON: case EAT_FROM_SPOON: case STIR: case POUR: case DRINK: {
            return(TRUE);
        }
        default: {
            return(FALSE);
        }
    }
}

static int count_actions()
{
    ActionList *tmp;
    int count = 0;

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        count++;
    }
    return(count);
}

static int count_independents()
{
    ActionList *tmp;
    int count = 0;

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        if (tmp->independent) {
            count++;
        }
    }
    return(count);
}

static int count_crux_errors()
{
    ActionList *tmp;
    int count = 0;

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        if ((tmp->crux) && (!tmp->correct)) {
            count++;
        }
    }
    return(count);
}

static int count_non_crux_errors()
{
    ActionList *tmp;
    int count = 0;

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        if ((!tmp->crux) && (!tmp->correct)) {
            count++;
        }
    }
    return(count);
}

/*----------------------------------------------------------------------------*/

static void categorise_a1s()
{
    ActionList *tmp;
    BOOLEAN prev_break;

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        tmp->a2_open = FALSE;
        tmp->a2_close = FALSE;
        tmp->subtask = get_subtask(tmp);
        tmp->crux = is_crux(tmp->act);
        tmp->independent = FALSE;
        tmp->correct = TRUE;
    }
    for (tmp = first_action, prev_break = FALSE; tmp != NULL; tmp = tmp->next) {
        if ((prev_break) && (tmp->subtask == A2_BREAK)) {
            tmp->prev->subtask = A2_DOUBLE_BREAK;
            tmp->subtask = A2_DOUBLE_BREAK;
        }
        else {
            prev_break = (tmp->subtask == A2_BREAK);
        }
    }
}

static void extend_bracketing_minimally(ActionList *action)
{
    ActionList *tmp;

    for (tmp = action->prev; ((tmp != NULL) && (tmp->subtask == action->subtask)); tmp = tmp->prev) {
        tmp->next->a2_open = FALSE;
        tmp->a2_open = TRUE;
    }

    for (tmp = action->next; ((tmp != NULL) && (tmp->subtask == action->subtask)); tmp = tmp->next) {
        tmp->prev->a2_close = FALSE;
        tmp->a2_close = TRUE;
    }
}

static void extend_bracketing_to_prior(ActionList *action)
{
    /* Look to the left. Is there another action from the current subtask */
    /* before the beginning of the previous fully bracketed subtask?      */

    ActionList *tmp = action;
    BOOLEAN crosses_a2 = FALSE;

    while ((tmp != NULL) && (tmp->subtask == action->subtask)) {
        tmp = tmp->prev;
    }

    while ((tmp != NULL) && (tmp->subtask != action->subtask)) {

        /* DON'T BRACKET OVER A DOUBLE A2_BREAK */

        if (tmp->subtask == A2_DOUBLE_BREAK) {
            return;
        }

        /* BE SURE WE DON'T CROSS ANOTHER COMPLETE A2 */
        /* IF WE DO, THEN TERMINATE THE BRACKETING    */ 

        if (!(crosses_a2) && (tmp->a2_close)) {
            crosses_a2 = TRUE;
        }
        else if ((crosses_a2) && (tmp->a2_open)) {
            return;
        }

        tmp = tmp->prev;
    }

    while ((tmp != NULL) && (tmp->subtask == action->subtask)) {
        tmp = tmp->prev;
    }

    if ((tmp == NULL) && (first_action->subtask == action->subtask)) {

    /* The open bracket should be moved to the start of the sequence */

        action->a2_open = FALSE;
        first_action->a2_open = TRUE;
    }
    else if (tmp != NULL) {

    /* We've found a prior action in the same subtask,  */
    /* so shift the open bracket                        */

        action->a2_open = FALSE;
        tmp->next->a2_open = TRUE;
        extend_bracketing_to_prior(tmp);
    }
}

static void extend_bracketing_to_post(ActionList *action)
{
    /* Look to the right. Is there another action from the current subtask */
    /* before the end of the following fully bracketed subtask?            */

    ActionList *tmp = action;
    BOOLEAN crosses_a2 = FALSE;

    while ((tmp != NULL) && (tmp->subtask == action->subtask)) {
        tmp = tmp->next;
    }

    while ((tmp != NULL) && (tmp->subtask != action->subtask)) {

        /* DON'T BRACKET OVER A DOUBLE A2_BREAK */

        if (tmp->subtask == A2_DOUBLE_BREAK) {
            return;
        }

        /* BE SURE WE DON'T CROSS ANOTHER COMPLETE A2 */
        /* IF WE DO, THEN TERMINATE THE BRACKETING    */ 

        if (!(crosses_a2) && (tmp->a2_open)) {
            crosses_a2 = TRUE;
        }
        else if ((crosses_a2) && (tmp->a2_close)) {
            return;
        }

        tmp = tmp->next;
    }

    while ((tmp != NULL) && (tmp->subtask == action->subtask)) {
        tmp = tmp->next;
    }

    if ((tmp == NULL) && (last_action->subtask == action->subtask)) {

    /* The close bracket should be moved to the end of the sequence */

        action->a2_close = FALSE;
        last_action->a2_close = TRUE;
    }
    else if (tmp != NULL) {

    /* We've found a post action in the same subtask,   */
    /* so shift the close bracket                       */

        action->a2_close = FALSE;
        tmp->prev->a2_close = TRUE;
        extend_bracketing_to_post(tmp);
    }
}

static void bracket_a1s()
{
    ActionList *tmp;

    /* First run through the list finding cruxes. Extend them to left/right   */
    /* so that the brackets extend over connected regions of one subtask.     */

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        if ((tmp->crux) && (tmp->subtask != A2_BREAK)) {
            tmp->a2_open = TRUE;
            tmp->a2_close = TRUE;
            extend_bracketing_minimally(tmp);
        }
    }

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        if (tmp->a2_open) {
            extend_bracketing_to_prior(tmp);
        }
        if (tmp->a2_close) {
            extend_bracketing_to_post(tmp);
        }
    }
}

static void categorise_indendepents()
{
    ActionList *tmp;
    int level = 0;

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        if (tmp->a2_open) {
            level++;
        }
        tmp->independent = (level == 0);
        if (tmp->a2_close) {
            level--;
        }
    }
}

static void categorise_errors()
{
    ActionList *tmp;

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        tmp->correct = !is_error(tmp);
    }
}

static void analyse_via_acs1()
{
    if (first_action != NULL) {
        categorise_a1s();
        bracket_a1s();
        categorise_indendepents();
        categorise_errors();
    }
}

/******************************************************************************/
/* Utility functions for repairing action lists:                              */

static void insert_before(ActionList *action, A2Label subtask, ActionType act, char *arg1, char *arg2)
{
    ActionList *new;

    if ((new = (ActionList *)malloc(sizeof(ActionList))) == NULL) {
        fprintf(stderr, "WARNING: Memory allocation failed in insert_before()\n");
    }
    else if (action == NULL) {
        new->act = act;
        new->cycle = 1;
        new->arg1 = arg1;
        new->arg2 = arg2;
        new->crux = is_crux(new->act);
        new->subtask = subtask;
        new->next = NULL;
        new->prev = NULL;

        first_action = new;
        last_action = new;
    }
    else {
        new->act = act;
        new->cycle = action->cycle - 1;
        new->arg1 = arg1;
        new->arg2 = arg2;
        new->crux = is_crux(new->act);
        new->subtask = subtask;
        new->next = action;
        new->prev = action->prev;

        action->prev = new;
        if (new->prev != NULL) {
            new->prev->next = new;
        }
        else {
            first_action = new;
        }
    }    
}

static void insert_after(ActionList *action, A2Label subtask, ActionType act, char *arg1, char *arg2)
{
    ActionList *new;

    if ((new = (ActionList *)malloc(sizeof(ActionList))) == NULL) {
        fprintf(stderr, "WARNING: Memory allocation failed in insert_after()\n");
    }
    else if (action == NULL) {
        new->act = act;
        new->cycle = 1;
        new->arg1 = arg1;
        new->arg2 = arg2;
        new->crux = is_crux(new->act);
        new->subtask = subtask;
        new->next = NULL;
        new->prev = NULL;

        first_action = new;
        last_action = new;
    }
    else {
        new->act = act;
        new->cycle = action->cycle + 1;
        new->arg1 = arg1;
        new->arg2 = arg2;
        new->crux = is_crux(new->act);
        new->subtask = subtask;
        new->next = action->next;
        new->prev = action;

        action->next = new;
        if (new->next != NULL) {
            new->next->prev = new;
        }
        else {
            last_action = new;
        }
    }
}

static void delete_current(ActionList *act)
{
    ActionList *tmp = act;

    if (act->prev != NULL) {
        act->prev->next = tmp->next;
    }
    else {
        first_action = tmp->next;
    }

    if (act->next != NULL) {
        act->next->prev = tmp->prev;
    }
    else {
        last_action = tmp->prev;
    }

    free(tmp);
}

static BOOLEAN arg_is_ok(char *arg_a, char *arg_b)
{
    if ((arg_a != NULL) && (arg_b != NULL)) {
        return(strcmp(arg_a, arg_b) == 0);
    }
    else {
        return(arg_a == arg_b);
    }
}

static BOOLEAN match_action(ActionList *action, A2Label subtask, ActionType act, char *arg1, char *arg2)
{
    if ((action != NULL) && (action->subtask == subtask) && (action->act == act)) {
        return(arg_is_ok(arg1, action->arg1) && arg_is_ok(arg2, action->arg2));
    }
    else {
        return(FALSE);
    }
}

static char *substance_for_task(A2Label subtask)
{
    switch (subtask) {
        case ADD_COFFEE: {
            return("coffee");
        }
        case ADD_SUGAR: {
            return("sugar");
        }
        case ADD_MILK: {
            return("milk");
        }
        default: {
            return(NULL);
        }
    }
}

static char *packet_for_task(A2Label subtask)
{
    switch (subtask) {
        case ADD_COFFEE: {
            return("coffee packet");
        }
        case ADD_SUGAR: {
            return("sugar packet");
        }
        case ADD_MILK: {
            return("milk carton");
        }
        default: {
            return(NULL);
        }
    }
}

static char *source_for_task(A2Label subtask)
{
    switch (subtask) {
        case ADD_COFFEE: {
            return("coffee jar");
        }
        case ADD_SUGAR: {
            return("sugar bowl");
        }
        default: {
            return(NULL);
        }
    }
}

/*----------------------------------------------------------------------------*/
/* Count/Correct object substitions:                                          */

static BOOLEAN check_set_arg(char **arg, char *correct)
{
    if (strcmp(*arg, correct) == 0) {
        return(FALSE);
    }
    else {
        *arg = correct;
        return(TRUE);
    }
}

static BOOLEAN ensure_appropriate_dipable_container(ActionList *action)
{
    switch (action->subtask) {
        case ADD_SUGAR: {
            return(check_set_arg(&(action->arg1), "sugar bowl"));
        }
        case ADD_COFFEE: {
            return(check_set_arg(&(action->arg1), "coffee jar"));
        }
        default: {
            return(FALSE);
        }
    }
}

static BOOLEAN ensure_appropriate_target(ActionList *action)
{
    return(check_set_arg(&(action->arg2), "coffee mug"));
}

static BOOLEAN ensure_appropriate_substance(ActionList *action)
{
    switch (action->subtask) {
        case ADD_MILK: {
            return(check_set_arg(&(action->arg1), "milk"));
        }
        case ADD_SUGAR: {
            return(check_set_arg(&(action->arg1), "sugar"));
        }
        case ADD_COFFEE: {
            return(check_set_arg(&(action->arg1), "coffee"));
        }
        default: {
            return(FALSE);
        }
    }
}

static BOOLEAN ensure_appropriate_packet(ActionList *action)
{
    switch (action->subtask) {
        case ADD_MILK: {
            return(check_set_arg(&(action->arg1), "milk carton"));
        }
        case ADD_SUGAR: {
            return(check_set_arg(&(action->arg1), "sugar packet"));
        }
        case ADD_COFFEE: {
            return(check_set_arg(&(action->arg1), "coffee packet"));
        }
        default: {
            return(FALSE);
        }
    }
}

static BOOLEAN incorrect_object(A2Label subtask, char *obj)
{
    switch(subtask) {
        case ADD_MILK: {
            return((strcmp(obj, "milk carton") != 0));
        }
        case ADD_COFFEE: {
            return((strcmp(obj, "coffee packet") != 0) && (strcmp(obj, "coffee jar") != 0) && (strcmp(obj, "teaspoon") != 0));
        }
        case ADD_SUGAR: {
            return((strcmp(obj, "sugar packet") != 0) && (strcmp(obj, "sugar bowl") != 0) && (strcmp(obj, "teaspoon") != 0));
        }
        default: {
            return(FALSE);
        }
    }
}

static char *get_appropriate_object(ActionList *action)
{
    switch(action->subtask) {
        case ADD_MILK: {
            return("milk carton");
        }
        case ADD_COFFEE: {
            if ((action->next != NULL) && (action->next->act == DIP_SPOON)) {
                return("teaspoon");
            }
            else {
                return("coffee packet");
	    }
        }
        case ADD_SUGAR: {
            if ((action->next != NULL) && (action->next->act == DIP_SPOON)) {
                return("teaspoon");
            }
            else {
                return("sugar packet");
	    }
        }
        default: {
            return(FALSE);
        }
    }
}

static void adjust_subsequent_actions(ActionList *action, char *bad_arg, char *good_arg)
{
    /* Correct all actions using bad_arg until we put it down. */

    while (action != NULL) {
        if ((action->arg1 != NULL) && (strcmp(action->arg1, bad_arg) == 0)) {
            action->arg1 = good_arg;
        }
        if ((action->arg2 != NULL) && (strcmp(action->arg2, bad_arg) == 0)) {
            action->arg2 = good_arg;
        }
        if ((action->act == PUT_DOWN) && (action->arg2 == good_arg)) {
            action = NULL; /* We've put the object down again -- force return from function */
        }
        else {
            action = action->next;
        }
    }
}

static BOOLEAN repair_argument(ActionList *action)
{
    switch (action->act) {
        case PICK_UP: {

            /* If we have picked up the wrong thing, correct all actions
            using that thing until we put it down again. Count just one
            argument error */

            if (incorrect_object(action->subtask, action->arg1)) {
                char *bad_arg = action->arg1;
                char *good_arg = get_appropriate_object(action);
                adjust_subsequent_actions(action, bad_arg, good_arg);
                return(TRUE);
            }
            else {
                return(FALSE);
            }
        }
        case PUT_DOWN: {
            return(FALSE);
        }
        case DIP_SPOON: {
            return(ensure_appropriate_dipable_container(action));
        }
        case EMPTY_SPOON:
        case POUR: {
            BOOLEAN tmp1, tmp2;

            tmp1 = ensure_appropriate_substance(action);
            tmp2 = ensure_appropriate_target(action);
            return(tmp1 || tmp2);
        }
        case TEAR: {
            return(ensure_appropriate_packet(action));
            break;
        }
        default: {
            return(FALSE);
        }
    }
}

static int repair_object_substitions()
{
    int os = 0;
    ActionList *tmp;

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        if (repair_argument(tmp)) {
            os++;
        }
    }

    return(os);
}

/*----------------------------------------------------------------------------*/
/* Count/Correct action additions:                                            */

static BOOLEAN is_addition(ActionList *action)
{
    switch (action->subtask) {
        case ADD_COFFEE:
        case ADD_SUGAR:
        case ADD_MILK: {
            return(FALSE);
        }
        default: {
            return(TRUE);
	}
    }
}

static int remove_additions()
{
    int ad = 0;
    ActionList *tmp;

    for (tmp = first_action; tmp != NULL; tmp = tmp->next) {
        if (is_addition(tmp)) {
            delete_current(tmp);
            ad++;
        }
    }

    return(ad);
}

/*----------------------------------------------------------------------------*/
/* Count/Correct action omissions:                                            */

static ActionList *find_crux(A2Label subtask)
{
    ActionList *tmp = first_action;

    while ((tmp != NULL) && !((tmp->subtask == subtask) && (tmp->crux))) {
        tmp = tmp->next;
    }
    return(tmp);
}

static ActionList *find_subtask(A2Label subtask)
{
    ActionList *tmp = first_action;

    while ((tmp != NULL) && (tmp->subtask != subtask)) {
        tmp = tmp->next;
    }
    return(tmp);
}

static void insert_subtask(A2Label subtask)
{
    switch (subtask) {
        case ADD_MILK: {
            insert_after(last_action, subtask, PICK_UP, "milk carton", NULL);
            insert_after(last_action, subtask, TEAR, "milk carton", NULL);
            insert_after(last_action, subtask, POUR, "milk", "coffee mug");
            insert_after(last_action, subtask, PUT_DOWN, "milk carton", NULL);
            break;
        }
        case ADD_COFFEE: {
            insert_after(last_action, subtask, PICK_UP, "coffee packet", NULL);
            insert_after(last_action, subtask, TEAR, "coffee packet", NULL);
            insert_after(last_action, subtask, POUR, "coffee", "coffee mug");
            insert_after(last_action, subtask, PUT_DOWN, "coffee packet", NULL);
            break;
        }
        case ADD_SUGAR: {
            insert_after(last_action, subtask, PICK_UP, "sugar packet", NULL);
            insert_after(last_action, subtask, TEAR, "sugar packet", NULL);
            insert_after(last_action, subtask, POUR, "sugar", "coffee mug");
            insert_after(last_action, subtask, PUT_DOWN, "sugar packet", NULL);
            break;
        }
        default: {
            break;
        }
    }
}

static BOOLEAN appropriate_object(A2Label subtask, char *object)
{
    switch (subtask) {
        case ADD_COFFEE: {
            return((strcmp(object, "teaspoon") == 0) ||
                   (strcmp(object, "coffee packet") == 0) ||
                   (strcmp(object, "coffee jar") == 0));
        }
        case ADD_SUGAR: {
            return((strcmp(object, "teaspoon") == 0) ||
                   (strcmp(object, "sugar packet") == 0) ||
                   (strcmp(object, "sugar bowl") == 0));
        }
        case ADD_MILK: {
            return((strcmp(object, "milk carton") == 0));
        }
        default: {
            return(TRUE);
        }
    }
}

static char *in_hand(ActionList *action)
{
    while (action != NULL) {
        if (action->act == PICK_UP) {
            return(action->arg1);
        }
        else {
            action = action->prev;
        }
    }
    return(NULL);
}

static BOOLEAN ensure_pickup_present(ActionList *crux)
{
    ActionList *tmp = crux;

    if (crux->act == EMPTY_SPOON) {
        while ((tmp->prev != NULL) && (tmp->prev->subtask == crux->subtask) && !match_action(tmp, crux->subtask, PICK_UP, "teaspoon", NULL)) {
            tmp = tmp->prev;
        }
        if (!match_action(tmp, crux->subtask, PICK_UP, "teaspoon", NULL)) {
            insert_before(tmp, crux->subtask, PICK_UP, "teaspoon", NULL);
            return(TRUE);
        }
        else {
            return(FALSE);
        }
    }
    else {
        while ((tmp->prev != NULL) && (tmp->prev->subtask == crux->subtask) && (tmp->act != PICK_UP)) {
            tmp = tmp->prev;
        }
        if ((tmp->act == PICK_UP) && appropriate_object(crux->subtask, tmp->arg1)) {
            return(FALSE);
        }
        else {
            insert_before(tmp, crux->subtask, PICK_UP, packet_for_task(crux->subtask), NULL);
            return(TRUE);
        }
    }
}

static BOOLEAN ensure_make_available_present(ActionList *crux)
{
    ActionType act;
    char *arg;

/*  What we need to do depends on the crux action. */

    if (crux->act == EMPTY_SPOON) {
        act = DIP_SPOON;
        arg = source_for_task(crux->subtask);
    }
    else {
        act = TEAR;
        arg = packet_for_task(crux->subtask);
    }

    if ((crux->act == POUR) && (in_hand(crux) != NULL) && (source_for_task(crux->subtask) != NULL) && (strcmp(in_hand(crux), source_for_task(crux->subtask)) == 0)) {
        return(FALSE);
    }
    else if (arg == NULL) {
        return(TRUE);
    }    
    else if ((crux->prev == NULL) || !match_action(crux->prev, crux->subtask, act, arg, NULL)) {
        insert_before(crux, crux->subtask, act, arg, NULL);
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

static ActionList *insert_crux(ActionList *start, A2Label subtask)
{
    ActionList *tmp = start, *crux;
    ActionType crux_action;
    char *held;

    /* The crux is always the penultimate action. So find the last action   */
    /* for the subtask.                                                     */

    while ((tmp->next != NULL) && (tmp->next->subtask == subtask)) {
        tmp = tmp->next;
    }

    held = in_hand(tmp);
    if (strcmp(held, "teaspoon") == 0) {
        crux_action = EMPTY_SPOON;
    }
    else {
        crux_action = POUR;
    }

    if (tmp->act == PUT_DOWN) {
        insert_before(tmp, subtask, crux_action, substance_for_task(subtask), "coffee mug");
        crux = tmp->prev;
    }
    else {
        insert_after(tmp, subtask, crux_action, substance_for_task(subtask), "coffee mug");
        crux = tmp->next;
    }
    return(crux);
}

static BOOLEAN ensure_discard_present(ActionList *crux)
{
    char *object = in_hand(crux);

    if ((crux->next == NULL) || (crux->next->act != PUT_DOWN) || (strcmp(crux->next->arg1, object) != 0)) {
        insert_after(crux, crux->subtask, PUT_DOWN, object, NULL);
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

static void check_presence(A2Label subtask, int *om1, int *om2)
{
    ActionList *crux, *tmp;

    if ((crux = find_crux(subtask)) != NULL) {
        /* We've found the crux. Check everything else is in order */
        if (ensure_pickup_present(crux)) { (*om2)++; }
        if (ensure_make_available_present(crux)) { (*om2)++; }
        if (ensure_discard_present(crux)) { (*om2)++; }
    }
    else if ((tmp = find_subtask(subtask)) != NULL) {
        /* No crux, but some subtask actions. Insert crux and check everything else is in order */
        crux = insert_crux(tmp, subtask); (*om2)++;
        if (ensure_pickup_present(crux)) { (*om2)++; }
        if (ensure_make_available_present(crux)) { (*om2)++; }
        if (ensure_discard_present(crux)) { (*om2)++; }
    }
    else {
        /* No crux, and no subtask actions. Insert complete subtask */
        insert_subtask(subtask);
        (*om1)++;
    }
}

static void insert_omissions(int *om1, int *om2)
{
    *om1 = 0; *om2 = 0;

    check_presence(ADD_COFFEE, om1, om2);
    check_presence(ADD_SUGAR, om1, om2);
    check_presence(ADD_MILK, om1, om2);
}

/*----------------------------------------------------------------------------*/
/* Count/Correct sequence errors:                                             */

static int check_post_crux_sequence(ActionList *crux, char *held)
{
    int cs = 0;

    if ((crux->next == NULL) || (crux->next->subtask != crux->subtask)) {
        insert_after(crux, crux->subtask, PUT_DOWN, held, NULL);
        cs++;
    }
    else {
        while ((crux->next != NULL) && (crux->next->subtask == crux->subtask) && !match_action(crux->next, crux->subtask, PUT_DOWN, held, NULL)) {
            /* Remove continuous perseveration before the final "put down" */
            delete_current(crux->next);
            cs++;
        }
        if (!match_action(crux->next, crux->subtask, PUT_DOWN, held, NULL)) {
            /* Insert the final "put down" if we didn't find it */
            insert_after(crux, crux->subtask, PUT_DOWN, held, NULL);
            cs++;
        }
    }

    /* Check/delete continuous perseveration after the final "put down" */

    while ((crux->next->next != NULL) && (crux->next->next->subtask == crux->subtask)) {
        delete_current(crux->next->next);
        cs++;
    }
    return(cs);
}

static ActionList *find_appropriate_action(ActionList *crux, ActionType act, char *arg1, char *arg2, int *cs)
{

/* Look for act(arg1, arg2) prior to the crux. If not found, insert it and  */
/* increment the error count. Return a pointer to the action.               */

    ActionList *tmp = crux;

    while ((tmp->prev != NULL) && (tmp->prev->subtask == crux->subtask) && !match_action(tmp, crux->subtask, act, arg1, arg2)) {
        tmp = tmp->prev;
    }

    if (!match_action(tmp, crux->subtask, act, arg1, arg2)) {
        insert_before(tmp, crux->subtask, act, arg1, arg2);
        (*cs)++;
        return(tmp->prev);
    }
    else {
        return(tmp);
    }
}

static int check_crux_sequence(ActionList *crux)
{
    int cs = 0;
    char *held = NULL;
    ActionList *tmp, *a1 = NULL, *a2 = NULL;

    /* Make sure the actions before the crux are consistent/correct.    */
    /* Also set held to the contents of the hand.                       */

    if (crux->act == EMPTY_SPOON) {
        held = "teaspoon";

/* Look for dip_spoon(source) prior to the crux. Assign a2 to this. If  */
/* it doesn't exist, insert it as the first act in this subtask.        */

        a2 = find_appropriate_action(crux, DIP_SPOON, source_for_task(crux->subtask), NULL, &cs);

/* Look for pick_up(held) prior to the crux. Assign a1 to this. If it   */
/* doesn't exist, insert it as the first act in this subtask.           */

        a1 = find_appropriate_action(crux, PICK_UP, held, NULL, &cs);

    }
    else if (crux->act == POUR) {

        char *obj1 = source_for_task(crux->subtask);
        char *obj2 = in_hand(crux);

        if ((obj1 != NULL) && (obj2 != NULL) && (strcmp(obj1, obj2) == 0)) {

/* If we're holding a jar/bowl, we don't need an enabling/opening act:  */

            held = obj2;
            a2 = NULL;
        }
        else {

/* Look for tear(held) prior to the crux. Assign a2 to this. If it      */
/* doesn't exist, insert it as the first act in this subtask.           */

            held = packet_for_task(crux->subtask);
            a2 = find_appropriate_action(crux, TEAR, held, NULL, &cs);
        }

/* Look for pick_up(held) prior to the crux. Assign a1 to this. If it   */
/* doesn't exist, insert it as the first act in this subtask.           */

        a1 = find_appropriate_action(crux, PICK_UP, held, NULL, &cs);
    }
    else {
        fprintf(stderr, "WARNING: Anomolous crux: \n");
        print_action(stderr, crux);
    }

/* Delete anything in the subtask prior to the crux that isn't a1 or a2.      */

    for (tmp = crux->prev; ((tmp != NULL) && (tmp->subtask == crux->subtask)); tmp = tmp->prev) {
        if ((tmp != a1) && (tmp != a2)) {
            delete_current(tmp);
            cs++;
        }
    }

/* Ensure a1 comes before a2.                                                */

    if ((a1 != NULL) && (a2 != NULL) && (a1->next != a2)) {
        a1->prev = a2->prev;
        a1->next = a2;
        a2->prev = a1;
        a2->next = crux;
        if (first_action == a2) {
            first_action = a1;
        }
        cs++;
    }

    /* Now check that the following action is the appropriate put down  */
    /* If not, carry out the appropriate repair.                        */

    cs += check_post_crux_sequence(crux, held);

    return(cs);
}

static int remove_recurrent_perseverations(ActionList *current)
{
    A2Label st = current->subtask;
    int p = 0;

    /* Skip to end of the subtask */

    while ((current != NULL) && (current->subtask == st)) {
        current = current->next;
    }

    /* Now run through any remaining actions, removing any for st. */

    while (current != NULL) {
        if (current->subtask != st) {
            /* Skip other subtasks */
            current = current->next;
        }
        else {
            /* We've got recurrent preseverations: */
            ActionList *tmp = current->next;
            delete_current(current);
            current = tmp;
            p++;
        }
    }

    return(p);
}

static int check_sequence(A2Label subtask)
{
    ActionList *tmp = first_action;
    int sq = 0;

    /* Find the (first) crux of the given subtask: */

    while ((tmp != NULL) && ((tmp->subtask != subtask) || (!tmp->crux))) {
        tmp = tmp->next;
    }

    if (tmp != NULL) {       /* We've found the crux */
        /* Check actions immediately before/after the crux */
        sq += check_crux_sequence(tmp);

        /* Remove any extraneous actions relating to the specified subtask */
        sq += remove_recurrent_perseverations(tmp);
    }
    return(sq);
}

static int repair_sequence()
{
    int sq = 0;

    sq += check_sequence(ADD_COFFEE);
    sq += check_sequence(ADD_SUGAR);
    sq += check_sequence(ADD_MILK);
    return(sq);
}

/*----------------------------------------------------------------------------*/

static void analyse_via_acs2(TaskType task, int *os, int *ad, int *om1, int *om2, int *sq)
{
    if (first_action != NULL) {
        switch (task) {
            case COFFEE_PREPARATION: {
                *os = repair_object_substitions();
                *ad = remove_additions();
                insert_omissions(om1, om2);
                *sq = repair_sequence();
                break;
            }
            default: {
                fprintf(stderr, "No ACS2 analysis available\n");
            }
        }
    }
}

/******************************************************************************/

static void print_internal_errors(FILE *fp)
{
    fprintf(fp, "Internal errors:\n");
    fprintf(fp, "  Failed argument selections: %3d\n", ARG_SEL_FAIL_COUNT);
    fprintf(fp, "  Failed actions:             %3d\n", ACTION_FAIL_COUNT);
}

static void print_analysis1(FILE *fp)
{
    if (first_action != NULL) {
        fprintf(fp, "Action analysis:\n");
        fprintf(fp, "  First action:           %4d\n", first_action->cycle);
        fprintf(fp, "  Last action:            %4d\n", last_action->cycle);
        fprintf(fp, "  Number of actions:      %4d\n", count_actions());
        fprintf(fp, "  Actions per 100 cycles:   %6.4f\n",  100.0 * count_actions() / (double) (last_action->cycle - first_action->cycle));
    }
    else {
        fprintf(fp, "No actions attempted\n");
    }
}

static void print_analysis2(FILE *fp)
{
    if (first_action != NULL) {
        fprintf(fp, "ACS 1 Error Analysis:\n");
        fprintf(fp, "  Independents:           %4d\n", count_independents());
        fprintf(fp, "  Crux Errors:            %4d\n", count_crux_errors());
        fprintf(fp, "  Non-crux Errors:        %4d\n", count_non_crux_errors());
    }
}

static void print_analysis3(FILE *fp, int os, int ad, int om1, int om2, int sq)
{
    if (first_action != NULL) {
        fprintf(fp, "ACS 2 Error Analysis:\n");
        fprintf(fp, "  Object Substitutions:   %4d\n", os);
        fprintf(fp, "  Additions:              %4d\n", ad);
        fprintf(fp, "  Subtask Omissions:      %4d\n", om1);
        fprintf(fp, "  Step Omissions:         %4d\n", om2);
        fprintf(fp, "  Sequence:               %4d\n", sq);
    }
}

static void delete_remainder(ActionList *list)
{
    ActionList *tmp;

    while (list != NULL) {
        tmp = list->next;
        delete_current(list);
        list = tmp;
    }
}

static BOOLEAN terminate_task()
{
    /* Return TRUE if 1) there are some actions, and 2) there's a NONE action */
    /* (i.e., the task completed normally). If one is found, delete it and    */
    /* all subsequent actions.                                                */

    if ((first_action == NULL) || (last_action == NULL)) {
        return(FALSE);
    }
    else {
        ActionList *tmp = first_action;

        while (tmp != NULL) {
            if (tmp->act == NONE) {
                delete_remainder(tmp);
                return(first_action != NULL);
            }
            else {
                tmp = tmp->next;
            }
        }
        return(FALSE);
    }
}

void ActAnalyseList(FILE *fp, TaskType task)
{
    int os = 0, ad = 0, om1 = 0, om2 = 0, sq = 0;

    terminate_task();
    fprintf(fp, "\n================================================================================\n\n");
    print_internal_errors(fp);
    analyse_via_acs1();
    print_list(fp);
    print_analysis1(fp);
    print_analysis2(fp);
    analyse_via_acs2(task, &os, &ad, &om1, &om2, &sq);
    print_analysis3(fp, os, ad, om1, om2, sq);
    fprintf(fp, "\n================================================================================\n");
}

/******************************************************************************/
/********* For multiple runs: Produce average stats: **************************/

static int first_count;
static int last_count;;
static int action_count;
static int independents;
static int crux_errors;
static int non_crux_errors;
static int arg_sel_fail_count;
static int action_fail_count;
static int runs;

static int step_omissions;
static int subtask_omissions;
static int additions;
static int substitutions;
static int sequence;

void ActMultiReset()
{
    first_count =        0;
    last_count =         0;
    action_count =       0;
    independents =       0;
    crux_errors  =       0;
    non_crux_errors =    0;
    arg_sel_fail_count = 0;
    action_fail_count  = 0;
    runs =               0;

    step_omissions =     0;
    subtask_omissions =  0;
    additions =          0;
    substitutions =      0;
    sequence =           0;
}

void ActMultiAnalyseList(TaskType task)
{
    int os, ad, om1, om2, sq;

    print_list(stderr);

    if (terminate_task()) {

        analyse_via_acs1();

        first_count +=  first_action->cycle;
        last_count +=   last_action->cycle;
        action_count += count_actions();
        independents +=    count_independents();
        crux_errors  +=    count_crux_errors();
        non_crux_errors += count_non_crux_errors();
    
        arg_sel_fail_count += ARG_SEL_FAIL_COUNT;
        action_fail_count  += ACTION_FAIL_COUNT;

        print_list(stderr);

        analyse_via_acs2(task, &os, &ad, &om1, &om2, &sq);

        subtask_omissions += om1;
        step_omissions += om2;
        additions += ad;
        substitutions += os;
        sequence += sq;

        runs++;
    }
}

static void print_mean_statistic(char *prefix, char *suffix, double par, double val)
{
    char file[32];
    FILE *fp;

    sprintf(file, "%s.%s", prefix, suffix);
    if ((fp = fopen(file, "a")) != NULL) {
        fprintf(fp, "%6.4f %6.4f\n", par, val);
        fclose(fp);
    }
}

void ActMultiPrint(char *prefix, double par)
{
    if (runs != 0) {
        print_mean_statistic(prefix, "init", par, (double) first_count / (double) runs);
        print_mean_statistic(prefix, "total", par, (double) last_count / (double) runs);
        print_mean_statistic(prefix, "acts", par, (double) action_count / (double) runs);
        print_mean_statistic(prefix, "rate", par, (double) 100.0 * action_count / (double) (last_count - first_count));

        print_mean_statistic(prefix, "ind", par, (double) independents / (double) runs);
        print_mean_statistic(prefix, "prop_ind", par, (double) independents / (double) action_count);
        print_mean_statistic(prefix, "crux", par, (double) crux_errors / (double) runs);
        print_mean_statistic(prefix, "non_crux", par, (double) non_crux_errors / (double) runs);
        print_mean_statistic(prefix, "err", par, (double) (crux_errors + non_crux_errors) / (double) runs);

        print_mean_statistic(prefix, "os", par, (double) (substitutions) / (double) runs);
        print_mean_statistic(prefix, "ad", par, (double) (additions) / (double) runs);
        print_mean_statistic(prefix, "om.step", par, (double) (step_omissions) / (double) runs);
        print_mean_statistic(prefix, "om.subtask", par, (double) (subtask_omissions) / (double) runs);
        print_mean_statistic(prefix, "sq", par, (double) (sequence) / (double) runs);
    }
}

/******************************************************************************/
