/******************************************************************************/

#include "cs.h"
#include <math.h>

extern double apply_equations(double previous, double input);
extern double squash(double x);

#undef SELF_INHIBIT

/******************************************************************************/
/*  SCHEMA ACTIVATION FLOW                                                    */
/******************************************************************************/

static int length(Subgoal *l)
{
    int count = 0;

    while (l != NULL) {
        count++;
        l = l->next;
    }
    return(count);
}

double SchemaIntentionalInfluence(int cycle, CsSchema *s)
{
    extern BOOLEAN ordering_constraints_satisfied(CsSchema *s, CsSchema *p);
    extern BOOLEAN goal_of_schema_is_achieved(CsSchema *s);
    extern CsSchema *selected_parent_of(CsSchema *s);
    extern BOOLEAN intended_goal(char *goal);
    extern BOOLEAN schema_has_been_activated(CsSchema *s);

    CsSchema *parent;

    if (intended_goal(s->goal)) {
        return(1.0);
    }
#ifndef SELF_INHIBIT
    else if (goal_of_schema_is_achieved(s)) {
        return(0.0);
    }
#endif
    else if ((parent = selected_parent_of(s)) != NULL) {
        if (!ordering_constraints_satisfied(s, parent)) {
            return(0.0);
        }
/*  Added Sat Feb  6 18:43:53 1999                  */
/*        else if (schema_has_been_activated(s)) {  */
/*            return(0.0);                          */
/*	}                                           */
/*  End of Sat Feb  6 18:44:05 1999                 */
        else {
            int n = length(parent->children);
            return( (parent->activation[cycle-1]-MIN_ACT) / ((MAX_ACT-MIN_ACT) * n) );
        }
    }
    else {
        return(0.0);
    }
}

/******************************************************************************/

double SchemaEnvironmentalInfluence(int cycle, CsSchema *schema)
{
    return(squash(schema->triggering_situation(cycle)));
}

/******************************************************************************/

double SchemaLateralInfluence(int cycle, CsSchema *s)
{
    extern BOOLEAN competitor_of(CsSchema *s1, CsSchema *s2);

    CsSchema *t;
    double sum = 0.0;

    for (t = schemas; t != NULL; t = t->next) {
        if (competitor_of(s, t)) {
            sum += (REST_ACT - t->activation[cycle-1]);
        }
    }
    return( sum / (MAX_ACT - MIN_ACT) );
}

/******************************************************************************/

static int self_influence_factor(CsSchema *s)
{
    extern BOOLEAN goal_of_schema_is_achieved(CsSchema *s);

#ifndef SELF_INHIBIT
    if ((s->selected) && goal_of_schema_is_achieved(s)) {
#else
    if (goal_of_schema_is_achieved(s)) {
#endif
        return(-1);
    }
    else {
        return(1);
    }
}

double SchemaSelfInfluence(int cycle, CsSchema *s)
{
    return( self_influence_factor(s) * (s->activation[cycle-1]-MIN_ACT) / (MAX_ACT-MIN_ACT) );
}

/******************************************************************************/

double CsCalculateSchemaActivation(int cycle, CsSchema *s)
{
    extern int count_competitors(CsSchema *s);

    double parent = SCHEMA_PE * SchemaIntentionalInfluence(cycle, s);
    double environment = (1.0-SCHEMA_PE) * SchemaEnvironmentalInfluence(cycle, s);
    double self = SCHEMA_SL * SchemaSelfInfluence(cycle, s);
    double lateral = (1.0-SCHEMA_SL) * SchemaLateralInfluence(cycle, s);
    double system =  SCHEMA_SI * (lateral + self);
    double input = (1.0-SCHEMA_SI) * (parent + environment);

    return( apply_equations(s->activation[cycle-1], system + input) );

}

/******************************************************************************/
/*  OBJECT ACTIVATION FLOW                                                    */
/******************************************************************************/

static double schema_affects_object(CsSchema *schema, CsObject *object, ArgumentFunction function, int cycle)
{
    extern BOOLEAN object_can_serve_function();

    if (object_can_serve_function(object, function) && schema->object_triggering_function(object, function, cycle)) {
        return((schema->activation[cycle-1] - REST_ACT) / (MAX_ACT - MIN_ACT));
    }
    else {
        return(0.0);
    }
}

double ObjectIntentionalInfluence(int cycle, CsObject *object, ArgumentFunction f)
{
    double total = 0.0;
    CsSchema *t;

    for (t = schemas; t != NULL; t= t->next) {
        total += schema_affects_object(t, object, f, cycle);
    }
    return(squash(total));
}

/******************************************************************************/

static double ObjectEnvironmentalInfluence(int cycle, CsObject *object, ArgumentFunction f)
{
    return(squash(object->triggering_situation(cycle, f)));
}

/******************************************************************************/

double ObjectLateralInfluence(int cycle, CsObject *object, ArgumentFunction f)
{
    CsObject *tmp;
    double total = 0.0;

    for (tmp = objects; tmp != NULL; tmp = tmp->next) {
        if (tmp != object) {
            total += (REST_ACT - tmp->activation[f][cycle-1]);
        }
    }
    return(total / (MAX_ACT - MIN_ACT));
}

/******************************************************************************/

double ObjectSelfInfluence(int cycle, CsObject *object, ArgumentFunction f)
{
    return(object->activation[f][cycle-1]);
}

/******************************************************************************/

double CsCalculateObjectActivation(int cycle, CsObject *object, ArgumentFunction f)
{
    double parent = OBJECT_PE * ObjectIntentionalInfluence(cycle, object, f);
    double environment = (1.0-OBJECT_PE) * ObjectEnvironmentalInfluence(cycle, object, f);
    double self = OBJECT_SL * ObjectSelfInfluence(cycle, object, f);
    double lateral = (1.0-OBJECT_SL) * ObjectLateralInfluence(cycle, object, f);
    double system = OBJECT_SI * (lateral + self);
    double input = (1.0-OBJECT_SI) * (parent + environment);

    return( apply_equations(object->activation[f][cycle-1], system + input) );
}

/******************************************************************************/
/*  RESOURCE ACTIVATION FLOW                                                  */
/******************************************************************************/

static double schema_affects_resource(CsSchema *schema, CsResource *effector, int cycle)
{
    if (schema->effector_triggering_function(effector, cycle)) {
        return((schema->activation[cycle-1]-REST_ACT) / (MAX_ACT - MIN_ACT));
    }
    else {
        return(0.0);
    }
}

double ResourceIntentionalInfluence(int cycle, CsResource *effector)
{
    double total = 0.0;
    CsSchema *t;

    for (t = schemas; t != NULL; t = t->next) {
        total += schema_affects_resource(t, effector, cycle);
    }
    return(squash(total));
}

double ResourceEnvironmentalInfluence(int cycle, CsResource *effector)
{
    return(0.0);
}

double ResourceSelfInfluence(int cycle, CsResource *effector)
{
    return(effector->activation[cycle-1]);
}

double ResourceLateralInfluence(int cycle, CsResource *effector)
{
    CsResource *tmp;
    double total = 0.0;

    for (tmp = resources; tmp != NULL; tmp = tmp->next) {
        if (tmp != effector) {
            total += (REST_ACT - tmp->activation[cycle-1]);
        }
    }
    return(total / (MAX_ACT - MIN_ACT));
}

/******************************************************************************/

double CsCalculateResourceActivation(int cycle, CsResource *effector)
{
    double parent = EFFECT_PE * ResourceIntentionalInfluence(cycle, effector);
    double environment = (1.0-EFFECT_PE) * ResourceEnvironmentalInfluence(cycle, effector);
    double self = EFFECT_SL * ResourceSelfInfluence(cycle, effector);
    double lateral = (1.0-EFFECT_SL) * ResourceLateralInfluence(cycle, effector);
    double system = EFFECT_SI * (lateral + self);
    double input = (1.0-EFFECT_SI) * (parent + environment);

    return( apply_equations(effector->activation[cycle-1], system + input) );
}

/******************************************************************************/
