/******************************************************************************/

#include <math.h>
#include "cs.h"

/******************************************************************************/

extern void CsMessage();
extern double CsRandom();
extern void post_action_trace(GENERIC_LIST *features);
extern void CsActionFailureMessage(int cycle, char *reason);
extern void CsArgumentFailureMessage(int cycle, char *action);

extern TERM *get_feature(GENERIC_LIST *features, char *functor);
extern BOOLEAN has_feature(GENERIC_LIST *features, char *functor);

/******************************************************************************/

BOOLEAN ColavitaIsVisualObject(CsObject *object)
{
    return( strcmp(object->name, "visual") == 0 );
}

BOOLEAN ColavitaIsVisualStimulus(CsObject *object)
{ 
    return(ColavitaIsVisualObject(object) && !has_feature(object->features, "light off"));
}

BOOLEAN ColavitaIsVisualTarget(CsObject *object)
{
    return(ColavitaIsVisualStimulus(object) && has_feature(object->features, "light on"));
}

BOOLEAN ColavitaIsAuditoryObject(CsObject *object)
{
    return( strcmp(object->name, "auditory") == 0 );
}

BOOLEAN ColavitaIsAuditoryStimulus(CsObject *object)
{
    return(ColavitaIsAuditoryObject(object) && !has_feature(object->features, "tone off"));
}

BOOLEAN ColavitaIsAuditoryTarget(CsObject *object)
{
    return(ColavitaIsAuditoryStimulus(object) && has_feature(object->features, "tone on"));
}

BOOLEAN ColavitaIsLeftHand(CsResource *effector)
{
    return( (strcmp(effector->name, "Left Hand") == 0) );
}

BOOLEAN ColavitaIsRightHand(CsResource *effector)
{
    return( (strcmp(effector->name, "Right Hand") == 0) );
}

BOOLEAN ColavitaIsLeftButton(CsObject *object)
{
    return( strcmp(object->name, "l_button") == 0 );
}

BOOLEAN ColavitaIsRightButton(CsObject *object)
{
    return( strcmp(object->name, "r_button") == 0 );
}

BOOLEAN ColavitaIsPressedButton(CsObject *object)
{
    return(has_feature(object->features, "pressed"));
}

BOOLEAN ColavitaIsUnpressedButton(CsObject *object)
{
    return(has_feature(object->features, "unpressed"));
}

BOOLEAN RtAwaitingResponse(char *button_name)
{
    CsObject *button;

    for (button = objects; button != NULL; button = button->next) {
        if (strcmp(button->name, button_name) == 0) {
            if ((ColavitaIsPressedButton(button)) || (ColavitaIsUnpressedButton(button))) {
                return(FALSE);
            }
        }
    }
    return(TRUE);
}


BOOLEAN ColavitaIsHand(CsResource *effector)
{
    return(TRUE);
}

/*******************************************************************************

For the RT task we need a procedure to manage external changes to the world.

This is performed by external_world/1 (and all the minor static functions
below), which controls the stimulus in the world:
    1. check if it's time to turn the stimulus on/off
    2. depending on the experiment, change the stimulus.
The return value is used for keeping track of the number of trials:
    return TRUE on completion of a trial, FALSE otherwise.

A major complication is introduced by the need to introduce a time delay
between the occurence of changes in the world and their reflection in the
model. This delay is intended to model perceptual processes, and is therefore
modality dependent. Visual changes, for example, are delayed more than
auditory changes. Separate parameters for each modality control this delay.

*******************************************************************************/

static CsObject *find_visual_object(CsObject *objects)
{
    while ((objects != NULL) && (!ColavitaIsVisualObject(objects))) {
        objects = objects->next;
    }
    return(objects);
}

static void record_visual_change(int cycle, char *state, char *type)
{
    char buf[80];

    sprintf(buf, "        visual stimulus: %s", state);
    CsMessage(ACTION, cycle, buf);
}

static char *set_visual_off(int cycle)
{
    record_visual_change(cycle, "light_off", "f");
    return(NULL);
}

static char *set_visual_on(int cycle, char trial_type)
{
    char *COLOUR = "light on";
    static char tmp[10];

    sprintf(tmp, "t(%c)", trial_type);
    record_visual_change(cycle, COLOUR, tmp);
    return(COLOUR);
}

static void delayed_off_visual(int cycle)
{

    CsObject *stimulus = find_visual_object(objects);
    TERM *functor = get_feature(stimulus->features, "light on");

    if (functor == NULL) {
        functor = get_feature(stimulus->features, "light off");
    }
    if (functor != NULL) {
        functor->functor = copy_string("light off");
    }
}

static void delayed_set_visual(int cycle, char *colour)
{
    if (colour == NULL) {
        delayed_off_visual(cycle);
    }
    else {
        CsObject *stimulus = find_visual_object(objects);
        TERM *functor = get_feature(stimulus->features, "light off");

        if (functor == NULL) {
            fprintf(stderr, "Error: Cycle %d: Visual stimulus should have been off.\n", cycle); 
            delayed_off_visual(cycle);
            delayed_set_visual(cycle, colour);

        }
        else {
            functor->functor = copy_string(colour);
        }
    }
}

/*----------------------------------------------------------------------------*/

static CsObject *find_auditory_object(CsObject *object)
{
    while ((object != NULL) && (!ColavitaIsAuditoryObject(object))) {
        object = object->next;
    }
    return(object);
}

static void record_auditory_change(int cycle, char *state, char *type)
{
    char buf[80];

    sprintf(buf, "        auditory stimulus: %s", state);
    CsMessage(ACTION, cycle, buf);
}

static char *set_auditory_off(int cycle)
{
    record_auditory_change(cycle, "tone off", "f");
    return(NULL);
}

static char *set_auditory_on(int cycle, char trial_type)
{
    char *TONE = "tone on";
    static char tmp[10];

    sprintf(tmp, "t(%c)", trial_type);
    record_auditory_change(cycle, TONE, tmp);
    return(TONE);
}

static void delayed_off_auditory(int cycle)
{

    CsObject *stimulus = find_auditory_object(objects);
    TERM *functor = get_feature(stimulus->features, "tone on");

    if (functor == NULL) {
        functor = get_feature(stimulus->features, "tone off");
    }
    if (functor != NULL) {
        functor->functor = copy_string("tone off");
    }
}

static void delayed_set_auditory(int cycle, char *tone)
{
    if (tone == NULL) {
            delayed_off_auditory(cycle);
    }
    else {
        CsObject *stimulus = find_auditory_object(objects);
        TERM *functor = get_feature(stimulus->features, "tone off");

        if (functor == NULL) {
            fprintf(stderr, "Error: Cycle %d: Auditory stimulus should have been off.\n", cycle); 
            delayed_off_auditory(cycle);
            delayed_set_auditory(cycle, tone);
        }
        else {
            functor->functor = copy_string(tone);
        }
    }
}

/*----------------------------------------------------------------------------*/

static void set_button(CsObject *button, char *state)
{
    TERM *functor;

    if (button == NULL) {
        fprintf(stderr, "NULL button in set_button");
    }
    else if ((functor = get_feature(button->features, "off")) != NULL) {
        functor->functor = copy_string(state);
    }
    else if ((functor = get_feature(button->features, "pressed")) != NULL) {
        functor->functor = copy_string(state);
    }
    else if ((functor = get_feature(button->features, "unpressed")) != NULL) {
        functor->functor = copy_string(state);
    }
}

static void set_l_button(char *state)
{
    CsObject *tmp = objects;

    while ((tmp != NULL) && (!ColavitaIsLeftButton(tmp))) {
        tmp = tmp->next;
    }
    set_button(tmp, state);
}

static void set_r_button(char *state)
{
    CsObject *tmp = objects;

    while ((tmp != NULL) && (!ColavitaIsRightButton(tmp))) {
        tmp = tmp->next;
    }
    set_button(tmp, state);
}

static void press_button(int cycle, CsResource *effector)
{
    if (strcmp(effector->name, "Left Hand") == 0) {
        delayed_off_auditory(cycle);
        set_l_button("pressed");
    }
    else if (strcmp(effector->name, "Right Hand") == 0) {
        delayed_off_visual(cycle);
        set_r_button("pressed");
    }
}

static void dont_press_button(CsResource *effector)
{
    if (strcmp(effector->name, "Left Hand") == 0) {
        set_l_button("unpressed");
    }
    else if (strcmp(effector->name, "Right Hand") == 0) {
        set_r_button("unpressed");
    }
}

static void reset_responses()
{
    set_l_button("off");
    set_r_button("off");
}

/******************************************************************************/

static void adjust_visual_world(int cycle, BOOLEAN turn_on, BOOLEAN turn_off, char **colour)
{
    if (turn_on) {
        *colour = set_visual_on(cycle, 'b');
    }
    if (turn_off) {
        *colour = set_visual_off(cycle);
    }
}

static void adjust_auditory_world(int cycle, BOOLEAN turn_on, BOOLEAN turn_off, char **tone)
{
    if (turn_on) {
        *tone = set_auditory_on(cycle, 'b');
    }
    if (turn_off) {
       *tone = set_auditory_off(cycle);
    }
}

static void adjust_choice_world(int cycle, BOOLEAN turn_on, BOOLEAN turn_off, char **colour, char **tone)
{
    if (CsRandom() > 0.5) {
        adjust_visual_world(cycle, turn_on, turn_off, colour);
        *tone = NULL;
    }
    else {
        adjust_auditory_world(cycle, turn_on, turn_off, tone);
        *colour = NULL;
    }
}

static void adjust_conflict_world(int cycle, BOOLEAN turn_on, BOOLEAN turn_off, char **colour, char **tone)
{
    adjust_visual_world(cycle, turn_on, turn_off, colour);
    adjust_auditory_world(cycle, turn_on, turn_off, tone);
    *tone = NULL;
    *colour = NULL;
}

/******************************************************************************/

/* Should the stimulus be on at cycle n? */

static BOOLEAN is_stimulus_on(int n, int ms_on, int ms_off, int pause)
{
    double ms_elapsed = MS_PER_CYCLE * n - pause;

    while (ms_elapsed > (ms_on + ms_off)) {
        ms_elapsed = ms_elapsed - (ms_on + ms_off);
    }
    return((ms_elapsed > 0) && (ms_elapsed < ms_on));
}

BOOLEAN SimpleVisualStepWorld(int cycle)
{
    int MS_ON = 400, MS_OFF = 400, PAUSE = 400; /* okay with 1 ms/cycle */

    BOOLEAN is_on = is_stimulus_on(cycle, MS_ON, MS_OFF, PAUSE);
    BOOLEAN was_on = is_stimulus_on(cycle-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN turn_on = (is_on && !was_on);
    BOOLEAN turn_off = (was_on && !is_on);

    BOOLEAN v_is_on = is_stimulus_on(cycle-(int) (VISUAL_LAG/MS_PER_CYCLE), MS_ON, MS_OFF, PAUSE);
    BOOLEAN v_was_on = is_stimulus_on(cycle-(int) (VISUAL_LAG/MS_PER_CYCLE)-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN v_turn_on = (v_is_on && !v_was_on);
    BOOLEAN v_turn_off = (v_was_on && !v_is_on);

    static char *colour = NULL;

    if (turn_off) {
        reset_responses();
    }
    else if (turn_on) {
        adjust_visual_world(cycle, turn_on, turn_off, &colour);
    }

    if (v_turn_off) {
        delayed_set_visual(cycle, NULL);
    }
    else if ((v_turn_on) && (colour != NULL)) {
        delayed_set_visual(cycle, colour);
    }

    return(turn_on);
}

BOOLEAN SimpleAuditoryStepWorld(int cycle)
{
    int MS_ON = 400, MS_OFF = 400, PAUSE = 400; /* okay with 1 ms/cycle */

    BOOLEAN is_on = is_stimulus_on(cycle, MS_ON, MS_OFF, PAUSE);
    BOOLEAN was_on = is_stimulus_on(cycle-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN turn_on = (is_on && !was_on);
    BOOLEAN turn_off = (was_on && !is_on);

    BOOLEAN a_is_on = is_stimulus_on(cycle-(int) (AUDITORY_LAG/MS_PER_CYCLE), MS_ON, MS_OFF, PAUSE);
    BOOLEAN a_was_on = is_stimulus_on(cycle-(int) (AUDITORY_LAG/MS_PER_CYCLE)-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN a_turn_on = (a_is_on && !a_was_on);
    BOOLEAN a_turn_off = (a_was_on && !a_is_on);

    static char *tone = NULL;

    if (turn_off) {
        reset_responses();
    }
    else if (turn_on) {
        adjust_auditory_world(cycle, turn_on, turn_off, &tone);
    }

    if (a_turn_off) {
        delayed_set_auditory(cycle, NULL);
    }
    else if ((a_turn_on) && (tone != NULL)) {
        delayed_set_auditory(cycle, tone);
    }

    return(turn_on);
}

BOOLEAN ChoiceReactionTimeStepWorld(int cycle)
{
    int MS_ON = 400, MS_OFF = 400, PAUSE = 400; /* okay with 1 ms/cycle */

    BOOLEAN is_on = is_stimulus_on(cycle, MS_ON, MS_OFF, PAUSE);
    BOOLEAN was_on = is_stimulus_on(cycle-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN turn_on = (is_on && !was_on);
    BOOLEAN turn_off = (was_on && !is_on);

    BOOLEAN a_is_on = is_stimulus_on(cycle-(int) (AUDITORY_LAG/MS_PER_CYCLE), MS_ON, MS_OFF, PAUSE);
    BOOLEAN a_was_on = is_stimulus_on(cycle-(int) (AUDITORY_LAG/MS_PER_CYCLE)-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN a_turn_on = (a_is_on && !a_was_on);
    BOOLEAN a_turn_off = (a_was_on && !a_is_on);

    BOOLEAN v_is_on = is_stimulus_on(cycle-(int) (VISUAL_LAG/MS_PER_CYCLE), MS_ON, MS_OFF, PAUSE);
    BOOLEAN v_was_on = is_stimulus_on(cycle-(int) (VISUAL_LAG/MS_PER_CYCLE)-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN v_turn_on = (v_is_on && !v_was_on);
    BOOLEAN v_turn_off = (v_was_on && !v_is_on);

    static char *colour = NULL;
    static char *tone = NULL;

    if (cycle == 0) {
        tone = NULL;
        colour = NULL;
    }

    if (turn_off) {
        reset_responses();
    }
    else if (turn_on) {
        adjust_choice_world(cycle, turn_on, turn_off, &colour, &tone);
    }

    if (a_turn_off) {
        delayed_set_auditory(cycle, NULL);
    }
    else if ((a_turn_on) && (tone != NULL)) {
        delayed_set_auditory(cycle, tone);
    }
        
    if (v_turn_off) {
        delayed_set_visual(cycle, NULL);
    }
    else if ((v_turn_on) && (colour != NULL)) {
        delayed_set_visual(cycle, colour);
    }

    return(turn_on);
}

BOOLEAN ConflictReactionTimeStepWorld(int cycle)
{
    int MS_ON = 400, MS_OFF = 400, PAUSE = 400; /* okay with 1 ms/cycle */

    BOOLEAN is_on = is_stimulus_on(cycle, MS_ON, MS_OFF, PAUSE);
    BOOLEAN was_on = is_stimulus_on(cycle-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN turn_on = (is_on && !was_on);
    BOOLEAN turn_off = (was_on && !is_on);

    BOOLEAN a_is_on = is_stimulus_on(cycle-(int) (AUDITORY_LAG/MS_PER_CYCLE), MS_ON, MS_OFF, PAUSE);
    BOOLEAN a_was_on = is_stimulus_on(cycle-(int) (AUDITORY_LAG/MS_PER_CYCLE)-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN a_turn_on = (a_is_on && !a_was_on);
    BOOLEAN a_turn_off = (a_was_on && !a_is_on);

    BOOLEAN v_is_on = is_stimulus_on(cycle-(int) (VISUAL_LAG/MS_PER_CYCLE), MS_ON, MS_OFF, PAUSE);
    BOOLEAN v_was_on = is_stimulus_on(cycle-(int) (VISUAL_LAG/MS_PER_CYCLE)-1, MS_ON, MS_OFF, PAUSE);
    BOOLEAN v_turn_on = (v_is_on && !v_was_on);
    BOOLEAN v_turn_off = (v_was_on && !v_is_on);

    static char *colour = NULL;
    static char *tone = NULL;

    if (cycle == 0) {
        tone = NULL;
        colour = NULL;
    }

    if (turn_off) {
        reset_responses();
    }
    else if (turn_on) {
        adjust_conflict_world(cycle, turn_on, turn_off, &colour, &tone);
    }

    if (a_turn_off) {
        delayed_set_auditory(cycle, NULL);
    }
    else if ((a_turn_on) && (tone != NULL)) {
        delayed_set_auditory(cycle, tone);
    }
        
    if (v_turn_off) {
        delayed_set_visual(cycle, NULL);
    }
    else if ((v_turn_on) && (colour != NULL)) {
        delayed_set_visual(cycle, colour);
    }

    return(turn_on);
}

/******************************************************************************/
/*******************************************************************************

The dual task has three actions, pressing a button, preparing to press a button,
and relaxing that preparation. All can be done with either hand.

*******************************************************************************/

void press(int cycle, CsResource *effector)
{
    if (effector == NULL) {
        CsArgumentFailureMessage(cycle, "press");
    }
    else {
        char buf[80];

        sprintf(buf, "        Press with %s", effector->name);
        CsMessage(ACTION, cycle, buf);

        press_button(cycle, effector);
    }
}

void prepare(int cycle, CsResource *effector)
{
    if (effector == NULL) {
        CsArgumentFailureMessage(cycle, "Prepare");
    }
    else {
        char buf[80];

        sprintf(buf, "        Prepare with %s", effector->name);
        CsMessage(ACTION, cycle, buf);
    }
}

void relax(int cycle, CsResource *effector)
{
    if (effector == NULL) {
        CsArgumentFailureMessage(cycle, "Relax");
    }
    else {
        char buf[80];

        sprintf(buf, "        Relax with %s", effector->name);
        CsMessage(ACTION, cycle, buf);

        dont_press_button(effector);
    }
}

/******************************************************************************/
