#include "cs.h"

#define PI 3.141592653

#include <math.h>

/******************************************************************************/

BOOLEAN common_goal(CsSchema *i, CsSchema *j)
{
    return(strcmp(i->goal, j->goal) == 0);
}

/******************************************************************************/

BOOLEAN common_subgoal(CsSchema *i, CsSchema *j)
{
    Subgoal *l1 = i->children;
    Subgoal *l2 = j->children;

    for (l1 = i->children; l1 != NULL; l1 = l1->next) {
        for (l2 = j->children; l2 != NULL; l2 = l2->next) {
            if (strcmp(l1->subgoal, l2->subgoal) == 0) {
                return(TRUE);
            }
        }
    }
    return(FALSE);
}

/******************************************************************************/

BOOLEAN common_resource(CsSchema *i, CsSchema *j)
{
    GENERIC_LIST *l1 = i->res;
    GENERIC_LIST *l2 = j->res;

    if ((l1 == NULL) || (l2 == NULL)) {
        return(FALSE);
    }
    else {
        while ((l1 != NULL) && (l1->head != l2->head)) {
            while ((l2 != NULL) && (l1->head != l2->head)) {
                l2 = l2->next;
            }
            if (l2 == NULL) {
                l2 = j->res;
                l1 = l1->next;
            }
        }
        return( (l1 != NULL) && (l2 != NULL) );
    }
}

/******************************************************************************/

/*
 * competitor_of_schema(Schema, Competitor) :-
 *     schema_share_goal(Schema, Competitor),
 *     Schema \== Competitor.
 * competitor_of_schema(Schema, Competitor) :-
 *     schema_share_sub_goal(Schema, Competitor),
 *     Schema \== Competitor,
 *     \+ schema_share_goal(Schema, Competitor).
 * competitor_of_schema(Schema, Competitor) :-
 *     schema_share_resource(Schema, Competitor),
 *     Schema \== Competitor,
 *     \+ schema_share_goal(Schema, Competitor),
 *     \+ schema_share_sub_goal(Schema, Competitor).
 */                                         

BOOLEAN competitor_of(CsSchema *i, CsSchema *j)
{
    if (i == j) {
        return(FALSE);
    }
    else {
        return(common_goal(i, j) || common_subgoal(i, j) || common_resource(i, j));
    }
}

int count_competitors(CsSchema *s)
{
    int n = 0;
    CsSchema *t;

    for (t = schemas; t != NULL; t = t->next) {
        if (competitor_of(s, t)) {
            n++;
        }
    }
    return(n);
}


/******************************************************************************/

/*
 * There are three cases for when a schema's goal is achieved:
 *
 * 1) If the schema is selected and doesn't correspond to a primitive action,
 * then it is achieved if all its subgoals are marked as being achieved;
 * 2) If the schema is selected and does correspond to a primitive action,
 * then it is achieved if its corresponding action has been executed;
 * 3) Otherwise (if the schema is unselected), it is achieved if its goal
 * is marked as an achieved subgoal of some other (parent) schema.
 *
 * goal_of_schema_is_achieved(Schema) :-
 *     schema_status(Schema, selected, Subgoals),
 *     not( primitive_action(Schema, _, _, _) ),
 *     not( member(-_, Subgoals) ).
 *
 * goal_of_schema_is_achieved(Schema) :-
 *     schema_status(Schema, selected, _),
 *     primitive_action(Schema, Action, _, _),
 *     action_has_been_executed(Schema, Action).
 *
 * goal_of_schema_is_achieved(Schema) :-
 *     schema_status(Schema, unselected, _),
 *     schema_status(_Parent, selected, Subgoals),
 *     schema_goal(Schema, Goal),
 *     member(+Goal, Subgoals),
 *     not( member(-Goal, Subgoals) ).
 */

BOOLEAN continuous_goal(char *goal)
{
    if (strcmp(goal, "Prepare") == 0) {   /* The special "Prepare" schema is never achieved! */
        return(TRUE);
    }
    else if (strcmp(goal, "Relax2") == 0) {   /* The special "Relax2" schema is never achieved! */
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

static BOOLEAN all_subgoals_achieved(CsSchema *i)
{
    Subgoal *tmp;

    for (tmp = i->children; tmp != NULL; tmp = tmp->next) {
        if (!tmp->achieved) {
            return(FALSE);
        }
    }
    return(TRUE);
}

static BOOLEAN goal_achieved_by_other_schema(char *goal)
{
    CsSchema *tmp;
    Subgoal *sg;

    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        if (tmp->selected) {
            if ((strcmp(tmp->goal, goal) == 0) && all_subgoals_achieved(tmp)) {
                return(TRUE);
            }
            for (sg = tmp->children; sg != NULL; sg = sg->next) {
                if ((sg->achieved) && (strcmp(goal, sg->subgoal) == 0)) {
                    return(TRUE);
                }
            }
        }
    }
    return(FALSE);
}

BOOLEAN goal_of_schema_is_achieved(CsSchema *schema)
{
    if (continuous_goal(schema->goal)) {
        return(FALSE);
    }
    else if (schema->selected) {
        if ((schema->corresponding_action == NONE)) {
            return(all_subgoals_achieved(schema));
        }
        else {
            return(schema->executed);
        }
    }
    else {
        return(goal_achieved_by_other_schema(schema->goal));
    }
}

/******************************************************************************/
/******************************************************************************/

BOOLEAN ordering_constraints_satisfied(CsSchema *child, CsSchema *parent)
{
    Subgoal  *sg = parent->children;
    GENERIC_LIST *tmp;

    while (strcmp(sg->subgoal, child->goal) != 0) {
        sg = sg->next;
    }
    tmp = sg->prior_subgoals;

    while ((tmp != NULL) && (((Subgoal *)(tmp->head))->achieved)) {
        tmp = tmp->next;
    }
    return(tmp == NULL);
}

/******************************************************************************/

BOOLEAN precondition_of_achieved_crux(CsSchema *child, CsSchema *parent)
{
    Subgoal *tmp = parent->children;

    while (strcmp(tmp->subgoal, child->goal) != 0) {
        tmp = tmp->next;
    }
    if (tmp->sort == PRECONDITION) {
        for (tmp = parent->children; tmp != NULL; tmp = tmp->next) {
            if ((tmp->sort == CRUX) && (tmp->achieved)) {
                return(FALSE);
            }
        }
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

/******************************************************************************/

BOOLEAN schema_has_been_activated(CsSchema *schema)
{
    Subgoal *tmp;

/* A first pass: true when we've done at least one subgoal */

    for (tmp = schema->children; tmp != NULL; tmp = tmp->next) {
        if (tmp->achieved == TRUE) {
            return(TRUE);
        }
    }
    return(FALSE);
}

/******************************************************************************/

BOOLEAN parent_of(CsSchema *p, CsSchema *c)
{
    Subgoal *tmp;

    for (tmp = p->children; tmp != NULL; tmp = tmp->next) {
        if (strcmp(tmp->subgoal, c->goal) == 0) {
            return(TRUE);
        }
    }
    return(FALSE);
}

/******************************************************************************/

CsSchema *is_parent_of(CsSchema *c)
{
    CsSchema *tmp;

    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        if (parent_of(tmp, c)) {
            return(tmp);
        }
    }
    return(NULL);
}

/******************************************************************************/

CsSchema *selected_parent_of(CsSchema *i)
{
    CsSchema *tmp;

    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        if ((tmp->selected) && (parent_of(tmp, i))) {
            return(tmp);
        }
    }
    return(NULL);
}

/******************************************************************************/

BOOLEAN intended_goal(char *goal)
{
    extern char *terminating_goal;

    return(strcmp(goal, terminating_goal) == 0);
}

/******************************************************************************/
