/******************************************************************************/

#include "cs.h"

#include <math.h>
#include <varargs.h>

#ifndef PI
#define PI 3.141592653
#endif

/******************************************************************************/

double net_influence(n, va_alist)
int n;
va_dcl
{
    va_list ap;
    int i;
    double sum = 0.0;

    va_start(ap);

    for (i=0; i < n; i++) {
        sum += (va_arg(ap, double) - REST_ACT) / (MAX_ACT - MIN_ACT);
    }

    va_end(ap);
    return(sum/n);
}

/******************************************************************************/

double sigmoid(double x)
{
    return(1.0 / (1.0 + exp(-x)));
}

double inverse_sigmoid(double y)
{
    return(log(y / (1.0 - y)));
}

/*
 * squash(x) is the same as 2 * sigmoid(x) - 1
 * It returns a value between -1 and 1, such that
 *     squash(0) = 0;
 *     squash(x) = -squash(-x);
 */

double squash(double x) 
{
    double y = exp(x);
    return((y-1.0)/(y+1.0));
}

/******************************************************************************/

static double bounding_function_sig(double x)
{
    double r = (REST_ACT - ((MAX_ACT+MIN_ACT)/2.0))*(2.0/(MAX_ACT-MIN_ACT));
    double fx = x + log((1.0+r)/(1.0-r));

    return( (squash(fx) * (MAX_ACT-MIN_ACT)/2.0) + ((MAX_ACT+MIN_ACT)/2.0) );
}

/******************************************************************************/

static double inverse_bounding_function_sig(double x)
{
    double r = (REST_ACT - ((MAX_ACT+MIN_ACT)/2.0))*(2.0/(MAX_ACT-MIN_ACT));
    double t = (  x   -    ((MAX_ACT+MIN_ACT)/2.0))*(2.0/(MAX_ACT-MIN_ACT));
    double a = log((1.0+r)/(1.0-r));

    return( (log((1.0 + t) / (1.0 - t)) - a) );
}

/******************************************************************************/

static double bounding_function_tanh(double x)
{
    double r = (REST_ACT - ((MAX_ACT+MIN_ACT)/2.0))*(2.0/(MAX_ACT-MIN_ACT));
    double fx = (2.0 * x) + log((1.0+r)/(1.0-r));

    return( (squash(fx) * (MAX_ACT-MIN_ACT)/2.0) + ((MAX_ACT+MIN_ACT)/2.0) );
}

/******************************************************************************/

static double inverse_bounding_function_tanh(double x)
{
    double r = (REST_ACT - ((MAX_ACT+MIN_ACT)/2.0))*(2.0/(MAX_ACT-MIN_ACT));
    double t = (  x   -    ((MAX_ACT+MIN_ACT)/2.0))*(2.0/(MAX_ACT-MIN_ACT));
    double a = log((1.0+r)/(1.0-r));

    return( (log((1.0 + t) / (1.0 - t)) - a) / 2.0 );
}

/******************************************************************************/

static double bounding_function_halfx(double x)
{
    double r = (REST_ACT - ((MAX_ACT+MIN_ACT)/2.0))*(2.0/(MAX_ACT-MIN_ACT));
    double fx = (x / 2.0) + log((1.0+r)/(1.0-r));

    return( (squash(fx) * (MAX_ACT-MIN_ACT)/2.0) + ((MAX_ACT+MIN_ACT)/2.0) );
}

/******************************************************************************/

static double inverse_bounding_function_halfx(double x)
{
    double r = (REST_ACT - ((MAX_ACT+MIN_ACT)/2.0))*(2.0/(MAX_ACT-MIN_ACT));
    double t = (  x   -    ((MAX_ACT+MIN_ACT)/2.0))*(2.0/(MAX_ACT-MIN_ACT));
    double a = log((1.0+r)/(1.0-r));

    return( (log((1.0 + t) / (1.0 - t)) - a) * 2.0 );
}

/******************************************************************************/

/*
 * Random number generator (from SICStus 2.1 distribution)
 */

static short RAND1 = 27134;
static short RAND2 = 9213;
static short RAND3 = 17773;

double CsRandom()
{
  int X,Y,Z;
  double T;
  
  X = (RAND1*171) % 30269;
  Y = (RAND2*172) % 30307;
  Z = (RAND3*170) % 30323;
  T = X/30269.0 + Y/30307.0 + Z/30323.0;
  RAND1 = X;
  RAND2 = Y;
  RAND3 = Z;
  return(T-(int)T);
}    

char *random_seed()
{
    static char buffer[30];

    sprintf(buffer, "%d:%d:%d", RAND1, RAND2, RAND3);
    return(buffer);
}

/******************************************************************************/

static double normal_noise(double mean, double sd)
{
    return(mean + sd * sqrt(-2 * log(CsRandom())) * cos(2 * PI * CsRandom()));
}

/******************************************************************************/

double apply_equations(double previous, double input)
{
    double epsilon;
    double noisey_input = normal_noise(input, NOISE_SD);

    switch (MATHS) {
    case SIGMOID:
        return( bounding_function_sig(inverse_bounding_function_sig(previous) * PERSISTENCE + noisey_input) );
        break;
    case TANH:
        return( bounding_function_tanh(inverse_bounding_function_tanh(previous) * PERSISTENCE + noisey_input) );
        break;
    case HALFX:
        return( bounding_function_halfx(inverse_bounding_function_halfx(previous) * PERSISTENCE + noisey_input) );
        break;
    case M_R:
        if ((noisey_input > 0.0) && (previous < MAX_ACT)) {
            epsilon = noisey_input * (MAX_ACT - previous);
        }
        else if ((noisey_input < 0.0) && (previous > MIN_ACT)) {
            epsilon = noisey_input * (previous - MIN_ACT);
        }
        else {
            epsilon = 0.0;
        }
        return( (PERSISTENCE * previous) + ((1-PERSISTENCE)*REST_ACT) + epsilon);
        break;
    case HOUGHTON:
        if ((noisey_input > 0.0) && (previous < MAX_ACT)) {
            epsilon = squash(noisey_input) * (MAX_ACT - previous);
        }
        else if ((noisey_input < 0.0) && (previous > MIN_ACT)) {
            epsilon = squash(noisey_input) * (previous - MIN_ACT);
        }
        else {
            epsilon = 0.0;
        }
        return( (PERSISTENCE * previous) + ((1-PERSISTENCE)*REST_ACT) + epsilon);
        break;
    case NORMALISED:
        epsilon = ((1.0 - PERSISTENCE) * squash(noisey_input)) / 3.0;
        return((PERSISTENCE * (previous - MIN_ACT) / (MAX_ACT - MIN_ACT)) + epsilon);
        break;
    default:
        fprintf(stdout, "ERROR: Unknown maths system [%d]\n", MATHS);
        return( 0.0 );
        break;
    }
}

/******************************************************************************/
