/*******************************************************************************
This file contains:
    functions for making an object representation network from the definitions.

*******************************************************************************/

#include "cs.h"

/******************************************************************************/

extern double apply_equations(double previous, double input);
extern TERM *create_functor();
extern void free_term_list();

/******************************************************************************/

static CsObject *add_object(CsObject *list1, CsObject *list2)
{
    if (list1 == NULL) {
        return(list2);
    }
    else {
        CsObject *tmp = list1;

        while (tmp->next != NULL) {
            tmp = tmp->next;
        }
        tmp->next = list2;

        return(list1);
    }
}

CsObject *CsInstallObjectName(CsObject **objects, char *name)
{
    CsObject *new;

    if ((new = (CsObject *)malloc(sizeof(CsObject))) != NULL) {
        new->name = copy_string(name);
        new->features = NULL;
        new->output_stream = NULL;
        new->highlighted = FALSE;
        new->next = NULL;
        *objects = add_object(*objects, new);
        return(new);
    }
    else {
        fprintf(stderr, "WARNING: Insufficient memory for malloc in CsInstallObjectName()\n");
        return(NULL);
    }
}

TERM *CsInstallObjectFeatureFunctor(CsObject *object, char *functor)
{
    GENERIC_LIST *tmp;

    if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) != NULL) {

        /* Install functor with empty argument list */

        ((TERM *)(tmp->head)) = create_functor(functor);

        /* Add new feature to head of feature list */

        tmp->next = object->features;
        object->features = tmp;

        return(((TERM *)(tmp->head)));
    }
    else {
        fprintf(stderr, "WARNING: Insufficient memory for malloc in CsInstallObjectFeatureFunctor()\n");
        return(NULL);
    }
}

CsObject *CsGetObject(CsObject *object_list, char *object_name)
{
    CsObject *tmp = object_list;

    while ((tmp != NULL) && (strcmp(object_name, tmp->name) != 0)) {
        tmp = tmp->next;
    }
    return(tmp);
}

void CsResetObjectFeatures(CsObject *object)
{
    int f;
    
    for (f=0; f < MaxFunctions; f++) {
        object->activation[f][0] = apply_equations(REST_ACT, 0.0);
    }
    free_term_list(object->features);
    object->features = NULL;
}

void CsNormaliseObjects(CsObject *objects, int cycle)
{
    CsObject *tmp;
    double sum = 0.0;
    int f;

    for (f=0; f < MaxFunctions; f++) {
        for (tmp = objects; tmp != NULL; tmp = tmp->next) {
            sum = sum + tmp->activation[f][cycle];
        }
        for (tmp = objects; tmp != NULL; tmp = tmp->next) {
            tmp->activation[f][cycle] = MIN_ACT + (MAX_ACT-MIN_ACT) * (tmp->activation[f][cycle] / sum);
        }
    }
}


CsObject *CsInitialiseObjects(TaskType task)
{
    extern CsObject *CsInitialiseCoffeePreparationObjects();
    extern CsObject  *ColavitaInitialiseObjects();

    CsObject *tmp = NULL;

    for (tmp = objects; tmp != NULL; tmp = objects) {
        objects = tmp->next;
        free(tmp);
    }

    switch(task) {
        case COFFEE_PREPARATION: case CEREAL_CONSUMPTION: case BREAKFAST: {
            function_name[IMPLEMENT] = "Implement";
            function_name[SOURCE] = "Source";
            function_name[TARGET] = "Target";
            function_name[ATTEND] = NULL;
            function_name[PRESS] = NULL;
            return(CsInitialiseCoffeePreparationObjects());
        }
        case SIMPLE_VISUAL_RT: case SIMPLE_AUDITORY_RT: case CHOICE_RT: case CONFLICT_RT: {
            function_name[IMPLEMENT] = NULL;
            function_name[SOURCE] = NULL;
            function_name[TARGET] = NULL;
            function_name[ATTEND] = "Attend";
            function_name[PRESS] = "Press";
            return(ColavitaInitialiseObjects());
        }
        default: {
            fprintf(stderr, "Error: No suitable object initialisation procedure\n");
            return(NULL);
        }
    }
}

void CsReinitialiseObjects(CsObject *objects, TaskType task)
{
    extern void CsReinitialiseCoffeePreparationObjects(CsObject *objects);
    extern void ColavitaReinitialiseObjects(CsObject *objects);

    switch(task) {
        case COFFEE_PREPARATION: case CEREAL_CONSUMPTION: case BREAKFAST: {
            CsReinitialiseCoffeePreparationObjects(objects);
            break;
        }
        case SIMPLE_VISUAL_RT: case SIMPLE_AUDITORY_RT: case CHOICE_RT: case CONFLICT_RT: {
            ColavitaReinitialiseObjects(objects);
            break;
        }
        default: {
            fprintf(stderr, "Error: No suitable object re-initialisation procedure\n");
            break;
        }
    }
}

/******************************************************************************/

static CsResource *add_resource(CsResource *list1, CsResource *list2)
{
    if (list1 == NULL) {
        return(list2);
    }
    else {
        CsResource *tmp = list1;

        while (tmp->next != NULL) {
            tmp = tmp->next;
        }
        tmp->next = list2;

        return(list1);
    }
}

CsResource *CsInstallResourceName(CsResource **effectors, char *name)
{
    CsResource *new;

    if ((new = (CsResource *)malloc(sizeof(CsResource))) != NULL) {
        new->name = copy_string(name);
        new->features = NULL;
        new->output_stream = NULL;
        new->highlighted = FALSE;
        new->next = NULL;
        *effectors = add_resource(*effectors, new);
        return(new);
    }
    else {
        fprintf(stderr, "WARNING: Insufficient memory for malloc in CsInstallResourceName()\n");
        return(NULL);
    }
}

TERM *CsInstallResourceFeatureFunctor(CsResource *effector, char *functor)
{
    GENERIC_LIST *tmp;

    if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) != NULL) {

        /* Install functor with empty argument list */

        ((TERM *)(tmp->head)) = create_functor(functor);

        /* Add new feature to head of feature list */

        tmp->next = effector->features;
        effector->features = tmp;

        return(((TERM *)(tmp->head)));
    }
    else {
        fprintf(stderr, "WARNING: Insufficient memory for malloc in CsInstallResourceFeatureFunctor()\n");
        return(NULL);
    }
}

CsResource *CsGetResource(CsResource *effector_list, char *effector_name)
{
    CsResource *tmp;

    for (tmp = effector_list; tmp != NULL; tmp = tmp->next) {
        if (strcmp(effector_name, tmp->name) == 0) {
            return(tmp);
        }
    }
    return(NULL);
}

void CsNormaliseResources(CsResource *effectors, int cycle)
{
    CsResource *tmp;
    double sum = 0.0;

    for (tmp = effectors; tmp != NULL; tmp = tmp->next) {
        sum = sum + tmp->activation[cycle];
    }
    for (tmp = effectors; tmp != NULL; tmp = tmp->next) {
        tmp->activation[cycle] = MIN_ACT + (MAX_ACT-MIN_ACT) * (tmp->activation[cycle] / sum);
    }
}

void CsResetResourceFeatures(CsResource *effector)
{
    effector->activation[0] = apply_equations(REST_ACT, 0.0);
    free_term_list(effector->features);
    effector->features = NULL;
}

CsResource *CsInitialiseResources(TaskType task)
{
    extern CsResource *CsInitialiseCoffeePreparationResources();
    extern CsResource *ColavitaInitialiseResources();

    CsResource *tmp;

    for (tmp = resources; tmp != NULL; tmp = resources) {
        resources = tmp->next;
        free(tmp);
    }

    switch(task) {
        case COFFEE_PREPARATION: case CEREAL_CONSUMPTION: case BREAKFAST: {
            return(CsInitialiseCoffeePreparationResources());
        }
        case SIMPLE_VISUAL_RT: case SIMPLE_AUDITORY_RT: case CHOICE_RT: case CONFLICT_RT: {
            return(ColavitaInitialiseResources());
        }
        default: {
            fprintf(stderr, "Error: No suitable resource initialisation procedure\n");
            return(NULL);
        }
    }
}

void CsReinitialiseResources(CsResource *resources, TaskType task)
{
    extern void CsReinitialiseCoffeePreparationResources(CsResource *resources);
    extern void ColavitaReinitialiseResources(CsResource *resources);

    switch(task) {
        case COFFEE_PREPARATION: case CEREAL_CONSUMPTION: case BREAKFAST: {
            CsReinitialiseCoffeePreparationResources(resources);
            break;
        }
        case SIMPLE_VISUAL_RT: case SIMPLE_AUDITORY_RT: case CHOICE_RT: case CONFLICT_RT: {
            ColavitaReinitialiseResources(resources);
            break;
        }
        default: {
            fprintf(stderr, "Error: No suitable resource re-initialisation procedure\n");
            break;
        }
    }
}

/******************************************************************************/
