/*******************************************************************************

Public procedures:
    void CsInitialise()
    int CsRun(FILE *file, TaskType task)
    void CsSetDefaultParameters(TaskType task)
    void CsPrintParameters(FILE *file, TaskType task)

*******************************************************************************/

#include "cs.h"

#define TRIALS 1

/******************************************************************************/
/* Local variables: ***********************************************************/

static int     cycle = -1;
static int     trial_count;
static BOOLEAN DATA_LOGGING = TRUE;
static BOOLEAN TaskComplete = FALSE;

/******************************************************************************/
/* Global variables: **********************************************************/

/* 1: Parameters **************************************************************/

double    SCHEMA_SI;
double    SCHEMA_PE;
double    SCHEMA_SL;
double    OBJECT_SI;
double    OBJECT_PE;
double    OBJECT_SL;
double    EFFECT_SI;
double    EFFECT_PE;
double    EFFECT_SL;

double    MIN_ACT;
double    MAX_ACT;
double    REST_ACT;
double    PERSISTENCE;
MathsType MATHS;
double    SELECTION_THRESHOLD;
double    NOISE_SD;
double    MS_PER_CYCLE;

double    VISUAL_ROUTE_STRENGTH;
double    AUDITORY_ROUTE_STRENGTH;
int       VISUAL_LAG;
int       AUDITORY_LAG;

/* 2. Schema/object/effector lists: *******************************************/

CsSchema   *schemas = NULL;
CsObject   *objects = NULL;
CsResource *resources = NULL;
char *terminating_goal;

/* 3. Message displaying/logging **********************************************/

OUTPUTFLAGS display_message = {FALSE, FALSE, FALSE,  FALSE, FALSE};

extern void CsMessage();

/******************************************************************************/
/* The code: ******************************************************************/
/******************************************************************************/

static char *parameter_file(TaskType task)
{
    switch (task) {
        case COFFEE_PREPARATION: case CEREAL_CONSUMPTION: case BREAKFAST: {
            return("coffee_parameters");
        }
        case SIMPLE_VISUAL_RT: case SIMPLE_AUDITORY_RT: case CHOICE_RT: case CONFLICT_RT: {
            return("colavita_parameters");
        }
        default: {
            return(NULL);
        }
    }
}

/*----------------------------------------------------------------------------*/
/* Read the parameters file: **************************************************/

void CsSetDefaultParameters(TaskType task)
{
    FILE *file = fopen(parameter_file(task), "r");
    
    MIN_ACT = 0.0;
    MAX_ACT = 1.0;
    REST_ACT = 0.1;
    PERSISTENCE = 0.80;
    MATHS = TANH;
    SELECTION_THRESHOLD = 0.60;
    SCHEMA_SI = 0.60; SCHEMA_PE = 0.00; SCHEMA_SL = 0.50;
    OBJECT_SI = 0.60; OBJECT_PE = 0.10; OBJECT_SL = 0.50; 
    EFFECT_SI = 1.00; EFFECT_PE = 1.00; EFFECT_SL = 1.00; 
    NOISE_SD = 0.001;
    MS_PER_CYCLE = 1.0;
    VISUAL_ROUTE_STRENGTH = 1.00;
    AUDITORY_ROUTE_STRENGTH = 0.75;
    AUDITORY_LAG = 20;
    VISUAL_LAG = 60;

    if (file != NULL) {
        CsMessage(WARNING, 0, "Reading default parameters...\n");
        
        fscanf(file, "%lf %lf %lf", &SCHEMA_SI, &SCHEMA_PE, &SCHEMA_SL);
        fscanf(file, "%lf %lf %lf", &OBJECT_SI, &OBJECT_PE, &OBJECT_SL);
        fscanf(file, "%lf %lf %lf", &EFFECT_SI, &EFFECT_PE, &EFFECT_SL);
        fscanf(file, "%lf %lf %d %lf %d", &MS_PER_CYCLE, &VISUAL_ROUTE_STRENGTH, &VISUAL_LAG, &AUDITORY_ROUTE_STRENGTH, &AUDITORY_LAG);
        fscanf(file, "%d %d %d %d %d", &display_message.action, &display_message.selection, &display_message.world, &display_message.warning, &display_message.activations);

        fclose(file);
    }
    else {
        CsMessage(WARNING, 0, "Cannot open parameters file: using built-in defaults\n");
    }
}

void CsSaveDefaultParameters(TaskType task)
{
    FILE *file = fopen(parameter_file(task), "w");

    if (file != NULL) {
        CsMessage(WARNING, 0, "Writing default parameters...\n");
        
        fprintf(file, "%5.2f %5.2f %5.2f\n", SCHEMA_SI, SCHEMA_PE, SCHEMA_SL);
        fprintf(file, "%5.2f %5.2f %5.2f\n", OBJECT_SI, OBJECT_PE, OBJECT_SL);
        fprintf(file, "%5.2f %5.2f %5.2f\n", EFFECT_SI, EFFECT_PE, EFFECT_SL);
        fprintf(file, "%5.2f %5.2f %3d %5.2f %3d\n", MS_PER_CYCLE, VISUAL_ROUTE_STRENGTH, VISUAL_LAG, AUDITORY_ROUTE_STRENGTH, AUDITORY_LAG);
        fprintf(file, "%2d %2d %2d %2d %2d\n", display_message.action, display_message.selection, display_message.world, display_message.warning, display_message.activations);

        fclose(file);
    }
    else {
        CsMessage(WARNING, 0, "Cannot open parameters file: save failed\n");
    }
}

/*----------------------------------------------------------------------------*/
/* Print the current settings of all parameters (to the given file): **********/

char *task_name(TaskType m)
{
    switch (m) {
        case COFFEE_PREPARATION: { return("Coffee Preparation"); }
        case CEREAL_CONSUMPTION: { return("Cereal Consumption"); }
        case BREAKFAST: { return("Breakfast"); }
        case SIMPLE_VISUAL_RT: { return("Simple Visual RT"); }
        case SIMPLE_AUDITORY_RT: { return("Simple Auditory RT"); }
        case CHOICE_RT: { return("Choice RT"); }
        case CONFLICT_RT: { return("Conflict RT"); }
    }
    return(NULL);
}

void CsPrintParameters(FILE *file, TaskType task)
{
    extern char *random_seed();

    fprintf(file, "Schema Excitation = %4.2f * SYSTEM + %4.2f * INPUT\n", SCHEMA_SI, (1-SCHEMA_SI));
    fprintf(file, "  where SYSTEM = %4.2f * self activation + %4.2f * lateral inhibition\n", SCHEMA_SL, (1-SCHEMA_SL));
    fprintf(file, "   and   INPUT = %4.2f * parent + %4.2f * environment\n", SCHEMA_PE, (1-SCHEMA_PE));
    fprintf(file, "Object Excitation = %4.2f * SYSTEM + %4.2f * INPUT\n", OBJECT_SI, (1-OBJECT_SI));
    fprintf(file, "  where SYSTEM = %4.2f * self activation + %4.2f * lateral inhibition\n", OBJECT_SL, (1-OBJECT_SL));
    fprintf(file, "   and   INPUT = %4.2f * parent + %4.2f * environment\n", OBJECT_PE, (1-OBJECT_PE));
    fprintf(file, "Resource Excitation = %4.2f * SYSTEM + %4.2f * INPUT\n", EFFECT_SI, (1-EFFECT_SI));
    fprintf(file, "  where SYSTEM = %4.2f * self activation + %4.2f * lateral inhibition\n", EFFECT_SL, (1-EFFECT_SL));
    fprintf(file, "   and   INPUT = %4.2f * parent + %4.2f * environment\n", EFFECT_PE, (1-EFFECT_PE));
    fprintf(file, "Global activation parameters:\n");
    fprintf(file, "   Rest: %4.2f; Selection Threshold: %4.2f; Persistence: %4.2f\n", REST_ACT, SELECTION_THRESHOLD, PERSISTENCE);
    fprintf(file, "   Noise SD: %e; Seed: %s\n", NOISE_SD, random_seed());
    fprintf(file, "Task: %s\n", task_name(task));
    fprintf(file, "Timing parameters: %4.2f MS/Cycle\n", MS_PER_CYCLE);
    fprintf(file, "  Visual Lag:   %3d; Visual Route Strength:   %4.2f\n", VISUAL_LAG, VISUAL_ROUTE_STRENGTH);
    fprintf(file, "  Auditory Lag: %3d; Auditory Route Strength: %4.2f\n", AUDITORY_LAG, AUDITORY_ROUTE_STRENGTH);
}

/*----------------------------------------------------------------------------*/
/* Calculate (and record) all activations for the current cycle: **************/

static void CsStepActivations(int cycle)
{
    extern double CsCalculateSchemaActivation(int cycle, CsSchema *s);
    extern void   CsNormaliseSchemas(CsSchema *list, int cycle);
    extern double CsCalculateObjectActivation(int cycle, CsObject *s, int f);
    extern void   CsNormaliseObjects(CsObject *list, int cycle);
    extern double CsCalculateResourceActivation(int cycle, CsResource *s);
    extern void   CsNormaliseResources(CsResource *list, int cycle);
    extern void   CsSaveActivations();

    CsSchema *tmp1;
    CsObject *tmp2;
    CsResource *tmp3;

    for (tmp1 = schemas; tmp1 != NULL; tmp1 = tmp1->next) {
        tmp1->activation[cycle] = CsCalculateSchemaActivation(cycle, tmp1);
    }

    if (MATHS == NORMALISED) {
        CsNormaliseSchemas(schemas, cycle);
    }

    for (tmp2 = objects; tmp2 != NULL; tmp2 = tmp2->next) {
        int f;
        for (f=0; f < MaxFunctions; f++) {
            tmp2->activation[f][cycle] = CsCalculateObjectActivation(cycle, tmp2, f);
        }
    }

    if (MATHS == NORMALISED) {
        CsNormaliseObjects(objects, cycle);
    }

    for (tmp3 = resources; (tmp3 != NULL); tmp3 = tmp3->next) {
        tmp3->activation[cycle] = CsCalculateResourceActivation(cycle, tmp3);
    }

    if (MATHS == NORMALISED) {
        CsNormaliseResources(resources, cycle);
    }

    if (DATA_LOGGING) {
        CsSaveActivations(cycle);
    }
}

/*----------------------------------------------------------------------------*/
/* Set up dynamic data structures (schema, object and effector lists): ********/

void CsInitialise(TaskType task)
{
    extern CsSchema   *CsInitialiseSchemas(TaskType task);
    extern CsObject   *CsInitialiseObjects(TaskType task);
    extern CsResource *CsInitialiseResources(TaskType task);
    extern char       *CsTerminatingGoal(TaskType task);

    schemas = CsInitialiseSchemas(task);
    objects = CsInitialiseObjects(task);
    resources = CsInitialiseResources(task);
    terminating_goal = CsTerminatingGoal(task);
}

/*----------------------------------------------------------------------------*/
/* Reset all variables and files to their initial values or states: ***********/

void CsReinitialise(TaskType task)
{
    extern void    CsReinitialiseSchemas(CsSchema *schemas, TaskType task);
    extern void    CsReinitialiseObjects(CsObject *objects, TaskType task);
    extern void    CsReinitialiseResources(CsResource *effectors, TaskType task);
    extern void    CsInitialiseStats();
    extern FILE   *CsOpenLogFile();
    extern BOOLEAN CsOpenActivationFiles();
    extern void    CsPostCycleTrace(int c, FILE *f);
    extern void    CsSaveActivations(int c);
    extern void    ActInitialiseList();

    cycle = 0;
    trial_count = 0;
    TaskComplete = FALSE;

    ActInitialiseList();
    CsInitialiseStats();
    CsReinitialiseSchemas(schemas, task);
    CsReinitialiseObjects(objects, task);
    CsReinitialiseResources(resources, task);

    if (display_message.activations) {
        CsPostCycleTrace(cycle, stdout);
    }

    if (DATA_LOGGING) {
        if (CsOpenActivationFiles()) {
            CsSaveActivations(cycle);
        }
        else {
            DATA_LOGGING = FALSE;
        }
    }
}

/*----------------------------------------------------------------------------*/
/* Close all files (generally on completion of a run): ************************/

static void CsCloseLogFiles()
{
    extern void CsCloseActivationFiles();

    if (DATA_LOGGING) {
        CsCloseActivationFiles();
    }
}

/*----------------------------------------------------------------------------*/
/* Run one cycle, logging progress if necessary: ******************************/

static void CsRunOneCycle(FILE *file, TaskType task)
{
    extern void ActRecord(ActionType act, int cycle, char *arg1, char *arg2);
    extern void CsStepSelections();
    extern void CsPostCycleTrace();
    extern BOOLEAN external_world(int cycle, TaskType task);
    extern void AttachData(int cycle, BOOLEAN redraw);

    cycle++;
    CsStepActivations(cycle);
    CsStepSelections(cycle);
    if (display_message.warning) {
        fprintf(stdout, "."); fflush(stdout);
    }
    AttachData(cycle, FALSE);
    if (external_world(cycle, task)) { /* TRUE on trial completion */
        if (display_message.warning) {
            fprintf(stdout, "*"); fflush(stdout);
        }
        if (!TaskComplete) {
            ActRecord(NONE, cycle, NULL, NULL);
            TaskComplete = TRUE;
        }
        trial_count++;
    }
    if ((display_message.activations)) {
        CsPostCycleTrace(cycle, file);
    }
}

/*----------------------------------------------------------------------------*/
/* Run exactly n cycles: ******************************************************/

static void CsMultiRun(int n, FILE *file, TaskType task)
{
    while (n-- > 0) {
        CsRunOneCycle(file, task);
    }
}

/*----------------------------------------------------------------------------*/
/* Return TRUE if more cycles should be executed so as to complete the task: **/

static BOOLEAN CsTaskIsIncomplete(FILE *file, TaskType task, int trials)
{
    if (cycle >= (MaxCycles - 1)) {
        CsMessage(WARNING, cycle, "\nToo many cycles: terminating immediately.\n");
        return(FALSE);
    }
    else if (trial_count > trials) {
        char buffer[80];
        sprintf(buffer, "%d trial(s) complete.", trials);
        CsMessage(WARNING, cycle, buffer);
        CsMessage(WARNING, cycle, "Running 10 more cycles before halting.");
        CsMultiRun((cycle > (MaxCycles - 11)) ? (MaxCycles - cycle - 1) : 10, file, task);
        return(FALSE);
    }
    else {
        return(TRUE);
    }
}

/*----------------------------------------------------------------------------*/
/* Reinitialise and run the task until completion: ****************************/

int CsRun(FILE *file, TaskType task)
{
    CsReinitialise(task);
    do {
        CsRunOneCycle(file, task);
    } while (CsTaskIsIncomplete(file, task, TRIALS));
    CsCloseLogFiles();
    return(cycle);
}


int CsStep(FILE *file, TaskType task, int n)
{
    if (cycle < 0) {
        CsReinitialise(task);
    }
    while ((n-- > 0) && (CsTaskIsIncomplete(file, task, TRIALS))) {
        CsRunOneCycle(file, task);
    } 
    return(cycle);
}
/******** END OF FILE *********************************************************/
