/*******************************************************************************
This file contains:
    functions for making a dynamic schema network from the definitions.

*******************************************************************************/

#include "cs.h"

extern char *copy_string(char *string);

/******************************************************************************/

static CsSchema *add_schema(CsSchema *list1, CsSchema *list2)
{
    if (list1 == NULL) {
        return(list2);
    }
    else {
        CsSchema *tmp = list1;

        while (tmp->next != NULL) {
            tmp = tmp->next;
        }
        tmp->next = list2;

        return(list1);
    }
}

CsSchema *CsInstallSchemaNode(CsSchema **schemas, char *schema_name, char *schema_goal, int level, ActionType action)
{
    extern GENERIC_LIST *append();

    CsSchema *new;

    if ((new = (CsSchema *)malloc(sizeof(CsSchema))) != NULL) {
        new->name = copy_string(schema_name);
        new->goal = copy_string(schema_goal);
        new->level = level;
        new->res = NULL;
        new->children = NULL;
        new->corresponding_action = action;
        new->next = NULL;
        *schemas = add_schema(*schemas, new);
        return(new);
    }
    else {
        fprintf(stderr, "WARNING: Memory allocation failed in CsInstallSchemaNode()\n");
        return(NULL);
    }
}

void CsInstallSchemaTriggeringFunctions(CsSchema *schema, BOOLEAN (*obj)(), BOOLEAN (*res)(), double (*sit)())
{
    schema->object_triggering_function = obj;
    schema->effector_triggering_function = res;
    schema->triggering_situation = sit;
}

void CsInstallSchemaSubgoal(CsSchema *schema, char *subgoal, ArgumentFunction arg_function, GoalType sort)
{
    Subgoal *tmp;

    if ((tmp = (Subgoal *)malloc(sizeof(Subgoal ))) == NULL) {
        fprintf(stderr, "WARNING: Insufficient memory for malloc in CsInstallSchemaSubgoal()\n");
    }
    else {
        tmp->subgoal = copy_string(subgoal);
        tmp->arg_function = arg_function;
        tmp->prior_subgoals = NULL;

        /* Add new subgoal to head of subgoal list */

        tmp->next = schema->children;
        schema->children = tmp;
    }
}

void CsInstallOrderingConstraint(CsSchema *schema, char *subgoal1, char *subgoal2)
{
    Subgoal *tmp1 = schema->children;
    Subgoal *tmp2 = schema->children;
    GENERIC_LIST *tmp;

    while (strcmp(tmp1->subgoal, subgoal1) != 0) {
        tmp1 = tmp1->next;
    }

    while (strcmp(tmp2->subgoal, subgoal2) != 0) {
        tmp2 = tmp2->next;
    }

    if ((tmp1 == NULL) || (tmp2 == NULL)) {
        fprintf(stderr, "Cannot find subgoals in CsInstallOrderingConstraint");
    }
    else if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) == NULL) {
        fprintf(stderr, "WARNING: Insufficient memory for malloc in CsInstallOrderingConstraint)(\n");
    }
    else {
        tmp->next = tmp2->prior_subgoals;
        tmp->head = tmp1;
        tmp2->prior_subgoals = tmp;
    }
}

void CsInstallSchemaResource(CsSchema *schema, Resource resource_id)
{
    GENERIC_LIST *tmp;

    if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) == NULL) {
        fprintf(stderr, "WARNING: Insufficient memory for malloc in CsInstallSchemaResource()\n");
    }
    else {

        ((Resource) (tmp->head)) = resource_id;

        /* Add new resource to head of resource list */

        tmp->next = schema->res;
        schema->res = tmp;
    }
}

/******************************************************************************/

static double CsInitialSchemaActivation()
{
    extern double apply_equations(double previous, double input);

    return( apply_equations(REST_ACT, 0.0) );
}

void CsNormaliseSchemas(CsSchema *schema_list, int cycle)
{
}

GENERIC_LIST *CsInitialiseSchemas(TaskType task)
{
    extern GENERIC_LIST *CsInitialiseCoffeePreparationSchemas();
    extern GENERIC_LIST *CsInitialiseSimpleAuditoryRTSchemas();
    extern GENERIC_LIST *CsInitialiseSimpleVisualRTSchemas();
    extern GENERIC_LIST *CsInitialiseChoiceRTSchemas();

    CsSchema *tmp = NULL;

    for (tmp = schemas; tmp != NULL; tmp = schemas) {
        schemas = tmp->next;
        free(tmp);
    }

    switch(task) {
        case COFFEE_PREPARATION: case CEREAL_CONSUMPTION: case BREAKFAST: {
            return(CsInitialiseCoffeePreparationSchemas());
        }
        case SIMPLE_VISUAL_RT: {
            return(CsInitialiseSimpleVisualRTSchemas());
        }
        case SIMPLE_AUDITORY_RT: {
            return(CsInitialiseSimpleAuditoryRTSchemas());
        }
        case CHOICE_RT: {
            return(CsInitialiseChoiceRTSchemas());
        }
        case CONFLICT_RT: {
            return(CsInitialiseChoiceRTSchemas());
        }
        default: {
            fprintf(stderr, "Error: No suitable schema initialisation procedure\n");
            return(NULL);
        }
    }
}

void CsReinitialiseSchemas(CsSchema *schema_list)
{
    CsSchema *tmp;

    for (tmp = schema_list; tmp != NULL; tmp = tmp->next) {
        tmp->activation[0] = CsInitialSchemaActivation();
        tmp->selected = FALSE;
        tmp->executed = FALSE;
    }
    if (MATHS == NORMALISED) {
        CsNormaliseSchemas(schema_list, 0);
    }
}

/******************************************************************************/

char *CsTerminatingGoal(TaskType task)
{
    switch(task) {
        case COFFEE_PREPARATION: {
            return("prepare_coffee");
        }
        case CEREAL_CONSUMPTION: { 
            return("consume_cereal");
        }
        case BREAKFAST: { 
            return("eat_breakfast");
        }
        case SIMPLE_VISUAL_RT: {
            return("Decide");
        }
        case SIMPLE_AUDITORY_RT: {
            return("Decide");
        }
        case CHOICE_RT: {
            return("Decide");
        }
        case CONFLICT_RT: {
            return("Decide");
        }
        default: {
            fprintf(stderr, "Error: No definition of terminating goal\n");
            return(NULL);
        }
    }
}

/******************************************************************************/
