#include "cs.h"

extern BOOLEAN competitor_of();
extern CsSchema *selected_parent_of();
extern CsSchema *is_parent_of();
extern BOOLEAN intended_goal();
extern void    CsSelectionMessage();

/******************************************************************************/

static BOOLEAN winning_schema(int cycle, CsSchema *schema)
{
    CsSchema *tmp = schemas;

    /* A schema is the winner if no competitor has greater activation */

    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        if (competitor_of(schema, tmp) && (tmp->activation[cycle] > (schema->activation[cycle]))) {
            return(FALSE);
        }
    }
    return(TRUE);
}

/******************************************************************************/

/*
 * An unselected schema whose activation is above threshold and which is the
 * winner amongst its competitors should be selected:
 *
 * schema_should_be_selected(Schema, Subgoals) :-
 *     schema_status(Schema, unselected, _),
 *     activation_above_threshold(Schema),
 *     winning_schema(Schema), 
 *     schema_children(Schema, Children),
 *     construct_unachieved_list(Children, Subgoals).
 */

static BOOLEAN schema_should_be_selected(int cycle, CsSchema *schema)
{
    return( (!(schema->selected)) &&
            (schema->activation[cycle-1] > SELECTION_THRESHOLD) &&
            (winning_schema(cycle-1, schema)) );
}

/******************************************************************************/

/*
 * There are three cases for deselection:
 *   1) A selected schema that is no longer the winner amongst its competitors
 *      should be deselected;
 *   2) A winning schema should be deselected if its parent is deselected; and
 *   3) A winning schema should be deselected if its parent is selected.
 * The second case solves a synchronisation problem that arose when the
 * activation of an achieved parent fell before its last child could be
 * deselected.
 */

static BOOLEAN schema_should_be_deselected(int cycle, CsSchema *schema)
{
    CsSchema *parent;

    if (schema->selected) {
        if (!winning_schema(cycle-1, schema)) {
            return(TRUE);
        }
        else if ((parent = selected_parent_of(schema)) != NULL) {
            return(schema_should_be_deselected(cycle, parent));
        }
        else if (intended_goal(schema->goal)) {
            return(FALSE);
        }
        else if ((parent = is_parent_of(schema)) != NULL) {
            return(schema_should_be_selected(cycle, parent));
        }
        else {
            return(FALSE);
        }
    }
    else {
        return(FALSE);
    }
}

static void set_unachieved_subgoals(CsSchema *schema)
{
    Subgoal *tmp;

    for (tmp = schema->children; tmp != NULL; tmp = tmp->next) {
        tmp->achieved = FALSE;
    }
}

/******************************************************************************/

/* Forward declaration: */

BOOLEAN goal_is_achieved_on_current_cycle(char *goal);

static BOOLEAN all_subgoals_will_be_achieved(CsSchema *schema)
{
    Subgoal *tmp;

    for (tmp = schema->children; tmp != NULL; tmp = tmp->next) {
        if (!(tmp->achieved) && !(goal_is_achieved_on_current_cycle(tmp->subgoal))) {
            return(FALSE);
        }
    }
    return(TRUE);
}

BOOLEAN goal_is_achieved_on_current_cycle(char *goal)
{
    extern BOOLEAN continuous_goal(char *goal);
    BOOLEAN found = FALSE;
    CsSchema *tmp = schemas;

    if (continuous_goal(goal)) {
        return(FALSE);
    }
    else {
        while (!(found) && (tmp != NULL)) {
            if ((tmp->selected) && (strcmp(tmp->goal, goal) == 0)) {
                found = all_subgoals_will_be_achieved(tmp);
            }
            tmp = tmp->next;
        }
        return(found);
    }
}

/******************************************************************************/

/* When a schema is deselected, make sure that its parent thinks it is        */
/* achieved. Under normal circumstances it will already be marked as such.    */
/* Under faulty operation the schema may be deselected before its goal has    */
/* been achieved. This should lead to genuine omissions at all levels.        */
/* Added by RPC on Thu Aug 13 13:54:45 1998                                   */

static void mark_schema_as_achieved(CsSchema *schema)
{
    CsSchema  *parent;
    Subgoal *tmp;

    if ((parent = selected_parent_of(schema)) != NULL) {
        for (tmp = parent->children; tmp != NULL; tmp = tmp->next) {
            if (strcmp(tmp->subgoal, schema->goal) == 0) {
                tmp->achieved = TRUE;
            }
        }
    }
}

/******************************************************************************/

/*
 * cycle:
 *     schema_status(Schema, unselected, []),
 *     schema_should_be_selected(Schema, UnachievedSubgoals),
 *     tab_level(Schema, Tab)
 *  => retract(schema_status(Schema, unselected, [])),
 *     wassert(schema_status(Schema, selected, UnachievedSubgoals)),
 *     message(selection, "~*c+~w~n", [Tab,32,Schema]).
 * 
 * cycle:
 *     schema_status(Schema, selected, Subgoals),
 *     schema_should_be_deselected(Schema),
 *     tab_level(Schema, Tab)
 *  => retract(schema_status(Schema, selected, Subgoals)),
 *     wassert(schema_status(Schema, unselected, [])),
 *     message(selection, "~*c-~w~n", [Tab,32,Schema]).
 *
 * cycle:
 *     schema_status(Schema, selected, BeforeSubgoals),
 *     not( schema_should_be_deselected(Schema) ),
 *     subgoals_achieved(BeforeSubgoals, AfterSubgoals)
 *  => retract(schema_status(Schema, selected, BeforeSubgoals)),
 *     wassert(schema_status(Schema, selected, AfterSubgoals)).
 */

GENERIC_LIST *update_subgoal_queue = NULL;
GENERIC_LIST *selection_queue = NULL;
GENERIC_LIST *deselection_queue = NULL;
GENERIC_LIST *execution_queue = NULL;

static void queue_for_update_achieved_subgoals(CsSchema *schema)
{
    GENERIC_LIST *tmp;

    if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) == NULL) {
        fprintf(stderr, "WARNING: Insufficient memory for malloc");
    }
    else {
        tmp->head = (void *)schema;
        tmp->next = update_subgoal_queue;
        update_subgoal_queue = tmp;
    }
}

static void queue_for_deselection(CsSchema *schema)
{
    GENERIC_LIST *tmp;

    if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) == NULL) {
        fprintf(stderr, "WARNING: Insufficient memory for malloc");
    }
    else {
        tmp->head = (void *)schema;
        tmp->next = deselection_queue;
        deselection_queue = tmp;
    }
}

static void queue_for_selection(CsSchema *schema)
{
    GENERIC_LIST *tmp;

    if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) == NULL) {
        fprintf(stderr, "WARNING: Insufficient memory for malloc");
    }
    else {
        tmp->head = (void *)schema;
        tmp->next = selection_queue;
        selection_queue = tmp;
    }
}

static void queue_for_execution(GENERIC_LIST *list_element)
{
    list_element->next = execution_queue;
    execution_queue = list_element;
}

void adjust_selection(int cycle, CsSchema *schema, BOOLEAN value)
{
    schema->selected = value;
    schema->executed = FALSE;
    set_unachieved_subgoals(schema);
    CsSelectionMessage(cycle, schema, value);
}

static void process_queues(int cycle)
{
    GENERIC_LIST *tmp;

    while (update_subgoal_queue != NULL) {
        tmp = update_subgoal_queue;
        update_subgoal_queue = update_subgoal_queue->next;
	free(tmp);
    }
        
    while (selection_queue != NULL) {
        tmp = selection_queue;
        adjust_selection(cycle, ((CsSchema *)(selection_queue->head)), TRUE);
        selection_queue = selection_queue->next;
        if (((CsSchema *)(tmp->head))->corresponding_action != NONE) {
            queue_for_execution(tmp);
        }
        else {
            free(tmp);
        }
    }

    while (deselection_queue != NULL) {
        tmp = deselection_queue;
        adjust_selection(cycle, ((CsSchema *)(deselection_queue->head)), FALSE);
        mark_schema_as_achieved((CsSchema *)(deselection_queue->head));
        deselection_queue = deselection_queue->next;
	free(tmp);
    }
}

/******************************************************************************/

static GENERIC_LIST *execute_actions(GENERIC_LIST *execution_queue, int cycle)
{
    extern GENERIC_LIST *execute_action();

    while (execution_queue != NULL) {
        GENERIC_LIST *tmp = execution_queue;
        execute_action(((CsSchema *)(tmp->head)), cycle);
        ((CsSchema *)(tmp->head))->executed = TRUE;
        execution_queue = tmp->next;
        free(tmp);
    }
    return(execution_queue);
}

void CsStepSelections(int cycle)
{
    CsSchema *tmp;

    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        if (tmp->selected) {
            if (schema_should_be_deselected(cycle, tmp)) {
                queue_for_deselection(tmp);
            }
            else {
                queue_for_update_achieved_subgoals(tmp);
            }
        }
        else if (!(tmp->selected) && (schema_should_be_selected(cycle, tmp))) {
            queue_for_selection(tmp);
        }
    }
    process_queues(cycle);
    execution_queue = execute_actions(execution_queue, cycle);
}

/******************************************************************************/
