/*******************************************************************************

    FILE: term.c
    AUTHOR: R. Cooper

This file contains a number of procedures for manipulating Prolog-like terms
encoded in C. The C data structure used to encode a Prolog terms is defined in
cs.h as:

    typedef struct Term {
        TermType sort;
        char functor[20];
        GENERIC_LIST *arguments;
    } TERM;

The first field of a structure of this sort is a scalar value which indicates
the type of the term (whether it represents a GROUND term, a VARIABLE, or a
LIST). The second field is a character array in which the functor of a ground
term is stored. (So if the term is on(cat, mat), the second field contains the
character string  "on".) The third field contains a list representation of the
arguments. The arguments will all be TERMs, so this is just a list of TERMs.
Because this program uses lots of linked lists of various types, we use a
single generic data structure to code lists, and use type casting when
refering to the individual elements.

A GENERIC_LIST is a data structure with two fields:

    typedef struct GenericList {
        void *head;
        struct GenericList *next;
    } GENERIC_LIST;

The head is a pointer to void. A linked list is encoded as a GENERIC_LIST by
type casting this field to the appropriate sort of list element. Thus, in a
GENERIC_LIST of TERM the head field is type cast to TERM. The second field of
a GENERIC_LIST is just a pointer to the tail of the list (which is NULL if the
list has no more elements, and a pointer to another GENERIC_LIST otherwise).

The public procedures in this file are:
    void free_term_list(GENERIC_LIST *terms):
        Given a GENERIC_LIST of terms, free all space associated with
        those terms (thus discarding the list).
    void write_term_list(FILE *file, GENERIC_LIST *terms)
        Given a GENERIC_LIST of terms, print that list as if it were
        Prolog in the file pointed to by its first argument.
    GENERIC_LIST *append(GENERIC_LIST *list1, GENERIC_LIST *list2)
        Given two lists, append the second to the end of the first
        (and return a pointer to the result).
    int length(GENERIC_LIST *l)
        Given a generic list, return the number of elements in that list.
    TERM *create_functor(char *functor)
        Create (and return) a ground TERM structure with the specified 
        functor (and no arguments).
    void install_argument_term(TERM *term, TERM *argument)
        Add an argument to the specified term. This predicate simply
        extends a Prolog-like term by appending a further argument
        to its argument list (effectively increasing the arity of
        the Prolog-like term by one.).
    TERM *install_argument_functor(TERM *term, char *functor)
        Create a gound term with the specified functor, and add it as
        the next argument of the specified term. Return a pointer to the
        argument.
    void install_list_argument(TERM *term, int n, ...)
        Given a term, an integer n, and a sequence of n character strings,
        create a Prolog-like list consisting of Prolog terms with those
        character strings as their functors, and add that list as an
        argument of the specified term.

None of these predicates should need to be changed.

*******************************************************************************/

#include "cs.h"
#include <stdarg.h>

/******************************************************************************/

char *copy_string(char *s)
{
    if (s == NULL) {
        return(NULL);
    }
    else {
        int l = strlen(s);
        char *copy = (char *)malloc((l+1)*sizeof(char));

        if (copy != NULL) {
            int j;
            for (j = 0; j < l; j++) {
                copy[j] = s[j];
            }
            copy[l] = '\0';
        }
        return(copy);
    }
}

/******************************************************************************/

static void free_term(TERM *term)
{
    void free_term_list();

    if (term->sort == LIST) {
        GENERIC_LIST *p, *q;
        for (p = term->arguments; p != NULL; p = q) {
            q = p->next;
            free(p->head);
        }
    }
    else {
        free_term_list(term->arguments);
    }
    free(term);
}

void free_term_list(GENERIC_LIST *terms)
{
    while (terms != NULL) {
        GENERIC_LIST *tmp = terms->next;
        free_term((TERM *)(terms->head));
        free(terms);
        terms = tmp;
    }
}

/******************************************************************************/

static void write_atom_list(FILE *fp, GENERIC_LIST *atoms)
{
    GENERIC_LIST *tmp;

    fprintf(fp, "[");
    for (tmp = atoms; (tmp != NULL); tmp = tmp->next) {
        fprintf(fp, "%s", ((char *)(tmp->head)));
        if (tmp->next != NULL) {
            fprintf(fp, ",");
        }
    }
    fprintf(fp, "]");
}

void write_term_list(FILE *fp, GENERIC_LIST *terms)
{
    GENERIC_LIST *tmp;

    for (tmp = terms; (tmp != NULL); tmp = tmp->next) {
        switch (((TERM *)(tmp->head))->sort) {
            case LIST: {
                write_atom_list(fp, ((TERM *)(tmp->head))->arguments);
                break;
            }
            case VARIABLE: {
                fprintf(fp, "_%p", tmp->head);
                break;
            }
            case GROUND: {
                fprintf(fp, "%s", ((TERM *)(tmp->head))->functor);
                if (((TERM *)(tmp->head))->arguments != NULL) {
                    fprintf(fp, "(");
                    write_term_list(fp, ((TERM *)(tmp->head))->arguments);
                    fprintf(fp, ")");
                }
                break;
            }
            default: {
                fprintf(fp, "WARNING: Unknown term type: %d\n", ((TERM *)(tmp->head))->sort);
                break;
            }
        }
        if (tmp->next != NULL) {
            fprintf(fp, ",");
        }
    }
}

/******************************************************************************/

GENERIC_LIST *append(GENERIC_LIST *list1, GENERIC_LIST *list2)
{
    if (list1 == NULL) {
        return(list2);
    }
    else {
        GENERIC_LIST *tmp = list1;

        while (tmp->next != NULL) {
            tmp = tmp->next;
        }
        tmp->next = list2;

        return(list1);
    }
}

/******************************************************************************/

int length(GENERIC_LIST *l)
{
    int count = 0;

    while (l != NULL) {
        count++;
        l = l->next;
    }
    return(count);
}

/******************************************************************************/

TERM *create_functor(char *functor)
{
    TERM *tmp = (TERM *)malloc(sizeof(TERM));

    if (tmp == NULL) {
        fprintf(stderr, "WARNING: Memory allocation failed in create_functor\n");
    }
    else {                        /* Install functor with empty argument list */
        tmp->sort = GROUND;
        tmp->functor = copy_string(functor);
        tmp->arguments = NULL;
    }
    return(tmp);                       /* Return a pointer to the new functor */
}

/******************************************************************************/

void install_argument_term(TERM *term, TERM *argument)
{
    GENERIC_LIST *tmp;

   if (term == NULL) {
        fprintf(stderr, "WARNING: NULL input term in install_argument_term\n");
    }
    else if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) == NULL) {
        fprintf(stderr, "WARNING: Memory allocation failure in install_argument_term\n");
    }
    else {
        tmp->head = (void *)argument;
        tmp->next = NULL;
        term->arguments = append(term->arguments, tmp);
    }
}

/******************************************************************************/

TERM *install_argument_functor(TERM *term, char *functor)
{
    TERM *tmp = create_functor(functor);

    install_argument_term(term, tmp);
    return(tmp);
}

/******************************************************************************/

static TERM *install_empty_list_argument(TERM *term)
{
    GENERIC_LIST *tmp;

    if (term == NULL) {
        fprintf(stderr, "WARNING: NULL term argument in install_empty_list_argument\n");
    }
    else if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) == NULL) {
        fprintf(stderr, "WARNING: Memory allocation failure in install_empty_list_argument\n");
    }
    else if ((tmp->head = (TERM *)malloc(sizeof(TERM))) == NULL) {
        fprintf(stderr, "WARNING: Memory allocation failure in install_empty_list_argument\n");
        free(tmp);
    }
    else {                    /* Install argument with empty subargument list */
        ((TERM *)(tmp->head))->functor = copy_string("");
        ((TERM *)(tmp->head))->sort = LIST;
        ((TERM *)(tmp->head))->arguments = NULL;
        tmp->next = NULL;
        term->arguments = append(term->arguments, tmp);
        return(((TERM *)(tmp->head)));
    }
    return(NULL);
}

static void install_list_argument_element(TERM *list, char *element)
{
    GENERIC_LIST *tmp;

    if ((tmp = (GENERIC_LIST *)malloc(sizeof(GENERIC_LIST))) == NULL) {
        fprintf(stderr, "WARNING: Memory allocation failure in install_list_argument_element\n");
    }
    else {       /* Add the new list element as the first element of the list */
        tmp->head = copy_string(element);
        tmp->next = list->arguments;
        list->arguments = tmp;
    }
}

void install_list_argument(TERM *term, int n, ...)
{
    va_list ap;
    TERM *list;
    char *name;

    va_start(ap, n);
    list = install_empty_list_argument(term);
    while (n-- > 0) {
        name = va_arg(ap, char *);
        install_list_argument_element(list, name);
    }
    va_end(ap);
}

/******************************** END OF FILE *********************************/
