/*
 *      xcs.c: CS with X plotting tool using GTK+ widgets
 */

#include "cs.h"

#include <gtk/gtk.h>

#include <math.h> 

#define  PLOTTER_HEIGHT 150

typedef enum {LINEAR_SLIDER, REAL_SLIDER, LOG_SLIDER, INTEGER_SLIDER, MATHS_SELECTOR, BOOLEAN_SELECTOR} ParameterType;

typedef struct parameter_list {
    GtkWidget *slider_widget;
    GtkWidget *value_widget;
    ParameterType type;
    void *value;
    struct parameter_list *next;
} ParameterList;

ParameterList *parameters = NULL;
 
static GtkWidget *plotter[3];
static GtkWidget *cycle_count_widget;
static TaskType current_task = COFFEE_PREPARATION;

static GdkFont *font_info[2];
/* font_info[0] = 10 point (small) */
/* font_info[1] = 12 point (normal) */

static GdkGC *gc[9];
/* gc[0,1,2] are black, with decreasing saturation */
/* gc[3,4,5] are blue,  with decreasing saturation */
/* gc[6,7,8] are red,   with decreasing saturation */

/* The limits of the graph within its canvas */
#define PLOT_WORLD_X_MIN 30
#define PLOT_WORLD_X_MAX -7
#define PLOT_WORLD_Y_MIN 5
#define PLOT_WORLD_Y_MAX -40

extern int  CsRun(FILE *file, TaskType task);
extern int  CsStep(FILE *file, TaskType task, int n);
extern void CsMessage(MessageType mt, int cycle, char *message);
extern void CsReinitialise(TaskType task);
extern void CsInitialise(TaskType task);
extern void CsSetDefaultParameters(TaskType task);
extern void CsSaveDefaultParameters(TaskType task);

extern char *task_name(TaskType m);

#define SELECTED_OBJ_FUNCTION   102938
#define PLOTTER_LABEL           395723

/*
 *	X stuff: callbacks, buttons, plotters, etc.
 */

static void create_gcs(GtkWidget *widget)
{
    static char dashed[2] = {2,1};
    static char dotted[2] = {1,3};
    GdkColor colour;
    int i, j;

    for (i = 0; i < 3; i++) {
        for (j = 0; j < 3; j++) {
            gc[(i*3)+j] = gdk_gc_new(widget->window);
        }

        colour.pixel = 0;
        colour.red = (i == 1)?65535:0;
        colour.green = 0;
        colour.blue = (i == 2)?65535:0;

        gdk_color_alloc(gdk_window_get_colormap(widget->window), &colour);

        for (j = 0; j < 3; j++) {
            gdk_gc_set_foreground(gc[(i*3)+j], &colour);
        }

        gdk_gc_set_line_attributes(gc[i*3], 0, GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER);
        gdk_gc_set_dashes(gc[(i*3)+1], 0, dashed, 2);
        gdk_gc_set_line_attributes(gc[(i*3)+1], 0, GDK_LINE_ON_OFF_DASH, GDK_CAP_BUTT, GDK_JOIN_MITER);
        gdk_gc_set_dashes(gc[(i*3)+2], 0, dotted, 2);
        gdk_gc_set_line_attributes(gc[(i*3)+2], 0, GDK_LINE_ON_OFF_DASH, GDK_CAP_BUTT, GDK_JOIN_MITER);
    }

    if ((font_info[0] = gdk_font_load("lucidasans-10")) == NULL) {
	font_info[0] = gdk_font_load("fixed");
    }

    if ((font_info[1] = gdk_font_load("lucidasans-12")) == NULL) {
	font_info[1] = gdk_font_load("fixed");
    }
}

static GdkPixmap *get_blank_graph(GtkWidget *widget)
{
    GdkPixmap *canvas = gtk_object_get_data(GTK_OBJECT(widget), "canvas");

    if (canvas != NULL) {
        gdk_pixmap_unref(canvas);
    }
    if (widget->window != NULL) {
        canvas = gdk_pixmap_new(widget->window, widget->allocation.width, widget->allocation.height, -1);
    }
    else {
        canvas = NULL;
    }
    if (canvas != NULL) {
        gdk_draw_rectangle(canvas, widget->style->white_gc, TRUE, 0, 0, widget->allocation.width, widget->allocation.height);
    }
    gtk_object_set_data(GTK_OBJECT(widget), "canvas", canvas);
    return(canvas);
}

static BOOLEAN expose_graph(GtkWidget *widget, GdkEventExpose *event, void *data)
{
    GdkPixmap *canvas = gtk_object_get_data(GTK_OBJECT(widget), "canvas");

    if (event == NULL) {
        gdk_draw_pixmap(widget->window,
                    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                    canvas,
                    0, 0, 0, 0,
                    widget->allocation.width, widget->allocation.height);
    }
    else {
        gdk_draw_pixmap(widget->window,
                    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                    canvas,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);

    }
    return(FALSE);
}

static GtkWidget *create_button(GtkWidget *box, char *label, void callback(), void *data)
{
    GtkWidget *button = gtk_button_new_with_label(label);
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(callback), data);
    gtk_box_pack_start(GTK_BOX(box), button, TRUE, TRUE, 0);
    gtk_widget_show(button);
    return(button);
}

static GtkWidget *create_graph(GtkWidget *vbox, int height, BOOLEAN repaint_proc(), char *label)
{
    GtkWidget *graph = gtk_drawing_area_new();
    gtk_widget_set_usize(graph, 0, height);
    gtk_box_pack_start(GTK_BOX(vbox), graph, TRUE, TRUE, 0);
    gtk_object_set_data(GTK_OBJECT(graph), "plotter label", label);
    gtk_signal_connect(GTK_OBJECT(graph), "expose_event", GTK_SIGNAL_FUNC(expose_graph), NULL);
    gtk_signal_connect(GTK_OBJECT(graph), "configure_event", GTK_SIGNAL_FUNC(repaint_proc), graph);
    gtk_widget_show(graph);
    return(graph);
}

static GtkWidget *create_menu(char *label, BOOLEAN menu_generate(), void *data)
{
    GtkWidget *menu_bar = gtk_menu_bar_new();
    GtkWidget *menu = gtk_menu_new();
    GtkWidget *item = gtk_menu_item_new_with_label(label);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);
    gtk_signal_connect(GTK_OBJECT(item), "enter_notify_event", GTK_SIGNAL_FUNC(menu_generate), data);
    gtk_widget_show(item);

    gtk_menu_bar_append(GTK_MENU_BAR(menu_bar), item);
    gtk_widget_show(menu_bar);
    return(menu_bar);
}

static char *maths_name(MathsType m)
{
    switch (m) {
        case SIGMOID: { return("Sigmoid"); }
        case TANH: { return("Tanh"); }
        case HALFX: { return("Half X"); }
        case M_R: { return("M/R"); }
        case HOUGHTON: { return("Houghton"); }
        case NORMALISED: { return("Normalised"); }
    }
    return(NULL);
}

static int calculate_x_step(int range)
{
    /* If range < 10^(n+1), make step 10^n */
    int n = (log(range) / log(10.0));
    return((int) pow(10.0, (double) n));
}

static void draw_axes(GtkWidget *graph, GdkPixmap *canvas, double x_scale, int x_range, double y_scale, double y_range)
{
    int height = graph->allocation.height;
    int width = graph->allocation.width;
    char buffer[16], *label;
    int i, j;

    gdk_draw_line(canvas, gc[0], PLOT_WORLD_X_MIN, height+PLOT_WORLD_Y_MAX+1, PLOT_WORLD_X_MIN, PLOT_WORLD_Y_MIN-1);
    gdk_draw_line(canvas, gc[0], PLOT_WORLD_X_MIN-1, height+PLOT_WORLD_Y_MAX, width+PLOT_WORLD_X_MAX+1, height+PLOT_WORLD_Y_MAX);

    /* Draw x-axis labels and vertical grid lines */

    i = calculate_x_step(x_range);

    for (j = 0; j <= x_range; j += i) {
        short w, x;

	g_snprintf(buffer, 16, "%d", j);
	w = gdk_text_width(font_info[0], buffer, strlen(buffer));
	x = ((short) (j * x_scale) + PLOT_WORLD_X_MIN);
	gdk_draw_text(canvas, font_info[0], gc[0], x - (w/2), height+PLOT_WORLD_Y_MAX+12, buffer, strlen(buffer));
	gdk_draw_line(canvas, gc[2], x, PLOT_WORLD_Y_MIN-2, x, height+PLOT_WORLD_Y_MAX+1);
    }

    /* Draw y-axis labels and horizontal grid lines */


    for (j = 0; j <= 5; j++) {
	short w, y;

	g_snprintf(buffer, 16, "%3.1f", y_range * j / 5.0);
	w = gdk_text_width(font_info[0], buffer, strlen(buffer));
	y = (height - (short) (j * 0.2 * y_scale) + PLOT_WORLD_Y_MAX);
	gdk_draw_text(canvas, font_info[0], gc[0], PLOT_WORLD_X_MIN-5-w, y+4, buffer, strlen(buffer));
	gdk_draw_line(canvas, gc[2], PLOT_WORLD_X_MIN-2, y, width+PLOT_WORLD_X_MAX+1, y);
    }

    if ((label = (char *)gtk_object_get_data(GTK_OBJECT(graph), "plotter label")) != NULL) {
	gdk_draw_text(canvas, font_info[1], gc[0], 5, height-5, label, strlen(label));
    }
}

static void draw_plot_line(GdkPixmap *canvas, double *values, BOOLEAN highlighted, int count, double x_scale, double y_scale, int height)
{
    GdkGC *current_gc = (highlighted?gc[3]:gc[6]);
    short x0, y0, x1, y1;
    int i;

    x0 = PLOT_WORLD_X_MIN;
    y0 = (short) (height - (values[0] * y_scale)) + PLOT_WORLD_Y_MAX;
    for (i = 1; i < count; i++) {
        x1 = (short) (i * x_scale) + PLOT_WORLD_X_MIN;
        y1 = (short) (height - values[i] * y_scale) + PLOT_WORLD_Y_MAX;
        gdk_draw_line(canvas, current_gc, x0, y0, x1, y1);
        x0 = x1;
        y0 = y1;
    }
}

static int calculate_range(int count)
{
    if (count > 0) {
        /* If count < 10^(n+1), round up to nearest 10^n */
        int n = (log(count) / log(10.0));
        int scale = (int) pow(10.0, (double) n);
        return(scale * (1 + (count / scale)));
    }
    else {
        return(1);
    }
}

static void redraw_schema_data(int count)
{
    GdkPixmap *canvas = get_blank_graph(plotter[0]);
    CsSchema *tmp;

    int height = plotter[0]->allocation.height;
    int width = plotter[0]->allocation.width;
    int x_range = calculate_range(count);
    double x_scale = (width - (PLOT_WORLD_X_MIN - PLOT_WORLD_X_MAX)) / (double) x_range;
    double y_scale = (height - (PLOT_WORLD_Y_MIN - PLOT_WORLD_Y_MAX)) / (double) (MAX_ACT - MIN_ACT);

    gtk_object_set_data(GTK_OBJECT(plotter[0]), "cycles", (void *)count);

    draw_axes(plotter[0], canvas, x_scale, x_range, y_scale, 1.0);
    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        draw_plot_line(canvas, &(tmp->activation[0]), tmp->highlighted, count, x_scale, y_scale, height);
    }
    expose_graph(plotter[0], NULL, NULL);
}

static void redraw_object_data(int count)
{
    GdkPixmap *canvas = get_blank_graph(plotter[1]);
    CsObject *tmp;

    int height = plotter[1]->allocation.height;
    int width = plotter[1]->allocation.width;
    int x_range = calculate_range(count);
    double x_scale = (width - (PLOT_WORLD_X_MIN - PLOT_WORLD_X_MAX)) / (double) x_range;
    double y_scale = (height - (PLOT_WORLD_Y_MIN - PLOT_WORLD_Y_MAX)) / (double) (0.5);
    static char label[40];

    int f = (int) gtk_object_get_data(GTK_OBJECT(plotter[1]), "selected obj function");

    sprintf(label, "Object Activations (%s)", function_name[f]);
    gtk_object_set_data(GTK_OBJECT(plotter[1]), "plotter label", label);

    gtk_object_set_data(GTK_OBJECT(plotter[1]), "cycles", (void *)count);

    draw_axes(plotter[0], canvas, x_scale, x_range, y_scale, 0.5);
    for (tmp = objects; tmp != NULL; tmp = tmp->next) {
        draw_plot_line(canvas, &(tmp->activation[f][0]), tmp->highlighted, count, x_scale, y_scale, height);
    }
    expose_graph(plotter[1], NULL, NULL);
}

static void redraw_resource_data(int count)
{
    GdkPixmap *canvas = get_blank_graph(plotter[2]);
    CsResource *tmp;

    int height = plotter[2]->allocation.height;
    int width = plotter[2]->allocation.width;
    int x_range = calculate_range(count);
    double x_scale = (width - (PLOT_WORLD_X_MIN - PLOT_WORLD_X_MAX)) / (double) x_range;
    double y_scale = (height - (PLOT_WORLD_Y_MIN - PLOT_WORLD_Y_MAX)) / (double) (MAX_ACT - MIN_ACT);

    gtk_object_set_data(GTK_OBJECT(plotter[2]), "cycles", (void *)count);

    draw_axes(plotter[2], canvas, x_scale, x_range, y_scale, 1.0);
    for (tmp = resources; tmp != NULL; tmp = tmp->next) {
        draw_plot_line(canvas, &(tmp->activation[0]), tmp->highlighted, count, x_scale, y_scale, height);
    }
    expose_graph(plotter[2], NULL, NULL);
}

static BOOLEAN repaint_schema_plotter(GtkWidget *widget, GdkEventConfigure *event, GtkWidget *graph)
{
    redraw_schema_data((int) gtk_object_get_data(GTK_OBJECT(graph), "cycles"));
}

static BOOLEAN repaint_object_plotter(GtkWidget *widget, GdkEventConfigure *event, GtkWidget *graph)
{
    redraw_object_data((int) gtk_object_get_data(GTK_OBJECT(graph), "cycles"));
}

static BOOLEAN repaint_resource_plotter(GtkWidget *widget, GdkEventConfigure *event, GtkWidget *graph)
{
    redraw_resource_data((int) gtk_object_get_data(GTK_OBJECT(graph), "cycles"));
}

void AttachData(int count, BOOLEAN redraw)
{
    char buffer[16];

    sprintf(buffer, "Cycle: %3d", count);
    gtk_label_set_text(GTK_LABEL(cycle_count_widget), buffer);

    if (redraw) {
        redraw_schema_data(count);
        redraw_object_data(count);
        redraw_resource_data(count);
    }
}

static void reset_parameter_widgets()
{
    char buffer[10];
    ParameterList *p;

    for (p = parameters; p != NULL; p = p->next) {
        if (p->type == LINEAR_SLIDER) {
            double d = 100.0 * *((double *)p->value);
            gtk_adjustment_set_value(GTK_ADJUSTMENT(p->slider_widget), d);
            g_snprintf(buffer, 10, "(%4.2f)", *((double *)p->value));
            gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
        }
        else if (p->type == INTEGER_SLIDER) {
            double d = *((int *)p->value) / 5.0;
            gtk_adjustment_set_value(GTK_ADJUSTMENT(p->slider_widget), d);
            g_snprintf(buffer, 10, "(%3d)", *((int *)p->value));
            gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
        }
        else if (p->type == REAL_SLIDER) {
            double d = 10.0 * *((double *)p->value);
            gtk_adjustment_set_value(GTK_ADJUSTMENT(p->slider_widget), d);
            g_snprintf(buffer, 10, "(%4.2f)", *((double *)p->value));
            gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
        }
        else if (p->type == LOG_SLIDER) {
            double d = 100.0 + (10.0 * log10(*((double *)p->value)));
            gtk_adjustment_set_value(GTK_ADJUSTMENT(p->slider_widget), d);
            g_snprintf(buffer, 10, "(%3.2e)", *((double *)p->value));
            gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
        }
        else if (p->type == MATHS_SELECTOR) {
            gtk_option_menu_set_history(GTK_OPTION_MENU(p->value_widget), *((int *)p->value));
	}
        else if (p->type == BOOLEAN_SELECTOR) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p->slider_widget), ((*(BOOLEAN *)p->value) != FALSE));
        }
    }
}

/******************************************************************************/
/*      The Button callbacks                                                  */

static void HighlightNoSchemaCallback(GtkWidget *mi, void *dummy) 
{
    CsSchema *tmp;

    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        if (tmp->highlighted) {
            tmp->highlighted = FALSE;
        }
    }
    redraw_schema_data((int) gtk_object_get_data(GTK_OBJECT(plotter[0]), "cycles"));
}

static void HighlightAllSchemaCallback(GtkWidget *mi, void *dummy)
{
    CsSchema *tmp;

    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        if (!(tmp->highlighted)) {
            tmp->highlighted = TRUE;
        }
    }
    redraw_schema_data((int) gtk_object_get_data(GTK_OBJECT(plotter[0]), "cycles"));
}

static void HighlightSchemaCallback(GtkWidget *mi, CsSchema *schema)
{
    schema->highlighted = !(schema->highlighted);
    redraw_schema_data((int) gtk_object_get_data(GTK_OBJECT(plotter[0]), "cycles"));
}

static void HighlightNoObjectCallback(GtkWidget *m, GtkWidget *mi)
{
    CsObject *tmp;

    for (tmp = objects; tmp != NULL; tmp = tmp->next) {
        if (tmp->highlighted) {
            tmp->highlighted = FALSE;
        }
    }
    redraw_object_data((int) gtk_object_get_data(GTK_OBJECT(plotter[1]), "cycles"));
 }

static void HighlightAllObjectCallback(GtkWidget *m, GtkWidget *mi)
{
    CsObject *tmp;

    for (tmp = objects; tmp != NULL; tmp = tmp->next) {
        if (!(tmp->highlighted)) {
            tmp->highlighted = TRUE;
        }
    }
    redraw_object_data((int) gtk_object_get_data(GTK_OBJECT(plotter[1]), "cycles"));
}

static void HighlightObjectCallback(GtkWidget *mi, CsObject *object)
{
    object->highlighted = !(object->highlighted);
    redraw_object_data((int) gtk_object_get_data(GTK_OBJECT(plotter[1]), "cycles"));
}

static void HighlightNoResourceCallback(GtkWidget *mi, void *dummy)
{
    CsResource *tmp;

    for (tmp = resources; tmp != NULL; tmp = tmp->next) {
        if (tmp->highlighted) {
            tmp->highlighted = FALSE;
        }
    }
    redraw_resource_data((int) gtk_object_get_data(GTK_OBJECT(plotter[2]), "cycles"));
}

static void HighlightAllResourceCallback(GtkWidget *mi, void *dummy)
{
    CsResource *tmp;

    for (tmp = resources; tmp != NULL; tmp = tmp->next) {
        if (!(tmp->highlighted)) {
            tmp->highlighted = TRUE;
        }
    }
    redraw_resource_data((int) gtk_object_get_data(GTK_OBJECT(plotter[2]), "cycles"));
}

static void HighlightResourceCallback(GtkWidget *mi, CsResource *current)
{
    current->highlighted = !(current->highlighted);
    redraw_resource_data((int) gtk_object_get_data(GTK_OBJECT(plotter[2]), "cycles"));
}

static void QuitCallback(GtkWidget *button, GtkWidget *frame)
{
    gtk_widget_destroy(frame);
}

static void RunCallback(GtkWidget *button, GtkWidget *dummy)
{
    extern void ActAnalyseList(FILE *fp);

    int cycles = CsRun(stdout, current_task);
    AttachData(cycles, TRUE);
    CsMessage(WARNING, cycles, "Run complete\n\n");
    ActAnalyseList(stdout);
}

static void ParamsCallback(GtkWidget *button, GtkWidget *params_popup)
{
    /* Ensure the popup up is open, visible, and */
    /* at the front of the window stack.         */

    gtk_widget_hide(params_popup);
    gtk_widget_show(params_popup);
}

static void DismissParamsCallback(GtkWidget *button, GtkWidget *params_popup)
{
    gtk_widget_hide(params_popup);
}

static void SaveParamsCallback(GtkWidget *button, GtkWidget *params_popup)
{
    CsSaveDefaultParameters(current_task);
}

static void DefaultParamsCallback(GtkWidget *button, GtkWidget *params_popup)
{
    CsSetDefaultParameters(current_task);
    reset_parameter_widgets();
}

static void InitialiseCallback(GtkWidget *button, GtkWidget *params_popup)
{
    CsReinitialise(current_task);
    AttachData(0, TRUE);
}

static void StepCallback(GtkWidget *button, GtkWidget *steps)
{
    AttachData(CsStep(stdout, current_task, (int) steps), TRUE);
}


static void slider_callback(GtkAdjustment *adjustment, ParameterList *p)
{
    if (p == NULL) {
	fprintf(stderr, "INTERNAL ERROR: NULL parameter in slider_callback\n");
    }
    else {
	double value = (double) adjustment->value;
	char buffer[10];

	switch (p->type) {
	    case LINEAR_SLIDER: {
		double real_value = (double) value / 100.0;
		g_snprintf(buffer, 10, "(%4.2f)", real_value);
		*((double *)p->value) = real_value;
		break;
	    }
	    case REAL_SLIDER: {
		double real_value = (double) value / 10.0;
		g_snprintf(buffer, 10, "(%4.2f)", real_value);
		*((double *)p->value) = real_value;
		break;
	    }
	    case LOG_SLIDER: {
		double real_value = pow(10.0, (((double) value - 100.0) / 10.0));
		g_snprintf(buffer, 10, "(%3.2e)", real_value);
		*((double *)p->value) = real_value;
		break;
	    }
	    case INTEGER_SLIDER: {
		int real_value = value * 5.0;
		g_snprintf(buffer, 10, "(%d)", real_value);
		*((int *)p->value) = real_value;
		break;
	    }
	    default: {
		return;
	    }
	}
	gtk_label_set_text(GTK_LABEL(p->value_widget), buffer);
    }
}

static void maths_callback(GtkWidget *widget, ParameterList *p)
{
    GtkWidget *mi = gtk_menu_get_active(GTK_MENU(p->slider_widget));

    *(MathsType *)p->value = (MathsType) gtk_object_get_data(GTK_OBJECT(mi), "data");
}


static void boolean_callback(GtkWidget *widget, ParameterList *p)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(p->slider_widget))) {
        *(BOOLEAN *)p->value = TRUE;
    }
    else {
        *(BOOLEAN *)p->value = FALSE;
    }
}

/*
 *      Create form and menu Buttons
 */

static void XcsCreateParamWidgets(char *label, void *param, ParameterType type, GtkWidget *box)
{
    ParameterList *p = (ParameterList *)malloc(sizeof(ParameterList));
    GtkWidget *tmp, *hbox;

    if (p != NULL) {

        p->value = param;

        p->type = type;

        hbox = gtk_hbox_new(FALSE, 4);
        gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 4);
        gtk_widget_show(hbox);

        tmp = gtk_label_new(label);
        gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
        gtk_widget_show(tmp);


        if (type == MATHS_SELECTOR) {
	    GtkWidget *mi;
            int i;

            p->value_widget = gtk_option_menu_new();
            p->slider_widget = gtk_menu_new();

            for (i = TANH; i <= NORMALISED; i++) {
                mi = gtk_menu_item_new_with_label(maths_name(i));
                gtk_menu_append(GTK_MENU(p->slider_widget), mi);
                gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(maths_callback), p);
                gtk_object_set_data(GTK_OBJECT(mi), "data", (void *)i);
                gtk_widget_show(mi);
            }
            gtk_option_menu_set_menu(GTK_OPTION_MENU(p->value_widget), p->slider_widget);
            gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(p->value_widget), FALSE, FALSE, 0);
            gtk_widget_show(p->value_widget);

        }
        else if (type == BOOLEAN_SELECTOR) {
	    p->slider_widget = (void *)gtk_check_button_new();
            gtk_signal_connect(GTK_OBJECT(p->slider_widget), "toggled", GTK_SIGNAL_FUNC(boolean_callback), p);
	    gtk_widget_set_usize(p->slider_widget, 200, 0);
            gtk_box_pack_end(GTK_BOX(hbox), GTK_WIDGET(p->slider_widget), FALSE, FALSE, 0);
            gtk_widget_show(GTK_WIDGET(p->slider_widget));
        }
	else { /* Some kind of slider widget ... */
            p->slider_widget = (void *)gtk_adjustment_new(0.0, 0.0, 100.0, 1.0, 1.0, 0.0);
	    p->value_widget = gtk_label_new(" ");
	    gtk_widget_set_usize(p->value_widget, 50, 0);
            gtk_box_pack_end(GTK_BOX(hbox), p->value_widget, FALSE, FALSE, 0);
	    gtk_widget_show(p->value_widget);

	    tmp = gtk_hscale_new(GTK_ADJUSTMENT(p->slider_widget));
            gtk_widget_set_usize(tmp, 160, 5);
            gtk_scale_set_draw_value(GTK_SCALE(tmp), FALSE);
            gtk_box_pack_end(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
	    gtk_widget_show(tmp);

            gtk_signal_connect(GTK_OBJECT(p->slider_widget), "value_changed", GTK_SIGNAL_FUNC(slider_callback), p);
	}

        p->next = parameters;
        parameters = p;
    }
    else {
        CsMessage(WARNING, 0, "Parameter allocation failed: Extra parameters ignored\n");
    }
}

static GtkWidget *make_params_popup(GtkWidget *frame)
{
    GtkWidget *params_popup, *vbox;
    GtkWidget *button_panel, *main_panel;

    params_popup = gtk_window_new(GTK_WINDOW_DIALOG);

    gtk_widget_set_usize(params_popup, 350, 0);
    gtk_container_set_border_width(GTK_CONTAINER(params_popup), 4);
    gtk_window_set_title(GTK_WINDOW(params_popup), "XCS Control Panel");

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_add(GTK_CONTAINER(params_popup), vbox);
    gtk_widget_show(vbox);

    button_panel = gtk_hbox_new(TRUE, 4);
    gtk_box_pack_start(GTK_BOX(vbox), button_panel, FALSE, FALSE, 0);
    gtk_widget_show(button_panel);

    create_button(button_panel, "Default", DefaultParamsCallback, params_popup);
    create_button(button_panel, "Run", RunCallback, NULL);
    create_button(button_panel, "Dismiss", DismissParamsCallback, params_popup);
    create_button(button_panel, "Save", SaveParamsCallback, params_popup);
    create_button(button_panel, "Quit", QuitCallback, frame);

    button_panel = gtk_hbox_new(TRUE, 4);
    gtk_box_pack_start(GTK_BOX(vbox), button_panel, FALSE, FALSE, 0);
    gtk_widget_show(button_panel);

    create_button(button_panel, "Initialise", InitialiseCallback, NULL);
    create_button(button_panel,   "1 Step",  StepCallback, (void *)   1);
    create_button(button_panel,  "25 Steps", StepCallback, (void *)  25);
    create_button(button_panel, "100 Steps", StepCallback, (void *) 100);

    main_panel = gtk_vbox_new(FALSE, 4);
    gtk_container_add(GTK_CONTAINER(vbox), main_panel);
    gtk_widget_show(main_panel);

/*
 * Create check boxes for the display parameters:
 */

    XcsCreateParamWidgets("Display actions:",     &display_message.action,      BOOLEAN_SELECTOR, main_panel);
    XcsCreateParamWidgets("Display selections:",  &display_message.selection,   BOOLEAN_SELECTOR, main_panel);
    XcsCreateParamWidgets("Display world:",       &display_message.world,       BOOLEAN_SELECTOR, main_panel);
    XcsCreateParamWidgets("Display warnings:",    &display_message.warning,     BOOLEAN_SELECTOR, main_panel); 
    XcsCreateParamWidgets("Display activations:", &display_message.activations, BOOLEAN_SELECTOR, main_panel);

/*
 * Create sliders for the parameters:
 */

    XcsCreateParamWidgets("Rest Activation:",     &REST_ACT,            LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Persistence:",         &PERSISTENCE,         LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Maths: ",              &MATHS,               MATHS_SELECTOR, main_panel);
    XcsCreateParamWidgets("Selection Threshold:", &SELECTION_THRESHOLD, LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Noise SD:",            &NOISE_SD,            LOG_SLIDER,     main_panel);
    XcsCreateParamWidgets("Cycle Time (ms): ",    &MS_PER_CYCLE,        REAL_SLIDER,    main_panel);

    XcsCreateParamWidgets("Visual Strength: ",    &VISUAL_ROUTE_STRENGTH,   LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Visual Lag (ms): ",    &VISUAL_LAG,              INTEGER_SLIDER, main_panel);
    XcsCreateParamWidgets("Auditory Strength: ",  &AUDITORY_ROUTE_STRENGTH, LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Auditory Lag (ms): ",  &AUDITORY_LAG,            INTEGER_SLIDER, main_panel);

    XcsCreateParamWidgets("Schema C:N:",          &SCHEMA_SI, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Schema P:E:",          &SCHEMA_PE, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Schema S:L:",          &SCHEMA_SL, LINEAR_SLIDER, main_panel);

    XcsCreateParamWidgets("Object C:N:",          &OBJECT_SI, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Object P:E:",          &OBJECT_PE, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Object S:L:",          &OBJECT_SL, LINEAR_SLIDER, main_panel);

    XcsCreateParamWidgets("Resource C:N:",        &EFFECT_SI, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Resource P:E:",        &EFFECT_PE, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Resource S:L:",        &EFFECT_SL, LINEAR_SLIDER, main_panel);

    reset_parameter_widgets();

    return(params_popup);
}

/******************************************************************************/
/*  MENU GENERATION PROCEDURES FOR SCHEMA/OBJECT/RESOUCE MENUS                */
/******************************************************************************/

static void SetTaskCallback(GtkWidget *item, void *task)
{
    CsInitialise((int) task);
}

static void select_object_role_callback(GtkWidget *item, void *dummy)
{
    int role = (int) gtk_object_get_data(GTK_OBJECT(item), "role");

    gtk_object_set_data(GTK_OBJECT(plotter[1]), "selected obj function", (void *)role);
    redraw_object_data((int) gtk_object_get_data(GTK_OBJECT(plotter[1]), "cycles"));
}

static BOOLEAN schemaMenuGenerate(GtkWidget *item, GdkEvent *event, void *dummy)
{
    GtkWidget *menu, *mi;
    CsSchema *tmp;
    char buffer[64];

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label(" None");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(HighlightNoSchemaCallback), dummy);
    gtk_menu_append(GTK_MENU(menu), mi);
    gtk_widget_show(mi);

    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        g_snprintf(buffer, 64, "%c %s", tmp->highlighted?'*':' ', tmp->name);
        mi = gtk_menu_item_new_with_label(buffer);
        gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(HighlightSchemaCallback), tmp);
        gtk_menu_append(GTK_MENU(menu), mi);
        gtk_widget_show(mi);
    }

    mi = gtk_menu_item_new_with_label(" All");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(HighlightAllSchemaCallback), dummy);
    gtk_menu_append(GTK_MENU(menu), mi);
    gtk_widget_show(mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

static BOOLEAN objectMenuGenerate(GtkWidget *item, GdkEvent *event, void *dummy)
{
    GtkWidget *menu, *mi;
    CsObject *tmp;
    char buffer[64];

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label(" None");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(HighlightNoObjectCallback), dummy);
    gtk_menu_append(GTK_MENU(menu), mi);
    gtk_widget_show(mi);

    for (tmp = objects; tmp != NULL; tmp = tmp->next) {
        g_snprintf(buffer, 64, "%c %s", tmp->highlighted?'*':' ', tmp->name);
        mi = gtk_menu_item_new_with_label(buffer);
        gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(HighlightObjectCallback), tmp);
        gtk_menu_append(GTK_MENU(menu), mi);
        gtk_widget_show(mi);
    }

    mi = gtk_menu_item_new_with_label(" All");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(HighlightAllObjectCallback), dummy);
    gtk_menu_append(GTK_MENU(menu), mi);
    gtk_widget_show(mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

static BOOLEAN resourceMenuGenerate(GtkWidget *item, GdkEvent *event, void *dummy)
{
    GtkWidget *menu, *mi;
    CsResource *tmp;
    char buffer[64];

    gtk_menu_item_remove_submenu(GTK_MENU_ITEM(item));
    menu = gtk_menu_new();

    mi = gtk_menu_item_new_with_label(" None");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(HighlightNoResourceCallback), dummy);
    gtk_menu_append(GTK_MENU(menu), mi);
    gtk_widget_show(mi);

    for (tmp = resources; tmp != NULL; tmp = tmp->next) {
        g_snprintf(buffer, 64, "%c %s", tmp->highlighted?'*':' ', tmp->name);
        mi = gtk_menu_item_new_with_label(buffer);
        gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(HighlightResourceCallback), tmp);
        gtk_menu_append(GTK_MENU(menu), mi);
        gtk_widget_show(mi);
    }

    mi = gtk_menu_item_new_with_label(" All");
    gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(HighlightAllResourceCallback), dummy);
    gtk_menu_append(GTK_MENU(menu), mi);
    gtk_widget_show(mi);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu);

    /* Continue processing the event... */

    return(FALSE);
}

static void exit_xvcs(GtkWidget *button, void *data)
{
    gtk_exit(0);
}

static GtkWidget *make_widgets()
{
    GtkWidget *frame, *params_popup;
    GtkWidget *vbox, *hbox;
    GtkWidget *menu;

    frame = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_container_set_border_width(GTK_CONTAINER(frame), 4);
    gtk_window_set_policy(GTK_WINDOW(frame), TRUE, TRUE, FALSE);
    gtk_widget_set_usize(frame, 640, 0);
    gtk_window_set_title(GTK_WINDOW(frame), "Contention Scheduling [V6; C/GTK+]");
    gtk_signal_connect(GTK_OBJECT(frame), "destroy", GTK_SIGNAL_FUNC(exit_xvcs), 0);
    gtk_widget_realize(frame);

    create_gcs(frame);

    params_popup = make_params_popup(frame);

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_add(GTK_CONTAINER(frame), vbox);
    gtk_widget_show(vbox);

    /* The top box of buttons: */

    hbox = gtk_hbox_new(TRUE, 4);
    gtk_widget_set_usize(hbox, 0, 26);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

/* Row 1: Open the control panel */

    create_button(hbox, "Run", RunCallback, NULL);
    create_button(hbox, "Params", ParamsCallback, params_popup);

    /* The task menu: */

    {
        GtkWidget *label, *tmp, *mi; int i;

        label = gtk_label_new("Task:");
        gtk_widget_show(label);

        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

        menu = gtk_option_menu_new();
        tmp = gtk_menu_new();
        for (i = 0; i <= (int) CONFLICT_RT; i++) {
            mi = gtk_menu_item_new_with_label(task_name((TaskType) i));
            gtk_menu_append(GTK_MENU(tmp), mi);
            gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(SetTaskCallback), (void *)i);
            gtk_widget_show(mi);
        }
        gtk_widget_show(menu);
        gtk_option_menu_set_menu(GTK_OPTION_MENU(menu), tmp);
        gtk_option_menu_set_history(GTK_OPTION_MENU(menu), current_task);
        gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(menu), FALSE, FALSE, 0);
    }

    cycle_count_widget = gtk_label_new("Cycle:   0");
    gtk_box_pack_start(GTK_BOX(hbox), cycle_count_widget, TRUE, TRUE, 0);
    gtk_widget_show(cycle_count_widget);

    create_button(hbox, "Quit", QuitCallback, frame);

/* Row 1: highlight schema menu */

    hbox = gtk_hbox_new(FALSE, 4);
    gtk_widget_set_usize(hbox, 0, 26);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    menu = create_menu("Highlight", schemaMenuGenerate, NULL);
    gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(menu), FALSE, FALSE, 0);

/* Row 2: The schema graph */

    plotter[0] = create_graph(vbox, PLOTTER_HEIGHT, repaint_schema_plotter, "Schema Activations");

/* Row 3: The object buttons */

    hbox = gtk_hbox_new(FALSE, 4);
    gtk_widget_set_usize(hbox, 0, 26);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    menu = create_menu("Highlight", objectMenuGenerate, NULL);
    gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(menu), FALSE, FALSE, 0);

    /* The object role menu: */

    {
        GtkWidget *label, *tmp, *mi; int i;

        label = gtk_label_new("     Role:");
        gtk_widget_show(label);
        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 5);

        menu = gtk_option_menu_new();
        tmp = gtk_menu_new();
        for (i = 0; i < MaxFunctions-2; i++) {
            mi = gtk_menu_item_new_with_label(function_name[i]);
            gtk_menu_append(GTK_MENU(tmp), mi);
            gtk_signal_connect(GTK_OBJECT(mi), "activate", GTK_SIGNAL_FUNC(select_object_role_callback), NULL);
            gtk_object_set_data(GTK_OBJECT(mi), "role", (void *)i);
            gtk_widget_show(mi);
        }
        gtk_widget_show(menu);
        gtk_option_menu_set_menu(GTK_OPTION_MENU(menu), tmp);
        gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(menu), FALSE, FALSE, 0);
    }

/* Row 4: The object graph */

    plotter[1] = create_graph(vbox, PLOTTER_HEIGHT, repaint_object_plotter, "Object Activations");

/* Row 5: The Resource buttons */

    hbox = gtk_hbox_new(FALSE, 4);
    gtk_widget_set_usize(hbox, 0, 26);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    menu = create_menu("Highlight", resourceMenuGenerate, NULL);
    gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(menu), FALSE, FALSE, 0);

/* Row 6: The Resource graph */

    plotter[2] = create_graph(vbox, PLOTTER_HEIGHT, repaint_resource_plotter, "Resource Activations");

    return(frame);
}

/*
 *      Main
 */

int main(int argc, char **argv)
{
    GtkWidget *context;

    gtk_set_locale();
    gtk_init(&argc, &argv);

    CsSetDefaultParameters(current_task);

/* Must initialise schemata etc BEFORE making the form, so as to get correct menus */

    CsInitialise(current_task);

    if ((context = make_widgets()) != NULL) {
        AttachData(0, TRUE);
        gtk_widget_show(context);

//        gtk_widget_set_events(plotter[0], GDK_EXPOSURE_MASK);
//        gtk_widget_set_events(plotter[1], GDK_EXPOSURE_MASK);
//        gtk_widget_set_events(plotter[2], GDK_EXPOSURE_MASK);

        gtk_main();
    }

    exit(0);
}

/******** END OF FILE *********************************************************/

