/*
 *      xvcs.c: CS with X plotting tool using Xview widgets
 */

#include "cs.h"

#include <X11/Intrinsic.h>

#include <xview/xview.h>
#include <xview/frame.h>
#include <xview/panel.h>
#include <xview/icon.h>
#include <xview/xv_xrect.h>

#include <math.h> 

#define  PLOTTER_HEIGHT 150

static short icon_bits[256] = {
#include "cs_icon.xview"
};

typedef enum {LINEAR_SLIDER, REAL_SLIDER, LOG_SLIDER, INTEGER_SLIDER, MATHS_SELECTOR, BOOLEAN_SELECTOR} ParameterType;

typedef struct parameter_list {
    Panel_item slider_widget;
    Panel_item value_widget;
    ParameterType type;
    void *value;
    struct parameter_list *next;
} ParameterList;

ParameterList *parameters = NULL;
 
static Canvas plotter[3];
static Panel_item task_widget, cycle_count_widget;
static TaskType current_task = COFFEE_PREPARATION;

static XFontStruct *font_info[2];
/* font_info[0] = 10 point (small) */
/* font_info[1] = 12 point (normal) */

static GC gc[9];
/* gc[0,1,2] are black, with decreasing saturation */
/* gc[3,4,5] are blue,  with decreasing saturation */
/* gc[6,7,8] are red,   with decreasing saturation */

/* The limits of the graph within its canvas */
#define PLOT_WORLD_X_MIN 30
#define PLOT_WORLD_X_MAX -7
#define PLOT_WORLD_Y_MIN 5
#define PLOT_WORLD_Y_MAX -40

extern int  CsRun(FILE *file, TaskType task);
extern int  CsStep(FILE *file, TaskType task, int n);
extern void CsMessage(MessageType mt, int cycle, char *message);
extern void CsReinitialise(TaskType task);
extern void CsInitialise(TaskType task);
extern void CsSetDefaultParameters(TaskType task);
extern void CsSaveDefaultParameters(TaskType task);

extern char *task_name(TaskType m);

#define SELECTED_OBJ_FUNCTION   102938
#define PLOTTER_LABEL           395723

/*
 *	X stuff: callbacks, buttons, plotters, etc.
 */

static void create_gcs(Frame frame)
{
    Display *display = (Display *)xv_get(frame, XV_DISPLAY);
    Window window = RootWindow(display, DefaultScreen(display));
    Colormap colour_map = DefaultColormap(display, DefaultScreen(display));
    static unsigned char dashed[2] = {2,1};
    static unsigned char dotted[2] = {1,3};
    XColor used, exact;
    XGCValues values;

    gc[0] = DefaultGC(display, DefaultScreen(display));
    values.line_style = LineOnOffDash;
    gc[1] = XCreateGC(display, window, GCLineStyle, &values);
    XSetDashes(display, gc[1], 0, dashed, 2);
    gc[2] = XCreateGC(display, window, GCLineStyle, &values);
    XSetDashes(display, gc[2], 0, dotted, 2);

    XAllocNamedColor(display,colour_map,"blue",&used,&exact);
    values.foreground = used.pixel;
    gc[3] = XCreateGC(display, window, GCForeground, &values);
    gc[4] = XCreateGC(display, window, GCForeground | GCLineStyle, &values);
    XSetDashes(display, gc[4], 0, dashed, 2);
    gc[5] = XCreateGC(display, window, GCForeground | GCLineStyle, &values);
    XSetDashes(display, gc[5], 0, dotted, 2);

    XAllocNamedColor(display,colour_map,"red",&used,&exact);
    values.foreground = used.pixel;
    gc[6] = XCreateGC(display, window, GCForeground, &values);
    gc[7] = XCreateGC(display, window, GCForeground | GCLineStyle, &values);
    XSetDashes(display, gc[7], 0, dashed, 2);
    gc[8] = XCreateGC(display, window, GCForeground | GCLineStyle, &values);
    XSetDashes(display, gc[8], 0, dotted, 2);

    if ((font_info[0] = XLoadQueryFont(display, "lucidasans-10")) == NULL) {
        font_info[0] = XQueryFont(display, XGContextFromGC(gc[0]));
    }
    if ((font_info[1] = XLoadQueryFont(display, "lucidasans-12")) == NULL) {
        font_info[1] = XQueryFont(display, XGContextFromGC(gc[0]));
    }
}

static char *maths_name(MathsType m)
{
    switch (m) {
        case SIGMOID: { return("Sigmoid"); }
        case TANH: { return("Tanh"); }
        case HALFX: { return("Half X"); }
        case M_R: { return("M/R"); }
        case HOUGHTON: { return("Houghton"); }
        case NORMALISED: { return("Normalised"); }
    }
    return(NULL);
}

static int calculate_x_step(int range)
{
    /* If range < 10^(n+1), make step 10^n */
    int n = (log(range) / log(10.0));
    return((int) pow(10.0, (double) n));
}

static void draw_axes(Canvas plotter, double x_scale, int x_range, double y_scale, double y_range)
{
    Display *display = (Display *)xv_get(plotter, XV_DISPLAY);
    Window window = (Window) xv_get(canvas_paint_window(plotter), XV_XID);

    int height = xv_get(plotter, XV_HEIGHT);
    int width = xv_get(plotter, XV_WIDTH);
    int i, j;
    double k, l;
    char string[32], *label;

    XDrawLine(display, window, gc[0], PLOT_WORLD_X_MIN, height+PLOT_WORLD_Y_MAX+1, PLOT_WORLD_X_MIN, PLOT_WORLD_Y_MIN-1);
    XDrawLine(display, window, gc[0], PLOT_WORLD_X_MIN-1, height+PLOT_WORLD_Y_MAX, width+PLOT_WORLD_X_MAX+1, height+PLOT_WORLD_Y_MAX);

/* Draw x-axis lables and vertical grid lines */

    XSetFont(display, gc[0], font_info[0]->fid);
    i = calculate_x_step(x_range);
    for (j = 0; (j <= x_range); j = j + i) {
        short w, x;

        sprintf(string, "%d", j);
        w = XTextWidth(font_info[0], string, strlen(string));
        x = ((short) (j * x_scale) + PLOT_WORLD_X_MIN);
        XDrawString(display, window, gc[0], x - (w/2), height+PLOT_WORLD_Y_MAX+12, string, strlen(string));
        XDrawLine(display, window, gc[2], x, PLOT_WORLD_Y_MIN-2, x, height+PLOT_WORLD_Y_MAX+1);
    }

/* Draw y-axis lables and horizontal grid lines */

    k = y_range / 5.0;
    for (l = 0.0; (l <= y_range); l = l + k) {
        short w, y;

        sprintf(string, "%4.2f", l);
        w = XTextWidth(font_info[0], string, strlen(string));
        y = (height - (short) (l * y_scale) + PLOT_WORLD_Y_MAX);
        XDrawString(display, window, gc[0], PLOT_WORLD_X_MIN-5-w, y+4, string, strlen(string));
        XDrawLine(display, window, gc[2], PLOT_WORLD_X_MIN-2, y, width+PLOT_WORLD_X_MAX+1, y);
    }

    XSetFont(display, gc[0], font_info[1]->fid);
    if ((label = (char *)xv_get(plotter, XV_KEY_DATA, PLOTTER_LABEL)) != NULL) {
        XDrawString(display, window, gc[0], 5, height-5, label, strlen(label));
    }
}

static void plot_line(Display *display, Window window, double *values, Boolean highlighted, int count, double x_scale, double y_scale, int height)
{
    GC current_gc = (highlighted?gc[6]:gc[4]);
    short x0, y0, x1, y1;
    int i;

    x0 = PLOT_WORLD_X_MIN;
    y0 = (short) (height - (values[0] * y_scale)) + PLOT_WORLD_Y_MAX;
    for (i = 1; i < count; i++) {
        x1 = (short) (i * x_scale) + PLOT_WORLD_X_MIN;
        y1 = (short) (height - values[i] * y_scale) + PLOT_WORLD_Y_MAX;
        XDrawLine(display, window, current_gc, x0, y0, x1, y1);
        x0 = x1;
        y0 = y1;
    }
}

static int calculate_range(int count)
{
    if (count > 0) {
        /* If count < 10^(n+1), round up to nearest 10^n */
        int n = (log(count) / log(10.0));
        int scale = (int) pow(10.0, (double) n);
        return(scale * (1 + (count / scale)));
    }
    else {
        return(1);
    }
}

static void redraw_schema_data(int count)
{
    CsSchema *tmp;

    Display *display = (Display *)xv_get(plotter[0], XV_DISPLAY);
    Window window = (Window) xv_get(canvas_paint_window(plotter[0]), XV_XID);

    int height = xv_get(plotter[0], XV_HEIGHT);
    int width = xv_get(plotter[0], XV_WIDTH);
    int x_range = calculate_range(count);
    double x_scale = (width - (PLOT_WORLD_X_MIN - PLOT_WORLD_X_MAX)) / (double) x_range;
    double y_scale = (height - (PLOT_WORLD_Y_MIN - PLOT_WORLD_Y_MAX)) / (double) (MAX_ACT - MIN_ACT);

    xv_set(plotter[0], WIN_CLIENT_DATA, count, NULL);
    XClearWindow(display, window);
    draw_axes(plotter[0], x_scale, x_range, y_scale, 1.0);
    for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
        plot_line(display, window, &(tmp->activation[0]), tmp->highlighted, count, x_scale, y_scale, height);
    }
}

static void redraw_object_data(int count)
{
    CsObject *tmp;

    Display *display = (Display *)xv_get(plotter[1], XV_DISPLAY);
    Window window = (Window) xv_get(canvas_paint_window(plotter[1]), XV_XID);

    int height = xv_get(plotter[1], XV_HEIGHT);
    int width = xv_get(plotter[1], XV_WIDTH);
    int x_range = calculate_range(count);
    double x_scale = (width - (PLOT_WORLD_X_MIN - PLOT_WORLD_X_MAX)) / (double) x_range;
    double y_scale = (height - (PLOT_WORLD_Y_MIN - PLOT_WORLD_Y_MAX)) / (double) (0.5);
    int f = xv_get(plotter[1], XV_KEY_DATA, SELECTED_OBJ_FUNCTION);

    char label[40];

    sprintf(label, "Object Activations (%s)", function_name[f]);
    xv_set(plotter[1], XV_KEY_DATA, PLOTTER_LABEL, label, NULL);

    xv_set(plotter[1], WIN_CLIENT_DATA, count, NULL);
    XClearWindow(display, window);
    draw_axes(plotter[1], x_scale, x_range, y_scale, 0.5);
    for (tmp = objects; tmp != NULL; tmp = tmp->next) {
        plot_line(display, window, &(tmp->activation[f][0]), tmp->highlighted, count, x_scale, y_scale, height);
    }
}

static void redraw_resource_data(int count)
{
    CsResource *tmp;

    Display *display = (Display *)xv_get(plotter[2], XV_DISPLAY);
    Window window = (Window) xv_get(canvas_paint_window(plotter[2]), XV_XID);

    int height = xv_get(plotter[2], XV_HEIGHT);
    int width = xv_get(plotter[2], XV_WIDTH);
    int x_range = calculate_range(count);
    double x_scale = (width - (PLOT_WORLD_X_MIN - PLOT_WORLD_X_MAX)) / (double) x_range;
    double y_scale = (height - (PLOT_WORLD_Y_MIN - PLOT_WORLD_Y_MAX)) / (double) (MAX_ACT - MIN_ACT);

    xv_set(plotter[2], WIN_CLIENT_DATA, count, NULL);
    XClearWindow(display, window);
    draw_axes(plotter[2], x_scale, x_range, y_scale, 1.0);
    for (tmp = resources; tmp != NULL; tmp = tmp->next) {
        plot_line(display, window, &(tmp->activation[0]), tmp->highlighted, count, x_scale, y_scale, height);
    }
}

static void repaint_schema_plotter(canvas, paint_window, xdpy, xwin,
xrects)
Canvas canvas;
Xv_Window paint_window;
Display *xdpy;
Window xwin;
Xv_xrectlist *xrects;
{
    redraw_schema_data(xv_get(plotter[0], WIN_CLIENT_DATA));
}

static void repaint_object_plotter(canvas, paint_window, xdpy, xwin,
xrects)
Canvas canvas;
Xv_Window paint_window;
Display *xdpy;
Window xwin;
Xv_xrectlist *xrects;
{
    redraw_object_data(xv_get(plotter[1], WIN_CLIENT_DATA));
}

static void repaint_resource_plotter(canvas, paint_window, xdpy, xwin,
xrects)
Canvas canvas;
Xv_Window paint_window;
Display *xdpy;
Window xwin;
Xv_xrectlist *xrects;
{
    redraw_resource_data(xv_get(plotter[2], WIN_CLIENT_DATA));
}

void AttachData(int count, BOOLEAN redraw)
{
    char buffer[16];

    sprintf(buffer, "Cycle: %3d", count);
    xv_set(cycle_count_widget, PANEL_LABEL_STRING, buffer, NULL);

    if (redraw) {
        redraw_schema_data(count);
        redraw_object_data(count);
        redraw_resource_data(count);
    }
}

static void reset_parameter_widgets()
{
    char buffer[10];
    ParameterList *p;

    for (p = parameters; p != NULL; p = p->next) {
        if (p->type == LINEAR_SLIDER) {
            xv_set(p->slider_widget, PANEL_VALUE, (int) (100 * *((double *)p->value)), NULL);
            sprintf(buffer, "(%4.2f)", *((double *)p->value));
            xv_set(p->value_widget, PANEL_LABEL_STRING, buffer, NULL);
        }
        else if (p->type == INTEGER_SLIDER) {
            xv_set(p->slider_widget, PANEL_VALUE, *((int *)p->value) / 5, NULL);
            sprintf(buffer, "(%3d)", *((int *)p->value));
            xv_set(p->value_widget, PANEL_LABEL_STRING, buffer, NULL);
        }
        else if (p->type == REAL_SLIDER) {
            xv_set(p->slider_widget, PANEL_VALUE, (int) (10 * *((double *)p->value)), NULL);
            sprintf(buffer, "(%4.2f)", *((double *)p->value));
            xv_set(p->value_widget, PANEL_LABEL_STRING, buffer, NULL);
        }
        else if (p->type == LOG_SLIDER) {
            xv_set(p->slider_widget, PANEL_VALUE, (int) (100 + (10 * log10(*((double *)p->value)))), NULL);
            sprintf(buffer, "(%3.2e)", *((double *)p->value));
            xv_set(p->value_widget, PANEL_LABEL_STRING, buffer, NULL);
        }
        else if (p->type == MATHS_SELECTOR) {
            xv_set(p->value_widget, PANEL_LABEL_STRING, maths_name(*((MathsType *)p->value)), NULL);
	}
        else if (p->type == BOOLEAN_SELECTOR) {
            xv_set(p->slider_widget, PANEL_VALUE, *((Boolean *)p->value), NULL);
        }
    }
}

/******************************************************************************/
/*      The Button callbacks                                                  */

static void DisplayFunctionCallback(Menu m, Menu_item mi)
{
    int df = xv_get(mi, MENU_CLIENT_DATA);
    xv_set(plotter[1], XV_KEY_DATA, SELECTED_OBJ_FUNCTION, df, NULL);

    redraw_object_data(xv_get(plotter[1], WIN_CLIENT_DATA));
}

static void HighlightNoSchemaCallback(Menu m, Menu_item mi) 
{
    CsSchema *tmp;
    int i;

    for (i = 2, tmp = schemas; tmp != NULL; tmp = tmp->next, i++) {
        if (tmp->highlighted) {
            tmp->highlighted = FALSE;
        }
    }
    redraw_schema_data(xv_get(plotter[0], WIN_CLIENT_DATA));
}

static void HighlightAllSchemaCallback(Menu m, Menu_item mi)
{
    CsSchema *tmp;
    int i;

    for (i = 2, tmp = schemas; tmp != NULL; tmp = tmp->next, i++) {
        if (!(tmp->highlighted)) {
            tmp->highlighted = TRUE;
        }
    }
    redraw_schema_data(xv_get(plotter[0], WIN_CLIENT_DATA));
}

static void HighlightSchemaCallback(Menu m, Menu_item mi)
{
    CsSchema *schema = (CsSchema *) xv_get(mi, MENU_CLIENT_DATA);

    schema->highlighted = !(schema->highlighted);
    redraw_schema_data(xv_get(plotter[0], WIN_CLIENT_DATA));
}

static void HighlightNoObjectCallback(Menu m, Menu_item mi)
{
    CsObject *tmp;
    int i;

    for (i = 2, tmp = objects; tmp != NULL; tmp = tmp->next, i++) {
        if (tmp->highlighted) {
            tmp->highlighted = FALSE;
        }
    }
    redraw_object_data(xv_get(plotter[1], WIN_CLIENT_DATA));
 }

static void HighlightAllObjectCallback(Menu m, Menu_item mi)
{
    CsObject *tmp;
    int i;

    for (i = 2, tmp = objects; tmp != NULL; tmp = tmp->next, i++) {
        if (!(tmp->highlighted)) {
            tmp->highlighted = TRUE;
        }
    }
    redraw_object_data(xv_get(plotter[1], WIN_CLIENT_DATA));
}

static void HighlightObjectCallback(Menu m, Menu_item mi)
{
    CsObject *object = (CsObject *) xv_get(mi, MENU_CLIENT_DATA);

    object->highlighted = !(object->highlighted);
    redraw_object_data(xv_get(plotter[1], WIN_CLIENT_DATA));
}

static void HighlightNoResourceCallback(Menu m, Menu_item mi)
{
    CsResource *tmp;
    int i;

    for (i = 2, tmp = resources; tmp != NULL; tmp = tmp->next, i++) {
        if (tmp->highlighted) {
            tmp->highlighted = FALSE;
        }
    }
    redraw_resource_data(xv_get(plotter[2], WIN_CLIENT_DATA));
}

static void HighlightAllResourceCallback(Menu m, Menu_item mi)
{
    CsResource *tmp;
    int i;

    for (i = 2, tmp = resources; tmp != NULL; tmp = tmp->next, i++) {
        if (!(tmp->highlighted)) {
            tmp->highlighted = TRUE;
        }
    }
    redraw_resource_data(xv_get(plotter[2], WIN_CLIENT_DATA));
}

static void HighlightResourceCallback(Menu m, Menu_item mi)
{
    CsResource *current = (CsResource *) xv_get(mi, MENU_CLIENT_DATA);

    current->highlighted = !(current->highlighted);
    redraw_resource_data(xv_get(plotter[2], WIN_CLIENT_DATA));
}

static void QuitCallback(Panel_item item, Event *event)
{
    xv_destroy_safe((Frame) xv_get(item, PANEL_CLIENT_DATA));
}

static void RunCallback()
{
    extern void ActAnalyseList(FILE *fp);

    int cycles = CsRun(stdout, current_task);
    AttachData(cycles, TRUE);
    CsMessage(WARNING, cycles, "Run complete\n\n");
    ActAnalyseList(stdout);
}

static void ParamsCallback(Panel_item p, Event *e)
{
    Frame params_popup = (Frame) xv_get(p, PANEL_CLIENT_DATA);

    /* Ensure the popup up is open, visible, and */
    /* at the front of the window stack.         */

    xv_set(params_popup, FRAME_CLOSED, FALSE, NULL);
    xv_set(params_popup, XV_SHOW, TRUE, NULL);
    XRaiseWindow((Display *)xv_get(params_popup, XV_DISPLAY), (Window) xv_get(params_popup, XV_XID));
}

static void DismissParamsCallback(Panel_item p, Event *e)
{
    Frame params_popup = (Frame) xv_get(p, PANEL_CLIENT_DATA);
    xv_set(params_popup, XV_SHOW, FALSE, NULL);
}

static void SaveParamsCallback(Panel_item p, Event *e)
{
    CsSaveDefaultParameters(current_task);
}

static void DefaultParamsCallback()
{
    CsSetDefaultParameters(current_task);
    reset_parameter_widgets();
}

static void InitialiseCallback(Panel_item item, Event *event)
{
    CsReinitialise(current_task);
    AttachData(0, TRUE);
}

static void StepCallback(Panel_item item, Event *event)
{
    AttachData(CsStep(stdout, current_task, xv_get(item, PANEL_CLIENT_DATA)), TRUE);
}

static void SliderCallback(Panel_item s, int v, Event *e)
{
    ParameterList *p = (ParameterList *)xv_get(s, PANEL_CLIENT_DATA);

    if (p == NULL) {
        CsMessage(WARNING, 0, "NULL parameter in SliderCallback");
    }
    else {
        int value = (int) xv_get(s, PANEL_VALUE);
        char buffer[10];
        if (p->type == LINEAR_SLIDER) {
            double real_value = (double) value / 100;
            sprintf(buffer, "(%4.2f)", real_value);
            *((double *)p->value) = real_value;
        }
        else if (p->type == REAL_SLIDER) {
            double real_value = (double) value / 10;
            sprintf(buffer, "(%4.2f)", real_value);
            *((double *)p->value) = real_value;
        }
        else if (p->type == INTEGER_SLIDER) {
            int real_value = value * 5;
            sprintf(buffer, "(%d)", real_value);
            *((int *)p->value) = real_value;
        }
        else if (p->type == LOG_SLIDER) {
            double real_value = pow(10.0, (((double) value - 100) / 10));
            sprintf(buffer, "(%3.2e)", real_value);
            *((double *)p->value) = real_value;
        }
        else {
            return;
        }
        xv_set(p->value_widget, PANEL_LABEL_STRING, buffer, NULL);
    }
}

static void MathsCallback(Menu menu, Menu_item item)
{
    ParameterList *p = (ParameterList *)xv_get(menu, MENU_CLIENT_DATA);
    *(MathsType *)p->value = (MathsType) xv_get(item, MENU_CLIENT_DATA);
    xv_set(p->value_widget, PANEL_LABEL_STRING, maths_name(*((MathsType *)p->value)), NULL);
}

static void BooleanCallback(Panel_item item, Event *event)
{
    ParameterList *p = (ParameterList *)xv_get(item, PANEL_CLIENT_DATA);
    *(Boolean *)p->value = (Boolean) xv_get(item, PANEL_VALUE);
}

/*
 *      Create form and menu Buttons
 */

static void XcsCreateParamWidgets(char *label, void *param, ParameterType type, Panel parent)
{
    ParameterList *p = (ParameterList *)malloc(sizeof(ParameterList));

    if (p != NULL) {

        p->value = param;

        p->type = type;

        if (type == MATHS_SELECTOR) {
            Menu menu = xv_create(XV_NULL, MENU,
                        MENU_NOTIFY_PROC, MathsCallback,
                        MENU_CLIENT_DATA, p,
                        MENU_ITEM, MENU_STRING, "Sigmoid",
                                   MENU_CLIENT_DATA, SIGMOID,
                                   NULL,
                        MENU_ITEM, MENU_STRING, "Tanh",
                                   MENU_CLIENT_DATA, TANH,
                                   NULL,
                        MENU_ITEM, MENU_STRING, "Half X",
                                   MENU_CLIENT_DATA, HALFX,
                                   NULL,
                        MENU_ITEM, MENU_STRING, "M/R",
                                   MENU_CLIENT_DATA, M_R,
                                   NULL,
                        MENU_ITEM, MENU_STRING, "Houghton",
                                   MENU_CLIENT_DATA, HOUGHTON,
                                   NULL,
                        MENU_ITEM, MENU_STRING, "Normalised",
                                   MENU_CLIENT_DATA, NORMALISED,
                                   NULL,
                        NULL);

            p->slider_widget = xv_create(parent, PANEL_ABBREV_MENU_BUTTON,
                        PANEL_LABEL_STRING, label,
                        PANEL_NEXT_ROW, -1,
                        PANEL_ITEM_MENU, menu,
                        NULL);
            p->value_widget = xv_create(parent, PANEL_MESSAGE,
                        PANEL_LABEL_STRING, maths_name(*((MathsType *)p->value)),
                        NULL);
        }
        else if (type == BOOLEAN_SELECTOR) {
            p->slider_widget = xv_create(parent, PANEL_CHECK_BOX,
                        PANEL_LABEL_STRING, label,
                        PANEL_NEXT_ROW, -1,
                        PANEL_LAYOUT,  PANEL_HORIZONTAL,
                        PANEL_NOTIFY_PROC, BooleanCallback,
                        PANEL_CLIENT_DATA, p,
                        PANEL_VALUE, *((Boolean *)p->value),
                        NULL);
        }
        else {
            p->slider_widget = xv_create(parent, PANEL_SLIDER,
                        PANEL_NEXT_ROW, -1,
                        PANEL_LABEL_STRING, label,
                        PANEL_LABEL_BOLD, FALSE,
                        PANEL_MIN_VALUE, 0,
                        PANEL_MAX_VALUE, 100,
                        PANEL_SHOW_RANGE, FALSE,
                        PANEL_SHOW_VALUE, FALSE,
                        PANEL_NOTIFY_PROC, SliderCallback,
                        PANEL_SLIDER_WIDTH, 120,
                        PANEL_CLIENT_DATA, p,
                        PANEL_VALUE,       param,
                        NULL);

            p->value_widget = xv_create(parent, PANEL_MESSAGE, NULL);
        }
        p->next = parameters;
        parameters = p;
    }
    else {
        CsMessage(WARNING, 0, "Parameter allocation failed: Extra parameters ignored\n");
    }
}

static Frame make_params_popup(Frame form)
{
    Frame params_popup;
    Panel button_panel, main_panel;

    params_popup = xv_create(form, FRAME,
  	                            FRAME_LABEL, 	" XVCS Control Panel ",
  	                            XV_WIDTH, 350,
  	                            NULL);

    button_panel = xv_create(params_popup, PANEL,
                                PANEL_LAYOUT, PANEL_HORIZONTAL,
				NULL);

    xv_create(button_panel, PANEL_BUTTON,
  	                           PANEL_LABEL_STRING, "Default",
  	                           PANEL_NOTIFY_PROC, DefaultParamsCallback,
  	                           NULL);

    xv_create(button_panel, PANEL_BUTTON,
  	                           PANEL_LABEL_STRING, "Run",
  	                           PANEL_NOTIFY_PROC, RunCallback,
  	                           NULL);

    xv_create(button_panel, PANEL_BUTTON,
  	                           PANEL_LABEL_STRING, "Dismiss", 
  	                           PANEL_NOTIFY_PROC, DismissParamsCallback,
  	                           PANEL_CLIENT_DATA, params_popup,
  	                           NULL);

    xv_create(button_panel, PANEL_BUTTON,
  	                           PANEL_LABEL_STRING, "Save",
  	                           PANEL_NOTIFY_PROC, SaveParamsCallback,
  	                           PANEL_CLIENT_DATA, params_popup,
  	                           NULL);

    xv_create(button_panel, PANEL_BUTTON,
  	                           PANEL_LABEL_STRING, "Quit",
  	                           PANEL_NOTIFY_PROC, QuitCallback,
  	                           PANEL_CLIENT_DATA, form,
  	                           NULL);

    window_fit_height(button_panel);

    button_panel = xv_create(params_popup, PANEL,
                                PANEL_LAYOUT, PANEL_HORIZONTAL,
				NULL);

    xv_create(button_panel, PANEL_BUTTON,
  	                           PANEL_LABEL_STRING, "Initialise",
  	                           PANEL_NOTIFY_PROC, InitialiseCallback,
  	                           NULL);

    xv_create(button_panel, PANEL_BUTTON,
  	                           PANEL_LABEL_STRING, "1 Step",
  	                           PANEL_NOTIFY_PROC, StepCallback,
  	                           PANEL_CLIENT_DATA, 1,
  	                           NULL);

    xv_create(button_panel, PANEL_BUTTON,
  	                           PANEL_LABEL_STRING, "25 Steps",
  	                           PANEL_NOTIFY_PROC, StepCallback,
  	                           PANEL_CLIENT_DATA, 25,
  	                           NULL);

    xv_create(button_panel, PANEL_BUTTON,
  	                           PANEL_LABEL_STRING, "100 Steps",
  	                           PANEL_NOTIFY_PROC, StepCallback,
  	                           PANEL_CLIENT_DATA, 100,
  	                           NULL);

    window_fit_height(button_panel);

    main_panel = xv_create(params_popup, PANEL,
                                PANEL_LAYOUT, PANEL_HORIZONTAL,
                                NULL);
/*
 * Create check boxes for the display parameters:
 */

    XcsCreateParamWidgets("Display actions:",     &display_message.action,      BOOLEAN_SELECTOR, main_panel);
    XcsCreateParamWidgets("Display selections:",  &display_message.selection,   BOOLEAN_SELECTOR, main_panel);
    XcsCreateParamWidgets("Display world:",       &display_message.world,       BOOLEAN_SELECTOR, main_panel);
    XcsCreateParamWidgets("Display warnings:",    &display_message.warning,     BOOLEAN_SELECTOR, main_panel); 
    XcsCreateParamWidgets("Display activations:", &display_message.activations, BOOLEAN_SELECTOR, main_panel);

/*
 * Create sliders for the parameters:
 */

    XcsCreateParamWidgets("Rest Activation:",     &REST_ACT,            LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Persistence:",         &PERSISTENCE,         LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Maths: ",              &MATHS,               MATHS_SELECTOR, main_panel);
    XcsCreateParamWidgets("Selection Threshold:", &SELECTION_THRESHOLD, LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Noise SD:",            &NOISE_SD,            LOG_SLIDER,     main_panel);
    XcsCreateParamWidgets("Cycle Time (ms): ",    &MS_PER_CYCLE,        REAL_SLIDER,    main_panel);

    XcsCreateParamWidgets("Visual Strength: ",    &VISUAL_ROUTE_STRENGTH,   LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Visual Lag (ms): ",    &VISUAL_LAG,              INTEGER_SLIDER, main_panel);
    XcsCreateParamWidgets("Auditory Strength: ",  &AUDITORY_ROUTE_STRENGTH, LINEAR_SLIDER,  main_panel);
    XcsCreateParamWidgets("Auditory Lag (ms): ",  &AUDITORY_LAG,            INTEGER_SLIDER, main_panel);

    XcsCreateParamWidgets("Schema C:N:",          &SCHEMA_SI, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Schema P:E:",          &SCHEMA_PE, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Schema S:L:",          &SCHEMA_SL, LINEAR_SLIDER, main_panel);

    XcsCreateParamWidgets("Object C:N:",          &OBJECT_SI, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Object P:E:",          &OBJECT_PE, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Object S:L:",          &OBJECT_SL, LINEAR_SLIDER, main_panel);

    XcsCreateParamWidgets("Resource C:N:",        &EFFECT_SI, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Resource P:E:",        &EFFECT_PE, LINEAR_SLIDER, main_panel);
    XcsCreateParamWidgets("Resource S:L:",        &EFFECT_SL, LINEAR_SLIDER, main_panel);

    reset_parameter_widgets();

    window_fit_height(main_panel);

    window_fit(params_popup);

    return(params_popup);
}

/******************************************************************************/
/*  MENU GENERATION PROCEDURES FOR SCHEMA/OBJECT/RESOUCE MENUS                */
/******************************************************************************/

static void destroy_all_menu_items(Menu menu)
{
    int i;

    for (i = (int)xv_get(menu, MENU_NITEMS); i > 0; i--) {
        xv_set(menu, MENU_REMOVE, i, NULL);
        xv_destroy(xv_get(menu, MENU_NTH_ITEM, i));
    }
}

/*----------------------------------------------------------------------------*/

static Menu create_schema_menu(Menu menu, Menu_generate op)
{
    if (op == MENU_DISPLAY) {
        Menu_item mi;
        CsSchema *tmp;

        destroy_all_menu_items(menu);

        mi = xv_create(XV_NULL, MENUITEM,
                        MENU_STRING, "  NONE",
                        MENU_NOTIFY_PROC, HighlightNoSchemaCallback,
                        NULL);

        xv_set(menu, MENU_APPEND_ITEM, mi, NULL);

        for (tmp = schemas; tmp != NULL; tmp = tmp->next) {
            char label[40];
            if (tmp->highlighted) {
                sprintf(label, "* %s", tmp->name);
            }
            else {
                sprintf(label, "  %s", tmp->name);
            }
            mi = xv_create(XV_NULL, MENUITEM,
                        MENU_STRING, copy_string(label),
                        MENU_NOTIFY_PROC, HighlightSchemaCallback,
                        MENU_CLIENT_DATA, tmp,
                        NULL);
            xv_set(menu, MENU_APPEND_ITEM, mi, NULL);
        }

        mi = xv_create(XV_NULL, MENUITEM,
                    MENU_STRING, "  ALL",
                    MENU_NOTIFY_PROC, HighlightAllSchemaCallback,
                    NULL);
        xv_set(menu, MENU_APPEND_ITEM, mi, NULL);
    }
    return(menu);
}  

/*----------------------------------------------------------------------------*/

static Menu create_object_menu(Menu menu, Menu_generate op)
{
    if (op == MENU_DISPLAY) {
        Menu_item mi;
        CsObject *tmp;

        destroy_all_menu_items(menu);
       
        mi = xv_create(XV_NULL, MENUITEM,
                        MENU_STRING, "  NONE",
                        MENU_NOTIFY_PROC, HighlightNoObjectCallback,
                        NULL);

        xv_set(menu, MENU_APPEND_ITEM, mi, NULL);

        for (tmp = objects; tmp != NULL; tmp = tmp->next) {
            char label[40];
            if (tmp->highlighted) {
                sprintf(label, "* %s", tmp->name);
            }
            else {
                sprintf(label, "  %s", tmp->name);
            }
            mi = xv_create(XV_NULL, MENUITEM,
                        MENU_STRING, copy_string(label),
                        MENU_NOTIFY_PROC, HighlightObjectCallback,
                        MENU_CLIENT_DATA, tmp,
                        NULL);
            xv_set(menu, MENU_APPEND_ITEM, mi, NULL);
        }

        mi = xv_create(XV_NULL, MENUITEM,
                    MENU_STRING, "  ALL",
                    MENU_NOTIFY_PROC, HighlightAllObjectCallback,
                    NULL);
        xv_set(menu, MENU_APPEND_ITEM, mi, NULL);
    }
    return(menu);
}  

/*----------------------------------------------------------------------------*/

static Menu create_function_menu(Menu menu, Menu_generate op)
{
    if (op == MENU_DISPLAY) {
        Menu_item mi; int i;

        destroy_all_menu_items(menu);
       
        for (i = 0; i < MaxFunctions; i++) {
            if (function_name[i] != NULL) {
                mi = xv_create(XV_NULL, MENUITEM,
                            MENU_STRING, function_name[i],
                            MENU_CLIENT_DATA, i,
                            NULL);
                xv_set(menu, MENU_APPEND_ITEM, mi, NULL);
            }
        }
    }
    return(menu);
}  

/*----------------------------------------------------------------------------*/

static Menu create_resource_menu(Menu menu, Menu_generate op)
{
    if (op == MENU_DISPLAY) {
        Menu_item mi;
        CsResource *tmp;

        destroy_all_menu_items(menu);

        mi = xv_create(XV_NULL, MENUITEM,
                        MENU_STRING, "  NONE",
                        MENU_NOTIFY_PROC, HighlightNoResourceCallback,
                        NULL);

        xv_set(menu, MENU_APPEND_ITEM, mi, NULL);

        for (tmp = resources; tmp != NULL; tmp = tmp->next) {
            char label[40];
            if (tmp->highlighted) {
                sprintf(label, "* %s", tmp->name);
            }
            else {
                sprintf(label, "  %s", tmp->name);
            }
            mi = xv_create(XV_NULL, MENUITEM,
                        MENU_STRING, copy_string(label),
                        MENU_NOTIFY_PROC, HighlightResourceCallback,
                        MENU_CLIENT_DATA, tmp,
                        NULL);
            xv_set(menu, MENU_APPEND_ITEM, mi, NULL);
        }

        mi = xv_create(XV_NULL, MENUITEM,
                    MENU_STRING, "  ALL",
                    MENU_NOTIFY_PROC, HighlightAllResourceCallback,
                    NULL);
        xv_set(menu, MENU_APPEND_ITEM, mi, NULL);
    }
    return(menu);
}  

/******************************************************************************/

static void add_task_to_menu(Menu menu, TaskType task)
{
    Menu_item mi = xv_create(XV_NULL, MENUITEM,
                        MENU_STRING, task_name(task),
                        MENU_CLIENT_DATA, task,
                        NULL);
    xv_set(menu, MENU_APPEND_ITEM, mi, NULL);
}

static void SetTaskCallback(Menu menu, Menu_item item)
{
    current_task = (TaskType) xv_get(item, MENU_CLIENT_DATA);
    xv_set(task_widget, PANEL_LABEL_STRING, task_name(current_task), NULL);
    CsInitialise(current_task);
}

static Frame make_widgets()
{
    Frame frame, params_popup;
    Panel form;
    Menu menu;

    frame = xv_create(XV_NULL, FRAME,
                            XV_WIDTH, 640,
                            FRAME_LABEL, "Contention Scheduling [V6; C/Xview]",
                            NULL);

    params_popup = make_params_popup(frame);

/* The first panel: */

    form = xv_create(frame, PANEL,
                            XV_HEIGHT, 26,
			    NULL);

/* Row 1: Open the control panel */

    xv_create(form, PANEL_BUTTON,
                             PANEL_LABEL_STRING, "Run",
                             PANEL_NOTIFY_PROC, RunCallback,
                             NULL);

    xv_create(form, PANEL_BUTTON,
                             PANEL_LABEL_STRING, "Parameters...",
                             PANEL_NOTIFY_PROC, ParamsCallback,
                             PANEL_CLIENT_DATA, params_popup,
                             XV_X, 110, XV_Y, 4,
                             NULL);

    menu = xv_create(XV_NULL, MENU,
                             MENU_NOTIFY_PROC, SetTaskCallback,
                             NULL);

    add_task_to_menu(menu, COFFEE_PREPARATION);
    add_task_to_menu(menu, CEREAL_CONSUMPTION);
    add_task_to_menu(menu, BREAKFAST);
    add_task_to_menu(menu, SIMPLE_VISUAL_RT);
    add_task_to_menu(menu, SIMPLE_AUDITORY_RT);
    add_task_to_menu(menu, CHOICE_RT);
    add_task_to_menu(menu, CONFLICT_RT);

    xv_create(form, PANEL_ABBREV_MENU_BUTTON,
                             PANEL_LABEL_STRING, "Task:",
                             PANEL_ITEM_MENU, menu,
                             XV_X, 220, XV_Y, 4,
                             NULL);

    task_widget = xv_create(form, PANEL_MESSAGE,
                             PANEL_LABEL_STRING, task_name(current_task),
                             NULL);

    cycle_count_widget = xv_create(form, PANEL_MESSAGE,
                             PANEL_LABEL_STRING, "Cycles: 0",
                             PANEL_LABEL_BOLD, TRUE,
                             XV_X, 440, XV_Y, 6,
                             NULL);

    xv_create(form, PANEL_BUTTON,
                             PANEL_LABEL_STRING, "Quit",
                             PANEL_NOTIFY_PROC, QuitCallback,
                             XV_X, 550, XV_Y, 4,
                             PANEL_CLIENT_DATA, frame,
                             NULL);

/* The second panel: */ 

    form = xv_create(frame, PANEL,
                            XV_HEIGHT, 26,
                            NULL);

/* Row 1: highlight schema menu */

    menu = xv_create(XV_NULL, MENU,
                            MENU_GEN_PROC,   create_schema_menu,
                            NULL);

    xv_create(form, PANEL_BUTTON,
                            PANEL_LABEL_STRING, "Highlight",
                            PANEL_ITEM_MENU, menu,
                            NULL);

/* Row 2: The schema graph */

    plotter[0] = xv_create(frame, CANVAS,
                                XV_HEIGHT, PLOTTER_HEIGHT,
                                CANVAS_REPAINT_PROC, repaint_schema_plotter,
                                XV_KEY_DATA, PLOTTER_LABEL, "Schema Activations",
				NULL);

/* Row 3: The object buttons */

    form = xv_create(frame, PANEL,
                            XV_HEIGHT, 26,
                            NULL);

    menu = xv_create(XV_NULL, MENU,
                            MENU_GEN_PROC,   create_object_menu,
                            NULL);

    xv_create(form, PANEL_BUTTON,
                            PANEL_LABEL_STRING, "Highlight",
                            PANEL_ITEM_MENU, menu,
                            NULL);

    menu = xv_create(XV_NULL, MENU,
                            MENU_NOTIFY_PROC, DisplayFunctionCallback,
                            MENU_GEN_PROC,   create_function_menu,
                            NULL);

    xv_create(form, PANEL_BUTTON,
                            PANEL_LABEL_STRING, "Function",
                            PANEL_ITEM_MENU, menu,
                            NULL);

/* Row 4: The object graph */

    plotter[1] = xv_create(frame, CANVAS,
                                XV_HEIGHT, PLOTTER_HEIGHT,
                                CANVAS_REPAINT_PROC, repaint_object_plotter,
                                XV_KEY_DATA, PLOTTER_LABEL, "Object Activations",
				NULL);

/* Row 5: The Resource buttons */

    form = xv_create(frame, PANEL,
                            XV_HEIGHT, 26,
                            NULL);

    menu = xv_create(XV_NULL, MENU,
                            MENU_GEN_PROC,   create_resource_menu,
                            NULL);

    xv_create(form, PANEL_BUTTON,
                            PANEL_LABEL_STRING, "Highlight",
                            PANEL_ITEM_MENU, menu,
                            NULL);

/* Row 6: The Resource graph */

     plotter[2] = xv_create(frame, CANVAS,
                                XV_HEIGHT, PLOTTER_HEIGHT,
                                CANVAS_REPAINT_PROC, repaint_resource_plotter,
                                XV_KEY_DATA, PLOTTER_LABEL, "Resource Activations",
				NULL);

    window_fit(frame);
    return(frame);
}

static void xview_error(Xv_object object, Attr_avlist avlist)
{
    fprintf(stderr, xv_error_format(object, avlist));
}

/*
 *      Main
 */

int main(int argc, char **argv)
{
    Frame        context;

    xv_init(XV_INIT_ARGC_PTR_ARGV, &argc, argv, XV_ERROR_PROC, xview_error, NULL);

    CsSetDefaultParameters(current_task);

/* Must initialise schemata etc BEFORE making the form, so as to get correct menus */

    CsInitialise(current_task);

    if ((context = make_widgets()) != XV_NULL) {
        Rect    image_rect, label_rect;
        Server_image closed_image;
        Icon         icon;

        closed_image = (Server_image) xv_create(XV_NULL, SERVER_IMAGE,
                        XV_WIDTH,           64,
                        XV_HEIGHT,          64,
                        SERVER_IMAGE_BITS,  icon_bits,
                        NULL);
        rect_construct(&image_rect, 0, 0, 64, 64);
        rect_construct(&label_rect, 0, 68, 64, 12);

        icon = (Icon) xv_create(context, ICON,
                        XV_WIDTH,           64,
                        XV_HEIGHT,          80,
                        ICON_LABEL,         "xvcs",
                        ICON_LABEL_RECT,    &label_rect,
                        ICON_IMAGE,         closed_image,
                        ICON_IMAGE_RECT,    &image_rect,
                        ICON_TRANSPARENT,   TRUE,
                        NULL);

        xv_set(context, FRAME_ICON, icon, NULL);

        create_gcs(context);

        AttachData(0, TRUE);

        xv_main_loop(context);
    }

    exit(0);
}

/******** END OF FILE *********************************************************/
